/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CORBAGeneral
// @module  CORBAGeneral | This module contains the implementation of the
//          CORBAGeneral class.  This class is responsible for hiding low-level
//          CORBA functionality and wrapping it up into an easy to use object.
//          This particular class is the base for the client and server
//          classes.  It contains any generic functionality.  There should
//          never be any need to construct an instance of this object in its
//          own right.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CORBA/CORBAGeneral.h>
#include <CORBA/CORBAHelper.inl>
using namespace CORBA;
using namespace CosNaming;

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates an instance of a CORBAGeneral object.  It
//          does not construct the ORB and BOA.  Instead they are passed as
//          parameters.
// @parm    The current ORB object.
// @parm    The current BOA object.
CORBAGeneral::CORBAGeneral( ORB_ptr orb, BOA_ptr boa )
{
    // Initialise ORB and BOA
    m_orb = orb;
    m_boa = boa;

    // This instance of the class did not start the ORB.
    m_bInit = false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates an instance of a CORBAGeneral object.  It
//          creates the ORB and BOA.
// @parm    The number of command line arguments.
// @parm    The array of command line arguments.
CORBAGeneral::CORBAGeneral( int argc, char** argv )
{
    // Start CORBA object and BOA.  This is very important!!!  It must be done
    // before anything else to do with CORBA otherwise things will start
    // crashing.
    m_orb = CORBA::ORB_init( argc, argv, "omniORB2" );
    m_boa = m_orb->BOA_init( argc, argv, "omniORB2_BOA" );

    // This class did start the ORB
    m_bInit = true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor needs to shutdown the ORB and BOA if this instance
//          of the class created them.
// @devnote The command to shutdown the BOA is impl_shutdown.  This used to be
//          destroy, but this meant that the BOA could not be started again
//          without restarting the entire application.
CORBAGeneral::~CORBAGeneral()
{
    // If this instance of the class started the ORB then
    // we need to shut down the ORB.
    if ( m_bInit == true )
    {
        // Shutdown BOA
        //      m_boa->impl_shutdown();
    }

    // Dispose of ORB and BOA references

    CORBA::release( m_orb );
    CORBA::release( m_boa );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method informs the BOA that it is ready to begin processing.
//          It is possible to block application execution or not by using this
//          method.
// @parm    If this parameter is false the method blocks, otherwise application
//          processing will continue.
void CORBAGeneral::Ready( bool cont )
{
    // Tell BOA to block or not
    m_boa->impl_is_ready( 0, cont );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a reference to the NameService or throws an
//          exception if one cannot be found.
// @rdesc   A reference to the NameService.
NamingContext_ptr CORBAGeneral::GetNamingService()
{
    // Return reference to NameService
    return resolve_init((CosNaming::NamingContext*)NULL, m_orb, (const char*)"NameService" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a list of elements registered with the
//          NameService.  The returned names are separated by a SEPARATOR
//          character. If no type is specified then all registered elements are
//          returned along with their types.  These are also separated by
//          SEPARATOR characters.
// @rdesc   The character string containing all the names, or the names and
//          their types.
// @parm    The string name of the type of object we are looking for. Defaults
//          to NULL.
string CORBAGeneral::GetRegisteredList( const char* type /* = NULL */ )
{
    string s = "";

    // Get a reference to the NameService
    CosNaming::NamingContext_var context = GetNamingService();

    // Get a list of all contexts on NameService
    CosNaming::BindingIterator_var bi;
    CosNaming::BindingList_var bl;
    CosNaming::Binding_var b;

    context->list( 0, bl, bi );

    // Loop through all contexts and get all objects
    while ( bi->next_one(b) )
    {
        // If type is NULL then add name AND type to list
        if ( type == NULL )
        {
            // Append name to string
            s += b->binding_name[0].id;
            s += SEPARATOR;
            s += b->binding_name[0].kind;
            s += SEPARATOR;
        }
        else if ( strcmp( type, b->binding_name[0].kind ) == 0 )
        {
            // Append name to string
            s += b->binding_name[0].id;
            s += SEPARATOR;
        }
    }

    // Add final terminator
    s += SEPARATOR;

    // Tidy up
    bi->destroy();

    return s;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a reference to a named object from the
//          NameService.  The object reference will still need to be narrowed
//          by the calling method.
// @rdesc   The object reference or nil if the object could not be located.
// @parm    The name of the object to locate.
// @parm    The type of the object to locate.
Object_ptr CORBAGeneral::GetObjectFromClass( const char* cls,
        const char* name )
{
    // No errors - yet
    int status = CYR_UNKNOWN_ERROR;
    const char* where = "CORBAGeneral::GetObjectFromClass";
    const char* error = NULL;

    // Get a reference to the NameService
    CosNaming::NamingContext_var nc = GetNamingService();

    // Construct name object
    CosNaming::Name n;
    n.length( 1 );
    n[0].id = name;
    n[0].kind = cls;

    try
    {
        // Resolve name object
        return nc->resolve( n );
    }
    catch( CORBA::Exception )
    {
        // If it is a CORBA exception then the error is that the requested
        // object cannot be found.
        error = "Could not resolve object";
        status = CYR_CORBA_OBJECT_NOT_FOUND;
    }
    catch( ... )
    {
        // Catch all other exceptions
        error = "Unknown exception";
    }

    // If we have an error then throw an exception
    if ( error != NULL )
        throw CyrCORBAException( error, where, status );

    return CORBA::Object::_nil();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a reference to a CORBA object by using an IOR
//          string.  The object reference will still need to be narrowed by the
//          calling method.
// @rdesc   The object reference or nil if the object could not be located.
// @parm    The IOR of the object to locate.
Object_ptr CORBAGeneral::GetObjectFromIOR( const char* ior )
{
    return m_orb->string_to_object( ior );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns the ORB pointer.
// @rdesc   The orb reference.
ORB_ptr CORBAGeneral::GetORB()
{
    return m_orb;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns the BOA pointer.
// @rdesc   The BOA reference.
BOA_ptr CORBAGeneral::GetBOA()
{
    return m_boa;
}

