/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// @doc
//
// @module CECTimer.cpp
//
// Implementation file for the <CECTimer> Class
//
// This class provides us with a mechanism to start, stop and get the slapsed times
// for timers with a granularity of 1 millisecond.
//
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


#include "CECTimer.h"

#if defined (UNIX) || defined (WIN32)
#include <time.h>           /* Time facilities */
#endif

//
// LINUX Port Changes
//
#ifdef LINUX
#include <sys/time.h>
#include <unistd.h>
#endif

#ifdef  WIN32
#include <sys/timeb.h>
#endif

#define MSECS_IN_24HRS 86400000
#define PSL1JAN70 40587                             // 1-JAN-1970 as PSL date

//
// @func Constructor
//
CECTimer::CECTimer()
{
    m_ElapsedTime = 0;
    m_Running = false;
}

//
// @func Destructor
//
CECTimer::~CECTimer()
{}

//
// @func    Start
//      Start the timer running. Note that this over-rides any other timer that
//      may alraedy be running.
//

int     CECTimer::Start ()
{
    GetTime (&m_StartDate ,&m_StartTime);
    m_Running = true;
    return 0;
}

//
// @func    Stop
//      Stop an existing running timer. Note that if the timer is NOT active we simply return,
//      otherwise we return the time which has elapsed between the start and this stop as a number
//      of 10 millisecond units (1/10th second units)
//
int     CECTimer::Stop  ()
{
    //
    //    If the timer is not active then simply return a status of -1
    //
    if (!m_Running) return -1;
    m_Running = false;                  // Time is no longer running

    //
    //    Get the time at which this timer stopped and calculate the alapsed period.
    //
    int StopDate,StopTime;
    GetTime (&StopDate ,&StopTime);

    SetElapsedTime (StopDate ,StopTime);
    return m_ElapsedTime;
}


//
// @func    SetElapsedTime
//
//      Calculate the time in centiseconds which has elapsed between a StartTimer and the time
//      supplied to this routine.
// @parm Date at which the timer was stopped (as returned from GetTime)
// @parm Time at which the timer was stopped (as returned from GetTime)
//
void CECTimer::SetElapsedTime (int StopDate ,int StopTime)
{

    //
    //  If the start date and end date are the same, the elapsed time
    //  is obtained by simply subtracting the earlier time from the later
    //  time and then dividing the result by 10 to get the time in
    //  centi-seconds.
    //
    if (StopDate == m_StartDate)
    {
        m_ElapsedTime = (StopTime - m_StartTime) / 10;
    }

    //
    //  Else if the times are from adjacent days, the elapsed time is obtained
    //  by adding the difference between the start time and midnight on the
    //  start date to the time for the end date and then dividing the result by
    //  10 to get the time in centi-seconds.
    //
    else if ((StopDate - m_StartDate) == 1)
    {
        m_ElapsedTime = ((MSECS_IN_24HRS - m_StartTime) + StopTime) / 10;
    }

    //
    //  Else there are a number of complete days between the start date and
    //  the end date. In this case, the elapsed time is obtained
    //  by adding the number of milliseconds for the number of COMPLETE
    //  days between the start date and the end date to the difference between
    //  the start time and midnight on the start date to the time for the end
    //  date and then dividing the result by 10 to get the time in
    //  centi-seconds.
    //
    else
    {
        m_ElapsedTime = (((StopDate - m_StartDate - 1) * MSECS_IN_24HRS) +
                         (MSECS_IN_24HRS - m_StartTime) + StopTime) / 10;
    }

    //
    //  Return to caller.
    //
    return;
}


//
//  @mfunc GetTime
//
//      Recover the current date and time
//
#ifdef  UNIX
void    CECTimer::GetTime   (int* pDate ,int* pTime)
{

    struct timeval  timenow;
    struct timezone zone;
    int      days;
    int      seconds;
#define PSL1JAN70 40587

    //
    //    set up the timezone information before getting the time of day and use it to make any
    //    timezone adjustements
    tzset ();
    (void) gettimeofday (&timenow, &zone);
    timenow.tv_sec -= timezone;

    //
    //  The daylight external variable shows whether we should make a summertime
    //  adjustment, 1 means yes, 0 means no and -1 means unavavilable.
    //
    if (daylight != 0) timenow.tv_sec += (60*60);
    days = timenow.tv_sec / (24*60*60);         // Calculate days since 1-Jan-1970
    *pDate = days + PSL1JAN70;                  // Convert to PSL date
    seconds = timenow.tv_sec % (24*60*60);      // Calculate seconds in current day
    *pTime = (seconds * 1000) + (timenow.tv_usec / 1000);
    // Convert to PSL time
    return;
}

#elif WIN32
void    CECTimer::GetTime   (int* pDate ,int* pTime)
{
    struct _timeb tstruct;
    long   days;             // Days since 1-JAN-1970
    long   seconds;          // Seconds in current day

    //
    //    Set time zone from TZ environment variable. If TZ is not set, operating system default is used,
    //    otherwise PST8PDT is used (Pacific standard time, daylight savings).
    //
    _tzset();

    //
    //    Get time in seconds since UTC
    //
    _ftime(&tstruct);

    //
    //    Modify for time zone notining that the time zone is the difference in minutes, moving westward,
    //    between UTC and local time
    //
    tstruct.time -= (tstruct.timezone * 60);

    //
    //  The daylight external variable shows whether we should make a summertime adjustment.
    //
    if (tstruct.dstflag) tstruct.time += (60*60);
    days = tstruct.time / (24*60*60);               // Calculate days since 1-Jan-1970
    *pDate = days + PSL1JAN70;                      // Convert to PSL date
    seconds = tstruct.time % (24*60*60);            // Calculate seconds in current day
    *pTime = (seconds * 1000) + tstruct.millitm;    // Convert to PSL time
    return;
}
#else
#error message "Platform not supported yet for routine GetTime"
#endif
