/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/* This file is common to the whole project,
 * and is therefore to be included quasi-systematically.
 * It deals with project-wide macros and library choices.
 * IMPORTANT: because under UNIX multithreaded program source
 * files must include the platform-dependant header containing
 * thread-related information as their FIRST system header,
 * the current include file must NOT contain any nested #include
 * statement.
 */


#ifndef __DEFINITION__
#define __DEFINITION__


// cyrtools log
#define CYRTOOLS_DEBUG "cyrtools"

#ifdef _WIN32
#ifndef WIN32 //make sure WIN32 defined
#    define WIN32   1
#endif
#    define DIR_SLASH '\\'
#else
#    define DIR_SLASH '/'
#endif


/* Definitions of system macros used for compile time.
 * Some macros are predefined, so we use them to define our macros.
 * WIN32 : Windows 95/NT.
 *         Must be defined in project settings.
 * UNIX  : Unix (5.4).
 *         We try to define UNIX using the predefined macros.
 * Our macros are :
 * MCH_OSTYPE     : Machine operating system type.
 *                  This macro is defined to determine the
 *                  platform on wich we are compiling :
 *   OSTYPE_WIN32 : Windows 95/NT.
 *   OSTYPE_UNIX  : Unix (5.4).
 * MCH_OS     : Machine real operating system.
 *              This macro is defined to determine the
 *              operating system on wich we are compiling :
 *   OS_WINNT : Windows NT.
 *   OS_WIN95 : Windows 95.
 *   OS_DGUX  : Data General DG/UX.
 *   OS_AIX   : IBM AIX.
 * MCH_ENDIAN      : Machine char order.
 *                   This macro is defined to determine the
 *                   machine char order.
 *   LITTLE_ENDIAN : Little endian type (Intel).
 *   BIG_ENDIAN    : Big endian type (Motorola).
 */

/* Predefined macros for UNIX. */
#ifndef DGUX
#  ifdef __DGUX__
#    define DGUX
#  endif        /* __DGUX__ */
#endif  /* DGUX */

#ifndef AIX
#  if defined (_AIX) || defined (_AIX32) || defined (_AIX41)
#    define AIX
#  endif        /* _AIX || _AIX32 || _AIX41 */
#endif  /* AIX */

#ifndef HPUX
#  if defined (__hpux) || defined (hpux)
#    define HPUX
#  endif        /* __hpux || hpux */
#endif  /* HPUX */

#ifndef UNIX
#  if defined (DGUX) || defined (AIX) || defined (HPUX) || defined (SOLARIS) || defined (ALPHA) || defined (LINUX)
#    define UNIX
#  else
#    if defined (__unix__) || defined (__UNIX__) || defined (unix)
#      define UNIX
#    endif      /* __unix__ || __UNIX__ || unix */
#  endif        /* DGUX || AIX */
#endif  /* UNIX */

/* Operating system types. */
#define OSTYPE_WIN32    1
#define OSTYPE_UNIX             2

/* Operating systems. */
#define OS_WINNT                1
#define OS_WIN95                2
#define OS_DGUX                 3
#define OS_AIX                  4
#define OS_HPUX                 5
#define OS_SOLARIS              6
#define OS_ALPHA                7
#define OS_LINUX                8

#if !defined(AIX) && !defined(LINUX) &&  !defined(ALPHA)

/* Machine char order. */
# ifdef LITTLE_ENDIAN
#  undef LITTLE_ENDIAN
# endif

# define LITTLE_ENDIAN   1

# ifdef BIG_ENDIAN
#  undef BIG_ENDIAN
# endif

# define BIG_ENDIAN 2

#endif

/* on AIX LITTLE/BIG ENDIAN are defined in sys/machine.h */
#if defined(AIX)
#include   <sys/machine.h>
#endif

/* on ALPHA LITTLE/BIG ENDIAN are defined in machine/endian.h */
#if defined(ALPHA)
#include   <machine/endian.h>
#endif


/* Macros for Windows 95/NT. */
#ifdef WIN32
#  define MCH_OSTYPE    OS_TYPEWIN32
#  ifdef _WIN32_WINNT
#    define MCH_OS              OS_WINNT
#  else
#    define MCH_OS              OS_WIN95
#  endif
#  define MCH_ENDIAN    LITTLE_ENDIAN   /* !! Alpha is not little endian !! */
#endif  /* WIN32 */

/* Macros for UNIX. */
#ifdef UNIX
#  define MCH_OSTYPE    OSTYPE_UNIX
#endif

#ifdef DGUX
#  define MCH_OS              OS_DGUX
#  define MCH_ENDIAN    BIG_ENDIAN
#elif defined AIX
#  define MCH_OS              OS_AIX
#  define MCH_ENDIAN    BIG_ENDIAN
#elif defined HPUX
#  define MCH_OS              OS_HPUX
#  define MCH_ENDIAN    BIG_ENDIAN
#elif defined SOLARIS
#  define MCH_OS              OS_SOLARIS
#  define MCH_ENDIAN    BIG_ENDIAN
#elif defined ALPHA
#  define MCH_OS              OS_ALPHA
#  define MCH_ENDIAN    LITTLE_ENDIAN
#elif defined LINUX
#  define MCH_OS              OS_LINUX
#  define MCH_ENDIAN    LITTLE_ENDIAN
#endif
/* !! Unixware is not on big endian !! */

/* Verify the macros. */
#ifndef MCH_OSTYPE
#  error "MCH_OSTYPE not defined."
#endif  /* MCH_OSTYPE */
#ifndef MCH_OS
#  error "MCH_OS not defined."
#endif  /* MCH_OS */
#ifndef MCH_ENDIAN
#  error "MCH_ENDIAN not defined."
#endif  /* MCH_ENDIAN */


/* Definitions of class export/import attributes.
 * Used to specify if the classes of a module are
 * to be exported or imported.
 */

/* Macros for Windows 95/NT. */
#ifdef WIN32
#  define __IMPORT__    __declspec (dllimport)
#  define __EXPORT__    __declspec (dllexport)
#endif  /* WIN32 */

#ifdef UNIX
#  define __IMPORT__
#  define __EXPORT__
#endif  /* UNIX */


/* Library export/import attributes macro constructor. */
#define DLLSPEC(name) DLLSPEC_##name

/* Predefined module macro can be one of the following :
 * BUILDDLLCYRTOOLS   : Cyrano Tools Library.
 * BUILDDLLCYRACOM : Cyrano Communication Library.
 * BUILDDLLxxxx : 
 */

/* SAMPLE: (xxxx is the module name).
 * #if defined(BUILDDLLxxxx)
 * #   define DLLSPEC_xxxx __EXPORT__
 * #else
 * #   define DLLSPEC_xxxx __IMPORT__
 * #endif
 */

/* TOOLS Library. */
#ifdef BUILDDLLCYRTOOLS
#  define DLLSPEC_CYRTOOLS __EXPORT__
#else
#  ifdef BUILD_STATIC
#    define DLLSPEC_CYRTOOLS
#  else
#    define DLLSPEC_CYRTOOLS __IMPORT__
#  endif
#endif  /* BUILDDLLTOOLS */


#if (__SUNPRO_CC == 0x500)
#define SUNPRO5
#endif


/* There are a lot of 4251 warning from Rogue includes.
   C4251 : 'identifier' : class 'type' needs to have dll-interface to be used by clients of class 'type2'.
   There are a lot of 4710 warning for Release target.
   C4710 : 'function' : function not inlined.
*/
#ifdef  WIN32
#  pragma warning(disable : 4251 4710 4786)
#endif  // WIN32
// This #define determines if a class is to be imported or exported from the
// including code.
#ifdef _WIN32
#  ifdef __CYR_STATIC__
#    define CYR_EXPORT
#  else
#    ifdef __CYR_DLL__
#      define CYR_EXPORT __declspec(dllexport)
#    else
#      define CYR_EXPORT __declspec(dllimport)
#    endif
#  endif
#else
#  define CYR_EXPORT // Non NT platforms
#endif


#endif /* __DEFINITION__ */
