/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __FUNCTOR__
#define __FUNCTOR__

//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifndef __EXCEPTION__
#  include <exception.hpp>
#endif  // __EXCEPTION__

#ifndef __MTSAFE__
#  include <mtsafe.hpp>
#endif  // __MTSAFE__


//===================================================================< Macros >
#define DLLSPEC_FUNCTOR DLLSPEC (CYRTOOLS)


//====================================================================< Class >

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// Class CTFunctorException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep functors exception.
class DLLSPEC_FUNCTOR CTFunctorException : public CTException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CTFunctorException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CTFunctorException& exception)
    {
        CTException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CTFunctorException

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////// End of Class CTFunctorException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Classes & Templates Classes Functor //
///////////////////////////////////////////////////////////////////////////////
// Functor is an object used to represent a function invocation (pointer to the
// function, arguments types and values, return type and value). This object
// can be passed to the thread classes constructors or service classes
// constructors to execute the function.
// Exemples :
// void vv (void) { }
// int ii (int) { return 1 ; }
//
// void titi (void)
// {
//  // vv has no arguments and return nothing.
//  CTFunctor VV = makeFunctorA0 (vv) ;     // Make a functor.
//  VV.call () ;                            // Call the function.
//  VV.getResultPointer () ;                // Generate an exception.
//
//  // ii has an integer (value 1) as argument and return an integer.
//  CTFunctor II = makeFunctorRA1 (int, ii, int, 1) ;   // Make a functor.
//  II.call () ;                                        // Call the function.
//  int i = *(int*) (II.getResultPointer () ) ;         // Get the return value.
// } ;

class CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                     // Control methodes.
    virtual void            call (void) = 0 ;
    virtual void*           getResultPointer (void) = 0 ;
    virtual CTFunctorXX*    duplicate (void) = 0 ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorXX

class CTFunctorA0 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:             // Types.
    typedef void (*func_t) (void) ;

public:             // Constructor(s) and destructor.
    CTFunctorA0 (func_t func)
: func_ (func)
    { }
    ;

public:             // Operators.
    CTFunctorA0&    operator= (const CTFunctorA0& functorA0)
    {
        func_ = functorA0.func_ ;
        return *this ;
    } ;

public:             // Control methodes.
    void            call (void)
    {
        func_ () ;
    } ;
    void*           getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*    duplicate (void)
    {
        return new CTFunctorA0 (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:            // Members.
    func_t          func_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA0

template <class A1> class CTFunctorA1 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                 // Types.
    typedef void (*func_t) (A1) ;

public:                 // Constructor(s) and destructor.
    CTFunctorA1 (func_t func, A1 a1)
: func_ (func), a1_ (a1)
    { }
    ;

public:                 // Operators.
    CTFunctorA1<A1>&    operator= (const CTFunctorA1<A1>& functorA1)
    {
        func_ = functorA1.func_ ;
        a1_ = functorA1.a1_ ;
        return *this ;
    } ;

public:                 // Control methodes.
    void                call (void)
    {
        func_ (a1_) ;
    } ;
    void*               getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*        duplicate (void)
    {
        return new CTFunctorA1<A1> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                // Members.
    func_t              func_ ;
    A1                  a1_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA1

template <class A1, class A2> class CTFunctorA2 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                     // Types.
    typedef void (*func_t) (A1, A2) ;

public:                     // Constructor(s) and destructor.
    CTFunctorA2 (func_t func, A1 a1, A2 a2)
: func_ (func), a1_ (a1), a2_ (a2)
    { }
    ;

public:                     // Operators.
    CTFunctorA2<A1, A2>&    operator= (const CTFunctorA2<A1, A2>& functorA2)
    {
        func_ = functorA2.func_ ;
        a1_ = functorA2.a1_ ;
        a2_ = functorA2.a2_ ;
        return *this ;
    } ;

public:                     // Control methodes.
    void                    call (void)
    {
        func_ (a1_, a2_) ;
    } ;
    void*                   getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*            duplicate (void)
    {
        return new CTFunctorA2<A1, A2> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                    // Members.
    func_t                  func_ ;
    A1                      a1_ ;
    A2                      a2_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA2

template <class A1, class A2, class A3> class CTFunctorA3 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    typedef void (*func_t) (A1, A2, A3) ;

public:                         // Constructor(s) and destructor.
    CTFunctorA3                 (func_t func, A1 a1, A2 a2, A3 a3)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3)
    { }
    ;

public:                         // Operators.
    CTFunctorA3<A1, A2, A3>&    operator= (const CTFunctorA3<A1, A2, A3>& functorA3)
    {
        func_ = functorA3.func_ ;
        a1_ = functorA3.a1_ ;
        a2_ = functorA3.a2_ ;
        a3_ = functorA3.a3_ ;
        return *this ;
    } ;

public:                         // Control methodes.
    void                        call (void)
    {
        func_ (a1_, a2_, a3_) ;
    } ;
    void*                       getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*                duplicate (void)
    {
        return new CTFunctorA3<A1, A2, A3> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                        // Members.
    func_t                      func_ ;
    A1                          a1_ ;
    A2                          a2_ ;
    A3                          a3_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA3

template <class A1, class A2, class A3, class A4> class CTFunctorA4 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    typedef void (*func_t) (A1, A2, A3, A4) ;

public:                         // Constructor(s) and destructor.
    CTFunctorA4                 (func_t func, A1 a1, A2 a2, A3 a3, A4 a4)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_(a4)
    { }
    ;

public:                         // Operators.
    CTFunctorA4<A1, A2, A3, A4>&    operator= (const CTFunctorA4<A1, A2, A3, A4>& functorA4)
    {
        func_ = functorA4.func_ ;
        a1_ = functorA4.a1_ ;
        a2_ = functorA4.a2_ ;
        a3_ = functorA4.a3_ ;
        a4_ = functorA4.a4_ ;
        return *this ;
    } ;

public:                         // Control methodes.
    void                        call (void)
    {
        func_ (a1_, a2_, a3_, a4_) ;
    } ;
    void*                       getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*                duplicate (void)
    {
        return new CTFunctorA4<A1, A2, A3, A4> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                        // Members.
    func_t                      func_ ;
    A1                          a1_ ;
    A2                          a2_ ;
    A3                          a3_ ;
    A4                          a4_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA4

template <class A1, class A2, class A3, class A4, class A5> class CTFunctorA5 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    typedef void (*func_t) (A1, A2, A3, A4, A5) ;

public:                         // Constructor(s) and destructor.
    CTFunctorA5                 (func_t func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_(a4), a5_(a5)
    { }
    ;

public:                         // Operators.
    CTFunctorA5<A1, A2, A3, A4, A5>&    operator= (const CTFunctorA5<A1, A2, A3, A4, A5>& functorA5)
    {
        func_ = functorA5.func_ ;
        a1_ = functorA5.a1_ ;
        a2_ = functorA5.a2_ ;
        a3_ = functorA5.a3_ ;
        a4_ = functorA5.a4_ ;
        a5_ = functorA5.a5_ ;
        return *this ;
    } ;

public:                         // Control methodes.
    void                        call (void)
    {
        func_ (a1_, a2_, a3_, a4_, a5_) ;
    } ;
    void*                       getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*                duplicate (void)
    {
        return new CTFunctorA5<A1, A2, A3, A4, A5> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                        // Members.
    func_t                      func_ ;
    A1                          a1_ ;
    A2                          a2_ ;
    A3                          a3_ ;
    A4                          a4_ ;
    A5                          a5_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA5

template <class A1, class A2, class A3, class A4, class A5, class A6> class CTFunctorA6 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    typedef void (*func_t) (A1, A2, A3, A4, A5, A6) ;

public:                         // Constructor(s) and destructor.
    CTFunctorA6                 (func_t func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_(a4), a5_(a5), a6_(a6)
    { }
    ;

public:                         // Operators.
    CTFunctorA6<A1, A2, A3, A4, A5, A6>&    operator= (const CTFunctorA6<A1, A2, A3, A4, A5, A6>& functorA6)
    {
        func_ = functorA6.func_ ;
        a1_ = functorA6.a1_ ;
        a2_ = functorA6.a2_ ;
        a3_ = functorA6.a3_ ;
        a4_ = functorA6.a4_ ;
        a5_ = functorA6.a5_ ;
        a6_ = functorA6.a6_ ;
        return *this ;
    } ;

public:                         // Control methodes.
    void                        call (void)
    {
        func_ (a1_, a2_, a3_, a4_, a5_, a6_) ;
    } ;
    void*                       getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*                duplicate (void)
    {
        return new CTFunctorA6<A1, A2, A3, A4, A5, A6> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                        // Members.
    func_t                      func_ ;
    A1                          a1_ ;
    A2                          a2_ ;
    A3                          a3_ ;
    A4                          a4_ ;
    A5                          a5_ ;
    A6                          a6_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA6

template <class A1, class A2, class A3, class A4, class A5, class A6, class A7> class CTFunctorA7 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    typedef void (*func_t) (A1, A2, A3, A4, A5, A6, A7) ;

public:                         // Constructor(s) and destructor.
    CTFunctorA7                 (func_t func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6, A7 a7)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_(a4), a5_(a5), a6_(a6), a7_(a7)
    { }
    ;

public:                         // Operators.
    CTFunctorA7<A1, A2, A3, A4, A5, A6, A7>&    operator= (const CTFunctorA7<A1, A2, A3, A4, A5, A6, A7>& functorA7)
    {
        func_ = functorA7.func_ ;
        a1_ = functorA7.a1_ ;
        a2_ = functorA7.a2_ ;
        a3_ = functorA7.a3_ ;
        a4_ = functorA7.a4_ ;
        a5_ = functorA7.a5_ ;
        a6_ = functorA7.a6_ ;
        a7_ = functorA7.a7_ ;
        return *this ;
    } ;

public:                         // Control methodes.
    void                        call (void)
    {
        func_ (a1_, a2_, a3_, a4_, a5_, a6_, a7_) ;
    } ;
    void*                       getResultPointer (void)
    {
        return NULL ;
    } ;
    CTFunctorXX*                duplicate (void)
    {
        return new CTFunctorA7<A1, A2, A3, A4, A5, A6, A7> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                        // Members.
    func_t                      func_ ;
    A1                          a1_ ;
    A2                          a2_ ;
    A3                          a3_ ;
    A4                          a4_ ;
    A5                          a5_ ;
    A6                          a6_ ;
    A7                          a7_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorA7

template <class R> class CTFunctorRA0 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                 // Types.
    typedef R (*func_t) (void) ;

public:                 // Constructor(s) and destructor.
    CTFunctorRA0        (func_t func)
: func_ (func)
    {
        uSize_ = sizeof (CTFunctorRA0<R>) ;
    } ;

public:                 // Operators.
    CTFunctorRA0<R>&    operator= (const CTFunctorRA0<R>& functorRA0)
    {
        r_ = functorRA0.r_ ;
        func_ = functorRA0.func_ ;
        return *this ;
    } ;

public:                 // Control methodes.
    void                call (void)
    {
        r_ = func_ () ;
    } ;
    void*               getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*        duplicate (void)
    {
        return new CTFunctorRA0<R> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                // Members.
    R                   r_ ;
    func_t              func_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorRA0

template <class R, class A1> class CTFunctorRA1 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                     // Types.
    typedef R (*func_t) (A1) ;

public:                     // Constructor(s) and destructor.
    CTFunctorRA1            (func_t func, A1 a1)
: func_ (func), a1_ (a1)
    { }
    ;

public:                     // Operators.
    CTFunctorRA1<R, A1>&    operator= (const CTFunctorRA1<R, A1>& functorRA1)
    {
        r_ = functorRA1.r_ ;
        func_ = functorRA1.func_ ;
        a1_ = functorRA1.a1_ ;
        return *this ;
    } ;

public:                     // Control methodes.
    void                    call (void)
    {
        r_ = func_ (a1_) ;
    } ;
    void*                   getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*            duplicate (void)
    {
        return new CTFunctorRA1<R, A1> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                    // Members.
    R                       r_ ;
    func_t                  func_ ;
    A1                      a1_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorRA1

template <class R, class A1, class A2> class CTFunctorRA2 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                         // Types.
    typedef R (*func_t) (A1, A2) ;

public:                         // Constructor(s) and destructor.
    CTFunctorRA2                (func_t func, A1 a1, A2 a2)
: func_ (func), a1_ (a1), a2_ (a2)
    { }
    ;

public:                         // Operators.
    CTFunctorRA2<R, A1, A2>&    operator= (const CTFunctorRA2<R, A1, A2>& functorRA2)
    {
        r_ = functorRA2.r_ ;
        func_ = functorRA2.func_ ;
        a1_ = functorRA2.a1_ ;
        a2_ = functorRA2.a2_ ;
        return *this ;
    } ;

public:                         // Control methodes.
    void                        call (void)
    {
        r_ = func_ (a1_, a2_) ;
    } ;
    void*                       getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*                duplicate (void)
    {
        return new CTFunctorRA2<R, A1, A2> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                        // Members.
    R                           r_ ;
    func_t                      func_ ;
    A1                          a1_ ;
    A2                          a2_ ;
}
; // CTFunctorRA2

template <class R, class A1, class A2, class A3> class CTFunctorRA3 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                             // Types.
    typedef R (*func_t) (A1, A2, A3) ;

public:                             // Constructor(s) and destructor.
    CTFunctorRA3                    (func_t func, A1 a1, A2 a2, A3 a3)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3)
    { }
    ;

public:                             // Operators.
    CTFunctorRA3<R, A1, A2, A3>&    operator= (const CTFunctorRA3<R, A1, A2, A3>& functorRA3)
    {
        r_ = functorRA3.r_ ;
        func_ = functorRA3.func_ ;
        a1_ = functorRA3.a1_ ;
        a2_ = functorRA3.a2_ ;
        a3_ = functorRA3.a3_ ;
        return *this ;
    } ;

public:                             // Control methodes.
    void                            call (void)
    {
        r_ = func_ (a1_, a2_, a3_) ;
    } ;
    void*                           getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*                    duplicate (void)
    {
        return new CTFunctorRA3<R, A1, A2, A3> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                            // Members.
    R                               r_ ;
    func_t                          func_ ;
    A1                              a1_ ;
    A2                              a2_ ;
    A3                              a3_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorRA3

template <class R, class A1, class A2, class A3, class A4> class CTFunctorRA4 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                                 // Types.
    typedef R (*func_t) (A1, A2, A3, A4) ;

public:                                 // Constructor(s) and destructor.
    CTFunctorRA4                        (func_t func, A1 a1, A2 a2, A3 a3, A4 a4)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_ (a4)
    { }
    ;

public:                                 // Operators.
    CTFunctorRA4<R, A1, A2, A3, A4>&    operator= (const CTFunctorRA4<R, A1, A2, A3, A4>& functorRA4)
    {
        r_ = functorRA4.r_ ;
        func_ = functorRA4.func_ ;
        a1_ = functorRA4.a1_ ;
        a2_ = functorRA4.a2_ ;
        a3_ = functorRA4.a3_ ;
        a4_ = functorRA4.a4_ ;
        return *this ;
    } ;

public:                                 // Control methodes.
    void                                call (void)
    {
        r_ = func_ (a1_, a2_, a3_, a4_) ;
    } ;
    void*                               getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*                        duplicate (void)
    {
        return new CTFunctorRA4<R, A1, A2, A3, A4> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                                // Members.
    R                                   r_ ;
    func_t                              func_ ;
    A1                                  a1_ ;
    A2                                  a2_ ;
    A3                                  a3_ ;
    A4                                  a4_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorRA4

template <class R, class A1, class A2, class A3, class A4, class A5> class CTFunctorRA5 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                                     // Types.
    typedef R (*func_t) (A1, A2, A3, A4, A5) ;

public:                                     // Constructor(s) and destructor.
    CTFunctorRA5                            (func_t func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_ (a4), a5_ (a5)
    { }
    ;

public:                                     // Operators.
    CTFunctorRA5<R, A1, A2, A3, A4, A5>&    operator= (const CTFunctorRA5<R, A1, A2, A3, A4, A5>& functorRA5)
    {
        r_ = functorRA5.r_ ;
        func_ = functorRA5.func_ ;
        a1_ = functorRA5.a1_ ;
        a2_ = functorRA5.a2_ ;
        a3_ = functorRA5.a3_ ;
        a4_ = functorRA5.a4_ ;
        a5_ = functorRA5.a5_ ;
        return *this ;
    } ;

public:                                     // Control methodes.
    void                                    call (void)
    {
        r_ = func_ (a1_, a2_, a3_, a4_, a5_) ;
    } ;
    void*                                   getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*                            duplicate (void)
    {
        return new CTFunctorRA5<R, A1, A2, A3, A4, A5> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                                    // Members.
    R                                       r_ ;
    func_t                                  func_ ;
    A1                                      a1_ ;
    A2                                      a2_ ;
    A3                                      a3_ ;
    A4                                      a4_ ;
    A5                                      a5_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorRA5

template <class R, class A1, class A2, class A3, class A4, class A5, class A6> class CTFunctorRA6 : public CTFunctorXX
{
    //-----------------------------------------------------------------------------
public:                                         // Types.
    typedef R (*func_t) (A1, A2, A3, A4, A5, A6) ;

public:                                         // Constructor(s) and destructor.
    CTFunctorRA6                                (func_t func, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6)
: func_ (func), a1_ (a1), a2_ (a2), a3_ (a3), a4_ (a4), a5_ (a5), a6_ (a6)
    { }
    ;

public:                                         // Operators.
    CTFunctorRA6<R, A1, A2, A3, A4, A5, A6>&    operator= (const CTFunctorRA6<R, A1, A2, A3, A4, A5, A6>& functorRA6)
    {
        r_ = functorRA6.r_ ;
        func_ = functorRA6.func_ ;
        a1_ = functorRA6.a1_ ;
        a2_ = functorRA6.a2_ ;
        a3_ = functorRA6.a3_ ;
        a4_ = functorRA6.a4_ ;
        a5_ = functorRA6.a5_ ;
        a6_ = functorRA6.a6_ ;
        return *this ;
    } ;

public:                                         // Control methodes.
    void                                        call (void)
    {
        r_ = func_ (a1_, a2_, a3_, a4_, a5_, a6_) ;
    } ;
    void*                                       getResultPointer (void)
    {
        return &r_ ;
    } ;
    CTFunctorXX*                                duplicate (void)
    {
        return new CTFunctorRA6<R, A1, A2, A3, A4, A5, A6> (*this) ;
    } ;

    //-----------------------------------------------------------------------------
private:                                        // Members.
    R                                           r_ ;
    func_t                                      func_ ;
    A1                                          a1_ ;
    A2                                          a2_ ;
    A3                                          a3_ ;
    A4                                          a4_ ;
    A5                                          a5_ ;
    A6                                          a6_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctorRA6

class DLLSPEC_FUNCTOR CTFunctor : virtual public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:             // Constructor(s) and destructor.
    CTFunctor (const CTFunctorXX& functorXX) ;
    CTFunctor (const CTFunctor& functor) ;
    ~CTFunctor () ;
public:             // Operators.
    CTFunctor&      operator= (const CTFunctor& functor) ;

public:             // Control methodes.
    void            call (void) ;
    void*           getResultPointer (void) ;

    //-----------------------------------------------------------------------------
private:            // Members.
    CTFunctorXX*    pFunctorXX_ ;

    //-----------------------------------------------------------------------------
}

; // CTFunctor

#define makeFunctorA0(f)                                                        CTFunctor (CTFunctorA0 (f) )
#define makeFunctorA1(f, t1, a1)                                                CTFunctor (CTFunctorA1<t1> (f, a1) )
#define makeFunctorA2(f, t1, a1, t2, a2)                                        CTFunctor (CTFunctorA2<t1, t2> (f, a1, a2) )
#define makeFunctorA3(f, t1, a1, t2, a2, t3, a3)                                CTFunctor (CTFunctorA3<t1, t2, t3> (f, a1, a2, a3) )
#define makeFunctorA4(f, t1, a1, t2, a2, t3, a3, t4, a4)                        CTFunctor (CTFunctorA4<t1, t2, t3, t4> (f, a1, a2, a3, a4) )
#define makeFunctorA5(f, t1, a1, t2, a2, t3, a3, t4, a4, t5, a5)                CTFunctor (CTFunctorA5<t1, t2, t3, t4, t5> (f, a1, a2, a3, a4, a5) )
#define makeFunctorA6(f, t1, a1, t2, a2, t3, a3, t4, a4, t5, a5, t6, a6)        CTFunctor (CTFunctorA6<t1, t2, t3, t4, t5, t6> (f, a1, a2, a3, a4, a5, a6) )
#define makeFunctorA7(f, t1, a1, t2, a2, t3, a3, t4, a4, t5, a5, t6, a6, t7, a7)        CTFunctor (CTFunctorA7<t1, t2, t3, t4, t5, t6, t7> (f, a1, a2, a3, a4, a5, a6, a7) )
#define makeFunctorRA0(tr, f)                                                   CTFunctor (CTFunctorRA0<tr> (f) )
#define makeFunctorRA1(tr, f, t1, a1)                                           CTFunctor (CTFunctorRA1<tr, t1> (f, a1) )
#define makeFunctorRA2(tr, f, t1, a1, t2, a2)                                   CTFunctor (CTFunctorRA2<tr, t1, t2> (f, a1, a2) )
#define makeFunctorRA3(tr, f, t1, a1, t2, a2, t3, a3)                           CTFunctor (CTFunctorRA3<tr, t1, t2, t3> (f, a1, a2, a3) )
#define makeFunctorRA4(tr, f, t1, a1, t2, a2, t3, a3, t4, a4)                   CTFunctor (CTFunctorRA4<tr, t1, t2, t3, t4> (f, a1, a2, a3, a4) )
#define makeFunctorRA5(tr, f, t1, a1, t2, a2, t3, a3, t4, a4, t5, a5)           CTFunctor (CTFunctorRA5<tr, t1, t2, t3, t4, t5> (f, a1, a2, a3, a4, a5) )
#define makeFunctorRA6(tr, f, t1, a1, t2, a2, t3, a3, t4, a4, t5, a5, t6, a6)   CTFunctor (CTFunctorRA6<tr, t1, t2, t3, t4, t5, t6> (f, a1, a2, a3, a4, a5, a6) )

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////// End of Classes & Templates Classes Functor //
///////////////////////////////////////////////////////////////////////////////

#endif  // __FUNCTOR__
