/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include <CORBA/CORBAHelper.inl>
#include "CyrNSC.h"
#include <fstream>
#define FULL_INTERVAL (6 * 30)
#define STEP_INTERVAL (6 * 3)
std::fstream LogFile;
void* NameServiceCleaner_Run(void* args)
{
    NameServiceCleaner* NSC=reinterpret_cast<NameServiceCleaner*>(args);
    NSC->Run();
    return NULL;
}
//number of retrys to test for non existence,
//CORBA::COMM_FAILURE seems to be both valid and invalid removal!!!
const int retrys=3;
NameServiceCleaner::NameServiceCleaner()
{
    LogFile.open(".\\NSC.log",std::ios::out);
    int i=0;
    m_orb = CORBA::ORB_init( i, NULL, "omniORB2" );
    m_boa = m_orb->BOA_init( i, NULL, "omniORB2_BOA" );
    m_boa->impl_is_ready( 0, true);
    try
    {
        // Create Cleaner thread and start it running.
        cleaner = new omni_thread(&NameServiceCleaner_Run, this);
        cleaner->start();
        m_running=true;
    }
    catch( ... )
    {}

    LogFile<<"NameService Cleaner Started"<<std::endl;
}
NameServiceCleaner::~NameServiceCleaner()
{
    // Request termination of thread.
    m_running = false;
    cleaner->join(NULL);
    CORBA::release( m_orb );
    CORBA::release( m_boa );
    LogFile.close();
}
void NameServiceCleaner::Run()
{
    // Loop until the daemon is being shut down
    while( m_running )
    {
        try
        {
            LogFile<<"Clean NameService,";
            Clean();//clean as soon as started
            // Sleep for alloted time
            for ( int i = 0; i < FULL_INTERVAL; i += STEP_INTERVAL)
            {
                omni_thread::sleep( STEP_INTERVAL );
                //if told to stop break out of this loop and fall through outer loop
                if(!m_running)break;
            }
        }
        catch( ... )
        {
            //just end!!;
        }

    }
}
void NameServiceCleaner::Clean()
{
    CosNaming::NamingContext_var context;

    try
    {
        // Obtain a reference to the root context of the Name service:
        CORBA::Object_var initServ;
        initServ = m_orb->resolve_initial_references("NameService");

        // Narrow the object returned by resolve_initial_references()
        // to a CosNaming::NamingContext object:
        context = CosNaming::NamingContext::_narrow(initServ);
        if (CORBA::is_nil(context))
        {
            LogFile << "Failed to narrow naming context." << std::endl;
            return;
        }
    }
    catch(CORBA::ORB::InvalidName )
    {
        LogFile << "Service required is invalid [does not exist]." << std::endl;
    }

    // Get a list of all contexts on NameService
    CosNaming::BindingIterator_var bi;
    CosNaming::BindingList_var bl;
    CosNaming::Binding_var b;

    context->list( 0, bl, bi );
    // Loop through all contexts and get all objects
    while ( bi->next_one(b) )
    {
        CORBA::Object_var obj;
        try
        {
            // Resolve name object
            obj = context->resolve( b->binding_name );
        }
        catch( CORBA::Exception &e)
        {
            // If it is a CORBA exception then the error is that the requested
            // object cannot be found. So de-register
            LogFile<< std::endl<<"Cannot resolve "<<(const char*)b->binding_name[0].id<<","<<(const char*)b->binding_name[0].kind<<" Caught:"<<e<<std::endl;
            context->unbind(  b->binding_name );
        }
        catch( ... )
        {
            // Catch all other exceptions
        }
        //Retry on CORBA::Exceptions break out on anything else

        bool deregister=false;
        for(int tryno=1;tryno<=retrys;tryno++)
        {
            deregister=false;//set to false on each try
            try// Check for existence
            {
                if(obj->_non_existent())//object does not exist
                {
                    LogFile<<std::endl<< "Object No Longer Exists "<<(const char*)b->binding_name[0].id<<","<<(const char*)b->binding_name[0].kind<<std::endl;
                    deregister=true;
                }
                break;
            }
            catch( CORBA::Exception &e)
            {
                // If it is a CORBA exception then the error is that the requested
                // object cannot be found. So de-register
                LogFile << std::endl<<"Thrown on try "<<tryno<<" "<<(const char*)b->binding_name[0].id<<","<<(const char*)b->binding_name[0].kind<<" Caught:"<<e<<std::endl;
                deregister=true;
            }
            catch( ... )
            {
                // Catch all other exceptions
            }

            Sleep(2000);//wait a couple of seconds before trying again
        }
        //If after retrys the object still not found then remove

        if(deregister)context->unbind(  b->binding_name );
        LogFile.flush();


    }
    // Tidy up
    bi->destroy();
}
