/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrWrapLib
// @module  Borrowed Items | This module contains the implementation of the
//          CECOrbBorrow class.  This class allows the continual borrowing of
//          objects from the Factory and prevents them being reaped.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbBorrow.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor is responsible for creating a CORBA helper class
//          and for connecting to a remote Factory.  This Factory must be
//          one that can create OpenSTARep interface objects.
// @parm    Pointer to a CORBA helper class.
// @parm    The name of the Factory as registered with the Name Service.
CECOrbBorrow::CECOrbBorrow( CORBAServer* corba, const char* factory )
{
    // Error information
    const char* where = "CECOrbBorrow::CECOrbBorrow";

    try
    {
        // Create CORBA functionality class
        m_corba = corba;

        // Connect to Factory
        CORBA::Object_ptr obj = m_corba->GetObjectFromClass( Factory::CLASS_NAME, factory );
        m_factory = narrow_object( (Factory*)NULL, obj );
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor should be used when Factory and CORBA helper
//          objects are already available.
// @parm    Pointer to a CORBA helper class.
// @parm    Reference to a Factory.
CECOrbBorrow::CECOrbBorrow( CORBAServer* corba, Factory_ptr factory )
{
    // Assign member variables
    m_corba = corba;
    m_factory = factory;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   As well as tidying up any allocated memory, the destructor is
//          responsible for disposing of the CORBA object in the Factory.
CECOrbBorrow::~CECOrbBorrow()
{
    // Dispose of CORBA object

    //
    // We first check for a valid class name to account for the CECOrbTestRun class.
    //
    // CECOrbTestRuns do not necessarily create a new testrun every time as they
    // are able to re-use an existing instance within the Daemon
    if (!m_class.empty())
    {
        m_factory->Dispose( m_class.c_str(), m_name.c_str() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method determines what name and class will be used when the
//          Borrow method is called.
// @devnote This method should really have been merged into the constructors.
//          However, the name parameter is not available until after the
//          underlying servant object is created and this must be done in the
//          derived constructor.  Therefore a separate method was needed.
// @parm    The registered name of the underlying servant object.
// @parm    The name of the class of the underlying object.
void CECOrbBorrow::Register( const char* name, const char* cls )
{
    // Assign member variables
    m_name = name;
    m_class = cls;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is called by a borrowing thread when the object is due
//          for renewal.
// @parm    The length of time the object's lease should be extended by - in
//          reaper periods.
void CECOrbBorrow::Borrow( long period )
{
    m_factory->Borrow( m_class.c_str(), m_name.c_str(), period );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the Factory object that maintains
//          the underlying servant object.  This reference should be used
//          to maintain the underlying servant object and can also be used
//          to create other OpenSTARep interfaces.  This call will increment the
//          reference count and therefore the returned value should be
//          released.
// @rdesc   The reference to the servant object.
Factory_ptr CECOrbBorrow::GetFactory()
{
    return Factory::_duplicate( m_factory );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a pointer to the CORBA helper class,
//          CORBAServer, that exists as a member within this class.  This will
//          prevent having to create more than one of these.
// @rdesc   A pointer to the CORBA helper class.
CORBAServer* CECOrbBorrow::GetHelper()
{
    return m_corba;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the registered name of the object within the
//          Factory.  This name must be used to extend the lease on a
//          particular object.
// @devnote Note that this name is supposed to be unique and can be used to
//          register this object in maps etc.
// @rdesc   The registered name of the object.
string CECOrbBorrow::GetRegisteredName() const
{
    return m_name;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method rdeturns the registered class name of the object within
//          the Factory.  This class name must be used to extend the lease on
//          a particular object.
// @rdesc   The registered class name of the object.
string CECOrbBorrow::GetRegisteredClass() const
{
    return m_class;
}
