/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrWrapLib
// @module  Server Wrapper | This module contains the implementation of
//          the CECOrbServer wrapper class.  This class simply wraps around the
//          CORBA interface CyrStServer and hides the CORBA code.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbServer.h>
#include <Factory/ParamList_i.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor initializes the underlying CORBA interface within
//          a remote Factory.
// @parm    Pointer to a CORBA helper class.
// @parm    The name of the Factory that contains the OpenSTARep Provider.
// @parm    The log on name of the user.
// @parm    The password of the user.
CECOrbServer::CECOrbServer( CORBAServer* corba, const char* factory,
                            const char* user, const char* password )
        : CECOrbBorrow( corba, factory )
{
    // Error information
    const char* where = "CECOrbServer::CECOrbServer";

    try
    {
        // Create a new parameter list
        FactoryParamList* params =
            new FactoryParamList( CyrStServer::CLASS_NAME, "", DEFAULT_PERIOD );
        m_corba->ReadyObject( params );
        params->AddValue( CyrStServer::PARAM_USER, user );
        params->AddValue( CyrStServer::PARAM_PASSWORD, password );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params );
        m_servant = narrow_object( (CyrStServer*)NULL, obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStServer::CLASS_NAME );

        // Destroy parameter object
        params->_dispose();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the list of Projects known about by the
//          Server.  The list is returned as a string that contains a list
//          of names separated by a '\|' character and terminated by two '\|'
//          characters.
// @rdesc   The separated list of Project names.
// @parm    Will contain the number of Projects found.
string CECOrbServer::GetProjects( long& count ) const
{
    // Error information
    const char* where = "CECOrbServer::GetProjects";
    const long status = CYR_CORBA_EXCEPTION;

    try
    {
        // Call underlying CORBA object's method, but store result in
        // a CORBA string so that memory is handled correctly.
        CORBA::String_var s = m_servant->GetProjects( count );

        // Convert the CORBA string into something else so that the CORBA
        // mechanism remains hidden.
        string ret( s );
        return ret;
    }
    catch( CORBA::Exception ex )
    {
        throw CyrRepositoryException( CorbaExceptionString( ex ),
                                      where, status );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the underlying CORBA servant object
//          to the calling code.  This will increment the reference count and
//          so will need to be released.
// @rdesc   The reference to the servant object.
CyrStServer_ptr CECOrbServer::GetServant()
{
    return CyrStServer::_duplicate( m_servant );
}

