/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrWrapLib
// @module  Test Run Wrapper | This module contains the implementation of
//          the CECOrbTestRun wrapper class.  This class simply wraps around
//          the CORBA interface CyrStTestRun and hides the CORBA code.  Also in
//          this module are several simple Test Output types
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbTestRun.h>
#include <Factory/ParamList_i.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor is used to retrieve a Test Run from the OpenSTARep
//          database.
// @parm    Pointer to the parent Test Store.
// @parm    The name of the Test Run to retrieve.
CECOrbTestRun::CECOrbTestRun( CECOrbFolder* parent, const char* name )
        : CECOrbFolder( parent )
{
    // Error information
    const char* where = "CECOrbTestRun::CECOrbTestRun";

    try
    {
        // Create a new parameter list for Test Run.  Remember that all
        // returned strings must be released so we copy them into
        // CORBA::String_var variables before passing them as parameters.  This
        // will cause them to be released as soon as the variable goes out of
        // scope.
        FactoryParamList* params =
            new FactoryParamList( CyrStTestRun::CLASS_NAME, name, DEFAULT_PERIOD );
        m_corba->ReadyObject( params );
        string paramStore = parent->GetRegisteredName();
        params->AddValue( CyrStTestRun::PARAM_STORE, paramStore.c_str() );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params );
        m_servant = narrow_object( (CyrStTestRun*)NULL, obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStTestRun::CLASS_NAME );

        // Destroy parameter list
        params->_dispose();

        // This instance of the Test Run is readonly and cannot be updated
        m_readonly = true;
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor is used when a brand new Test Run needs to be
//          created.
// @devnote Currently the Test version is hard-coded to be 1.  This is because
//          the COM interface does not appear to offer a method to get the
//          Folder version.  It will need changing when the interface does
//          allow this functionality, and also when Cyrano put versioned Tests
//          into Impact.
// @parm    Pointer to the parent Test Store.
CECOrbTestRun::CECOrbTestRun( CECOrbFolder* pParent, const char* desc,
                              const char* node, const char* executor )
        : CECOrbFolder( pParent )
{
    // Error information
    const char* where = "CECOrbTestRun::CECOrbTestRun";

    try
    {
        // Create a new parameter list for Test Run.  Remember that all
        // returned strings must be released so we copy them into
        // CORBA::String_var variables before passing them as parameters.  This
        // will cause them to be released as soon as the variable goes out of
        // scope.
        FactoryParamList* params =
            new FactoryParamList( CyrStTestRun::CLASS_NAME, "", DEFAULT_PERIOD );
        m_corba->ReadyObject( params );
        string paramProject = pParent->GetRegisteredName();
        params->AddValue( CyrStTestRun::PARAM_STORE, paramProject.c_str() );
        params->AddValue( CyrStTestRun::PARAM_DESC, desc );
        params->AddValue( CyrStTestRun::PARAM_NODENAME, node );
        params->AddValue( CyrStTestRun::PARAM_EXECNAME, executor );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params );
        m_servant = narrow_object( (CyrStTestRun*)NULL, obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStTestRun::CLASS_NAME );

        // Destroy parameter list
        params->_dispose();

        // This instance of the Test Run is not read-only and can be updated
        m_readonly = false;
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor allows a Test Run object to connect to an existing
//          and active Test Run.
// @parm    Pointer to the parent Test Store.
// @parm    Reference to active Test Run.
CECOrbTestRun::CECOrbTestRun( CECOrbFolder* pParent, CyrStTestRun_ptr pTestRun )
        : CECOrbFolder( pParent )
{
    // Error information
    const char* where = "CECOrbTestRun::CECOrbTestRun";

    try
    {
        // Assign member variables
        m_servant = CyrStTestRun::_duplicate( pTestRun );
        m_name = m_servant->GetRegisteredName();

        // This instance of the Test Run is not read-only and can be updated
        m_readonly = false;
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor is responsible for generating a stop time for the
//          Test Run if one is required.
CECOrbTestRun::~CECOrbTestRun()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the underlying CORBA servant object
//          to the calling code.  This will increment the reference count and
//          so will need to be released.
// @rdesc   The reference to the servant object.

CyrStTestRun_ptr CECOrbTestRun::GetServant()
{
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    return CyrStTestRun::_duplicate( pTestRun );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description of the Test Run object.
// @rdesc   The description string.
string CECOrbTestRun::GetDescription()
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var s = pTestRun->GetDescription();

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    string ret( s );
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the Test being run.
// @rdesc   The Test name.
string CECOrbTestRun::GetTestName()
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var s = pTestRun->GetTestName();

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    string ret( s );
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the current version of the Test being run.
// @rdesc   The version number.
long CECOrbTestRun::GetTestVersion()
{
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    return pTestRun->GetTestVersion();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the node that the current Test Run
//          is being generated on.
// @rdesc   The name of the node.
string CECOrbTestRun::GetNodeName()
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var s = pTestRun->GetNodeName();

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    string ret( s );
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the executor that is currently
//          running the Test.
// @rdesc   The name of the Executor.
string CECOrbTestRun::GetExecutorName()
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var s = pTestRun->GetExecutorName();

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    string ret( s );
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the time at which the Test Run was started.
// @rdesc   The date/time value.
CTTime CECOrbTestRun::GetTimeStarted()
{
    // Create date/time object
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CTTime tm( pTestRun->GetTimeStarted() );
    return tm;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the time at which the Test Run was stopped.
// @rdesc   The date/time value.
CTTime CECOrbTestRun::GetTimeStopped()
{
    // Create date/time object
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CTTime tm( pTestRun->GetTimeStopped() );
    return tm;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a single test output object to the OpenSTARep
//          database.
// @parm    Pointer to an instance of a test output object.
void CECOrbTestRun::AddTestOutput( CECTestOutput* output )
{
    // Determine which array we need to add output to
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    pTestRun->AddTestOutput( output->GetType().c_str(), output->ToString().c_str() );
}

string CECOrbTestRun::GetFirstOutputString( const char* outType )
{
    // Determine which array we need to add output to
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var str = pTestRun->GetFirstOutput( outType );
    return string( str );
}

string CECOrbTestRun::GetNextOutputString( const char* outType )
{
    // Determine which array we need to add output to
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var str = pTestRun->GetNextOutput( outType );
    return string( str );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method resets the iteration for timer output values.  It then
//          creates and returns the first timer in the sequence.  Note that
//          the returned pointer must be freed by the calling application else
//          there will be memory leaks.
// @rdesc   Pointer to the first timer output in the sequence.
CECTimerOutput* CECOrbTestRun::GetFirstTimer()
{
    CECTimerOutput* timer = NULL;

    // If we successfully retrieve an entry then continue
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var str = pTestRun->GetFirstOutput( CECTimerOutput::OUTPUT_TYPE );
    if ( strlen( str ) > 0 )
        timer = new CECTimerOutput( str );
    return timer;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the next timer output value in the sequence.
//          It calls the underlying CORBA Test Run object to retrieve a single
//          string from the remote OpenSTARep database.  A timer output object is
//          then created from that string and returned to the calling code.
//          Note that the returned pointer must be deleted by the calling
//          application to prevent memory leaks.
// @rdesc   The next timer value in the sequence.
CECTimerOutput* CECOrbTestRun::GetNextTimer()
{
    CECTimerOutput* timer = NULL;

    // If we successfully retrieve an entry then continue
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var str = pTestRun->GetNextOutput( CECTimerOutput::OUTPUT_TYPE );
    if ( strlen( str ) > 0 )
        timer = new CECTimerOutput( str );
    return timer;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method resets the iteration for test log output values.  It
//          then creates and returns the first test log in the sequence.  Note
//          that the returned pointer must be freed by the calling application
//          else there will be memory leaks.
// @rdesc   Pointer to the first test log output in the sequence.
CECTestLog* CECOrbTestRun::GetFirstTestLog()
{
    CECTestLog* log = NULL;

    // If we have successfully retrieved an entry then continue
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var str = pTestRun->GetFirstOutput( CECTestLog::OUTPUT_TYPE );
    if ( strlen( str ) > 0 )
        log = new CECTestLog( str, false );

    return log;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the next test log value in the sequence.
//          It calls the underlying CORBA Test Run object to retrieve a single
//          string from the remote OpenSTARep database.  A test log object is
//          then created from that string and returned to the calling code.
//          Note that the returned pointer must be deleted by the calling
//          application to prevent memory leaks.
// @rdesc   The next test log value in the sequence.
CECTestLog* CECOrbTestRun::GetNextTestLog()
{
    CECTestLog* log = NULL;

    // If we have successfully retrieved an entry then continue
    CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                (CORBA::Object_ptr)m_servant );
    CORBA::String_var str = pTestRun->GetNextOutput( CECTestLog::OUTPUT_TYPE );
    if ( strlen( str ) > 0 )
        log = new CECTestLog( str, false );

    return log;
}

void CECOrbTestRun::StopTestRun()
{
    try
    {
        // If we have created a brand new Test Run object we need to generate
        // a stop time in the same way that the start time was generated.
        if (( m_readonly == false ) && ( !m_class.empty() ))
        {
            // Update stop time in underlying entity
            CyrStTestRun_var pTestRun = narrow_object( (CyrStTestRun*)NULL,
                                        (CORBA::Object_ptr)m_servant );
            pTestRun->StopTestRun();
        }
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), ex.where().c_str(), ex.status() );
    }
}
