/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Factory
// @module  Factory | This module contains the implementation of the Factory
//          class.  The FactoryImpl class contains the actual implementation
//          code for a Factory object.  Factory objects provide a common
//          location for the creation and management of CORBA objects.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Factory/Factory_i.h>
#include <Logging/CECLog.h>

using namespace CORBA;

extern CECLog* flogger;

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor is responsible for allocating memory and
//          initialising the CORBA services for this class.  It also registers
//          the Factory with the NameService under the given name.
// @parm    Pointer to CORBA helper functionality
// @parm    The name of the Factory object.
FactoryImpl::FactoryImpl( CORBAServer* pServer, const char* name )
{
    LogEvent( 1, "Factory starting" );

    // Assign member variables
    m_pServer = pServer;

    // Register factory
    m_name = name;
    m_pServer->RegisterObject( this, Factory::CLASS_NAME, name );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor does nothing as we cannot guarantee when it will be
//          called.
FactoryImpl::~FactoryImpl()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is responsible for destroying any allocated memory
//          and shutting down the CORBA services.

void FactoryImpl::Shutdown()
{
    // Deregister this object
    m_pServer->DeregisterObject( Factory::CLASS_NAME, m_name.c_str() );

    LogEvent( 1, "Factory shutting down" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method registers a particular class and Provider with the
//          Factory object.
// @parm    The name of the class to register.
// @parm    The name of the Provider that contains the class.
// @parm    The type of the class being registered.
//
// @devnote Only unique classes are registered.  Therefore the order in which
//          a collection of Providers register themselves may be important.
//
// @xref    <mf FactoryImpl::Deregister>, <mf FactoryImpl::DeregisterAll>
void FactoryImpl::Register( const char* cls, const char* provider, long type )
{
    // Register class with list
    m_Reg.Register( cls, provider, type );

    // Log event
    string event( "Registered class [" );
    event += cls;
    event += "] with Provider [";
    event += provider;
    event += "]";
    LogEvent( 2, event.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method deregisters a particular class from the Factory object.
// @parm    The name of the class to remove.
// @parm    The name of the Provider that contains the class.
//
// @devnote Note that there is NO security on this call, so anyone can
//          deregister a class.  This has been done deliberately so that
//          future upgrades for objects can be carried out easily.  The latest
//          Provider of the upgraded object can simply deregister the old one
//          and register itself as the new Provider of those objects.
//
// @xref    <mf FactoryImpl::DeregisterAll>, <mf FactoryImpl::Register>
void FactoryImpl::Deregister( const char* cls, const char* provider )
{
    // Deregister class from list
    m_Reg.Deregister( cls, provider );

    // Log event
    string event( "Deregistered class [" );
    event += cls;
    event += "] from Provider [";
    event += provider;
    event += "]";
    LogEvent( 2, event.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes all classes associated with a particular
//          Provider from the Factory.
// @parm    The name of the Provider that contains the classes.
//
// @xref    <mf FactoryImpl::Register>, <mf FactoryImpl::Deregister>
void FactoryImpl::DeregisterAll( const char* provider )
{
    // Remove all classes from specified provider
    m_Reg.DeregisterAll( provider );

    // Log event
    string event( "Deregistered all classes from Provider: " );
    event += provider;
    LogEvent( 2, event.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves an actual Provider reference based on the
//          passed class name.
// @rdesc   A reference to a Provider object.
// @parm    The name of the Provider to locate.
Provider_ptr FactoryImpl::GetProvider( string name ) const
{
    const char* where = "Factory::GetProvider";

    try
    {
        // Get reference to Provider
        CORBA::Object_ptr obj =
            m_pServer->GetObjectFromClass( Provider::CLASS_NAME, name.c_str() );

        Provider_var p = narrow_object( (Provider*)NULL, obj );

        // Return Provider reference
        return p;
    }
    catch( CORBA::Exception )
    {
        // A CORBA exception caught here will be because the Provider
        // could not be found.
        throw CyrCORBAException( "Provider not found", where,
                                 CYR_CORBA_OBJECT_NOT_FOUND );
    }

    // Return a nil reference.  This will only occur if there has been some
    // kind of error that has not been caught by the exception handlers.
    return Provider::_nil();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method creates a new instance of an object based on the class
//          name passed into the function.  The new object's name is passed as
//          well. In actuality the request is simply passed on to the relevant
//          Provider.
// @rdesc   A reference to a new Product.
// @parm    The list of parameters.
//
// @xref    <mf FactoryImpl::Borrow>, <mf FactoryImpl::Reaper>
CORBA::Object_ptr FactoryImpl::New( ParamList_ptr params )
{
    // Get class type parameter out of parameter list
    CORBA::String_var cls = params->GetValue( Factory::PARAM_CLASS );

    // Locate the correct provider
    string s = m_Reg.GetFirstProvider( cls );

    try
    {
        // Get reference to Provider
        Provider_var p = GetProvider( s );

        // Pass request on to Provider
        return p->New( params );
    }
    catch( ... )
    {
        // Any exceptions should be ignored
        string error( "Exception: Cannot create class [" );
        error += cls ;
        error += "]";
        LogEvent( 3, error.c_str() );
    }

    return CORBA::Object::_nil();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the first class type that the Factory knows
//          how to construct.  Further classes can be retrieved by using the
//          <mf FactoryImpl::GetNext> method.
// @rdesc   The first element in the list, or an empty string if none can be
//          found.
//
// @ex      This example demonstrates a typical use of this method: |
//          // List all factory classes
//          CORBA::String_var cls = factory->GetFirstClass();
//          while( strlen( cls ) > 0 )
//          {
//              cout << cls << endl;
//              cls = factory->GetNext();
//          }
//
//
// @xref    <mf FactoryImpl::GetFirstClassFromType>, <mf FactoryImpl::GetNext>,
//          <mf FactoryImpl::GetFirstClassFromProvider>,
//          <mf FactoryImpl::GetFirstProviderFromClass>
char* FactoryImpl::GetFirstClass()
{
    // Get first class from list
    string s = m_Reg.GetFirstClass();

    // Return string
    return CORBA::string_dup( (const char*) s.data() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the first class type that the Factory knows
//          how to construct.  The list is filtered to only return classes of
//          the specified type.  Further classes can be retrieved by using the
//          <mf FactoryImpl::GetNext> method.
// @rdesc   The first element in the list, or an empty string if none can be
//          found.
// @parm    The types of classes to search for.
//
// @xref    <mf FactoryImpl::GetFirstClass>, <mf FactoryImpl::GetNext>,
//          <mf FactoryImpl::GetFirstClassFromProvider>,
//          <mf FactoryImpl::GetFirstProviderFromClass>
char* FactoryImpl::GetFirstClassFromType( long type )
{
    // Get first class from list
    string s = m_Reg.GetFirstClass( type );

    // Return string
    return CORBA::string_dup( s.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the first class type that the Factory knows how
//          to construct.  The list is filtered to only return classes that
//          belong to a specific Provider.
// @rdesc   The first element in the list, or an empty string if none can be
//          found.
// @parm    The name of the Provider containing the classes.
//
// @xref    <mf FactoryImpl::GetFirstClass>, <mf FactoryImpl::GetNext>,
//          <mf FactoryImpl::GetFirstClassFromType>,
//          <mf FactoryImpl::GetFirstProviderFromClass>
char* FactoryImpl::GetFirstClassFromProvider( const char* provider )
{
    // Get first class from list
    string s = m_Reg.GetFirstClass( provider );

    // Return string
    return CORBA::string_dup( s.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the first Provider that can be found that is
//          able to construct classes of a specified type.  Further Providers
//          can be retrieved using the <mf FactoryImpl::GetNext> method.
// @rdesc   The first element in the list, or an empty string if none can be
//          found.
// @parm    The name of the class to locate.
//
// @xref    <mf FactoryImpl::GetFirstClassFromType>, <mf FactoryImpl::GetNext>,
//          <mf FactoryImpl::GetFirstClassFromProvider>,
//          <mf FactoryImpl::GetFirstClass>, <mf FactoryImpl::GetFirstProvider>
char* FactoryImpl::GetFirstProviderFromClass( const char* cls )
{
    // Get first Provider from list
    string s = m_Reg.GetFirstProvider( cls );

    // Return string
    return CORBA::string_dup(s.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the first unique Provider that can be found.
//          Further Providers can be retrieved using the
//          <mf FactoryImpl::GetNext> method.
// @rdesc   The first element in the list, or an empty string if none can be
//          found.
// @xref    <mf FactoryImpl::GetFirstClassFromType>, <mf FactoryImpl::GetNext>,
//          <mf FactoryImpl::GetFirstClassFromProvider>,
//          <mf FactoryImpl::GetFirstClass>,
//          <mf FactoryImpl::GetFirstProviderFromClass>
char* FactoryImpl::GetFirstProvider()
{
    // Get first Provider from list
    string s = m_Reg.GetFirstProvider();

    // Return string
    return CORBA::string_dup(s.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Once any of the <lt>i<gt>GetFirstX<lt>/i<gt> methods have been
//          called, this method will allow the retrieval of further entries in
//          the list.
// @rdesc   The next element in the list, or any empty string if none can be
//          found.
char* FactoryImpl::GetNext()
{
    // Get next string from list
    string s = m_Reg.GetNext();

    // Return string
    return CORBA::string_dup( (const char*) s.data() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   When this method is called the Factory tells all its Providers to
//          dispose of any dead objects.  Note that calling this method will
//          invalidate any other searches being done.
// @xref    <mf FactoryImpl::Borrow>, <mf FactoryImpl::Dispose>
void FactoryImpl::Reaper()
{
    string s;

    try
    {
        string event( "About to reap objects..." );
        LogEvent( 100, event.c_str() );

        // Get list of unique elements from Provider list.
        s = m_Reg.GetFirstProvider();
        while ( !s.empty() )
        {
            string eventReap( "Reaping All Corba Objects for Provider = [ " );
            eventReap += s + "]";
            LogEvent( 100, eventReap.c_str() );

            // Get reference to Provider and call the reaper method.
            Provider_var p = GetProvider( s );

            p->Reaper();

            // Get next Provider name in list
            s = m_Reg.GetNext();
        }
    }
    catch( CyrCORBAException )
    {
        // A CORBA exception here implies that a Provider could not
        // be found to extend the lease on the requested object.

        string event( "Corba Exception Thrown on reaping Corba Object, Provider = [ " );
        event += s;
        event += "]";
        LogEvent( 100, event.c_str() );
    }
    catch(...)
    {
        string event( "General Exception Thrown on reaping Corba Object, Provider = [ " );
        event += s;
        event += "]";
        LogEvent( 100, event.c_str() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method extends the length of time an object has to live.  The
//          timing mechanism is in 'Reaper calls', i.e. specifying 3 will mean
//          the object exists until the Reaper has been called 3 times.  By
//          placing this call in a timed loop it is possible to implement a
//          more consistent Reaping mechanism. Note that calling this method
//          will invalidate any other searches currently being done.
// @parm    The name of the class to extend the lease on.
// @parm    The name of the object to extend the lease on.
// @parm    The length of time to extend the lease by in minutes.
//
// @xref    <mf FactoryImpl::Reaper>, <mf FactoryImpl::New>,
//          <mf FactoryImpl::Dispose>
void FactoryImpl::Borrow( const char* cls, const char* name, long period )
{
    string s;

    try
    {
        // Locate the correct provider
        s = m_Reg.GetFirstProvider( cls );

        // Get reference to Provider
        Provider_var p = GetProvider( s );

        // Extend period
        p->Borrow( cls, name, period );
    }
    catch( CyrCORBAException )
    {
        // A CORBA exception here implies that a Provider could not
        // be found to extend the lease on the requested object.

        string event( "Corba Exception Thrown on borrowing Corba Object, Provider = [ " );
        event += s;
        event += "] Class Name = [";
        event += cls;
        event += "] Name = [";
        event += name;
        event += "]";
        LogEvent( 100, event.c_str() );
    }
    catch(...)
    {
        string event( "General Exception Thrown on borrowing Corba Object, Provider = [ " );
        event += s;
        event += "] Class Name = [";
        event += cls;
        event += "] Name = [";
        event += name;
        event += "]";
        LogEvent( 100, event.c_str() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc    This method calls the _dispose() method on a registered object.
//           The _dispose() method does not delete the object immediately.  It
//           waits for all clients to drop their references and then deletes
//           the object.  This method should only be used if you can guarantee
//           that your code has created the object and all other code will
//           not need access to this object at a later point.
// @parm     The type of object to dispose.
// @parm     The name of the object to dispose.
//
// @xref     <mf FactoryImpl::Borrow>, <mf FactoryImpl::New>
void FactoryImpl::Dispose( const char* cls, const char* name )
{
    string s;

    try
    {
        // Locate the correct provider
        s = m_Reg.GetFirstProvider( cls );

        string event( "Collected the First Provider [ " );
        event += s + "] Class Name = [" + cls + "] Name = [" + name + "]";
        LogEvent( 100, event.c_str() );

        if (!s.empty())
        {
            // Get reference to Provider
            Provider_var p = GetProvider( s );

            // Dispose of object
            p->Dispose( cls, name );

            string disposeEvent( "Disposed " );
            LogEvent( 100, disposeEvent.c_str() );
        }
    }
    catch( CyrCORBAException )
    {
        // A CORBA exception here implies that a Provider could not
        // be found to dispose of the requested object.

        string event( "Corba Exception Thrown on disposing Corba Object, Provider = [ " );
        event += s;
        event += "]";
        LogEvent( 100, event.c_str() );
    }
    catch(...)
    {
        string event( "General Exception Thrown on disposing Corba Object, Provider = [ " );
        event += s;
        event += "]";
        LogEvent( 100, event.c_str() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method provides simplistic logging functionality.  Any Factory
//          event worthy of logging can be logged using this method.
// @devnote This method was added after most of the Factory had been written
//          and so it will be worth going through the Factory code and adding
//          in events worth logging.
void FactoryImpl::LogEvent( long level, const char* event )
{
    flogger->AddTrace( CYR_FACTORY, level, event );
}
