/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// CYRService.cpp
// --------------
//
//    This is the include file for the CYRService class
//
//=========================================================================

#include <windows.h>
#include <stdio.h>
#include "CYRService.h"

// static variables
CYRService* CYRService::m_pThis = NULL;

CYRService::CYRService(const char* szServiceName)
{
    // copy the address of the current object so we can access it from
    // the static member callback functions.
    // WARNING: This limits the application to only one CYRService object.
    m_pThis = this;

    // Set the default service name and version
    strncpy(m_szServiceName, szServiceName, sizeof(m_szServiceName)-1);
    m_iMajorVersion = 1;
    m_iMinorVersion = 0;
    m_hEventSource = NULL;

    // set up the initial service status
    m_hServiceStatus = NULL;
    m_Status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    m_Status.dwCurrentState = SERVICE_STOPPED;
    m_Status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
    m_Status.dwWin32ExitCode = 0;
    m_Status.dwServiceSpecificExitCode = 0;
    m_Status.dwCheckPoint = 0;
    m_Status.dwWaitHint = 0;
    m_bIsRunning = FALSE;
}

CYRService::~CYRService()
{
    if (m_hEventSource)
    {
        ::DeregisterEventSource(m_hEventSource);
    }
}

//
// ParseCommandLine
// ================
//
// See what mode we are to run in
BOOL CYRService::ParseCommandLine(int argc, char* argv[])
{
    // See if we have any command line args we recognise
    if (argc <= 1) return FALSE;

    // Version Info
    if (_stricmp(argv[1], "-v") == 0)
    {
        // Spit out version info
        printf("%s Version %d.%d\n",
               m_szServiceName, m_iMajorVersion, m_iMinorVersion);
        printf("The service is %s installed\n",
               IsInstalled() ? "currently" : "not");
        return TRUE; // say we processed the argument

    }

    // Install the service

    else if (_stricmp(argv[1], "-i") == 0)
    {
        // Request to install.- or modify the service
        if (IsInstalled())
        {
            printf("%s is already installed\n", m_szServiceName);
        }
        else
        {
            // Try and install the copy that's running
            if (Install())
            {
                printf("%s installed\n", m_szServiceName);
            }
            else
            {
                printf("%s failed to install. Error %d\n", m_szServiceName, GetLastError());
            }
        }
        return TRUE; // say we processed the argument
    }

    // Uninstall the service

    else if (_stricmp(argv[1], "-u") == 0)
    {
        // Request to uninstall.
        if (!IsInstalled())
        {
            printf("%s is not installed\n", m_szServiceName);
        }
        else
        {
            // Try and remove the copy that's installed
            if (Uninstall())
            {
                // Get the executable file path
                char szFilePath[_MAX_PATH];

                ::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));

                printf("%s removed. (You must delete the file (%s) yourself.)\n",
                       m_szServiceName, szFilePath);
            }
            else
            {
                printf("Could not remove %s. Error %d\n", m_szServiceName, GetLastError());
            }
        }
        return TRUE; // say we processed the argument

        // Make automatic, manual or disabled
    }

    else if ((_stricmp(argv[1], "-a") == 0) ||
             (_stricmp(argv[1], "-m") == 0) ||
             (_stricmp(argv[1], "-d") == 0))
    {
        if (IsInstalled())
        {
            ChangeConfig(argv[1]);
        }
        else
        {
            printf("Service %s is not installed. Error %d\n", m_szServiceName, GetLastError());
        }
    }

    // Unrecognized Argument
    return FALSE;
}

//
// IsInstalled
// ===========
//
// Test if the service is currently installed
BOOL CYRService::IsInstalled()
{
    BOOL bResult = FALSE;

    // Open the Service Control Manager
    SC_HANDLE hServiceManager = ::OpenSCManager(NULL,                   // local machine
                                NULL,                   // ServicesActive database
                                SC_MANAGER_ALL_ACCESS); // full access
    if (hServiceManager)
    {
        // Try to open the service
        SC_HANDLE hService = ::OpenService(hServiceManager,
                                           m_szServiceName,
                                           SERVICE_QUERY_CONFIG);
        if (hService)
        {
            bResult = TRUE;
            ::CloseServiceHandle(hService);
        }

        ::CloseServiceHandle(hServiceManager);
    }

    return bResult;
}

//
// Install
// =======
//
// Install the service
BOOL CYRService::Install()
{
    // Open the Service Control Manager
    SC_HANDLE hServiceManager = ::OpenSCManager(NULL,                   // local machine
                                NULL,                   // ServicesActive database
                                SC_MANAGER_ALL_ACCESS); // full access
    if (!hServiceManager)
        return FALSE;

    // Get the executable file path
    char szFilePath[_MAX_PATH];
    ::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));

    // Create the service
    SC_HANDLE hService = ::CreateService(hServiceManager,
                                         m_szServiceName,
                                         m_szServiceName,
                                         SERVICE_ALL_ACCESS,
                                         SERVICE_WIN32_OWN_PROCESS,
                                         SERVICE_DEMAND_START,        // start condition
                                         SERVICE_ERROR_NORMAL,
                                         szFilePath,
                                         NULL,
                                         NULL,
                                         NULL,
                                         NULL,
                                         NULL);
    if (!hService)
    {
        ::CloseServiceHandle(hServiceManager);
        return FALSE;
    }

    // tidy up
    ::CloseServiceHandle(hService);
    ::CloseServiceHandle(hServiceManager);

    return TRUE;
}

//
// ChangeConfig
// ============
//
// Change the Configuration of an installed service
BOOL CYRService::ChangeConfig(const char* startup)
{
    // Firstly decide what startup we need
    DWORD dwStartup = SERVICE_AUTO_START;

    if (_stricmp(startup, "-a") == 0)
    {
        dwStartup = SERVICE_AUTO_START;
    }
    else if (_stricmp(startup, "-m") == 0)
    {
        dwStartup = SERVICE_DEMAND_START;
    }
    else if (_stricmp(startup, "-d") == 0)
    {
        dwStartup = SERVICE_DISABLED;
    }

    // Open the Service Control Manager
    SC_HANDLE hServiceManager = ::OpenSCManager(NULL,                   // local machine
                                NULL,                   // ServicesActive database
                                SC_MANAGER_ALL_ACCESS); // full access
    if (!hServiceManager)
        return FALSE;

    // Now oen the service allowing access to change the configuration
    SC_HANDLE hService = ::OpenService(hServiceManager,
                                       m_szServiceName,
                                       SERVICE_CHANGE_CONFIG);

    // If not found then cleanup and exit
    if (!hService)
    {
        ::CloseServiceHandle(hServiceManager);
        return FALSE;
    }


    // So Modify the service
    BOOL bSuccess = ChangeServiceConfig(hService,
                                        SERVICE_WIN32_OWN_PROCESS,
                                        dwStartup,
                                        SERVICE_ERROR_NORMAL,
                                        NULL,
                                        NULL,
                                        NULL,
                                        NULL,
                                        NULL,
                                        NULL,
                                        m_szServiceName);

    // Close the service
    ::CloseServiceHandle(hService);

    // Close the service manager
    ::CloseServiceHandle(hServiceManager);

    return bSuccess;
}

//
// Uninstall
// =========
//
// Uninstall the service
BOOL CYRService::Uninstall()
{
    // Open the Service Control Manager
    SC_HANDLE hServiceManager = ::OpenSCManager(NULL,                   // local machine
                                NULL,                   // ServicesActive database
                                SC_MANAGER_ALL_ACCESS);// full access
    if (!hServiceManager)
        return FALSE;

    BOOL bResult = FALSE;
    SC_HANDLE hService = ::OpenService(hServiceManager,
                                       m_szServiceName,
                                       DELETE);
    if (hService)
    {
        if (::DeleteService(hService))
        {
            bResult = TRUE;
        }
        ::CloseServiceHandle(hService);
    }

    ::CloseServiceHandle(hServiceManager);

    return bResult;
}

//
// StartService
// ============
//
// Service startup and registration
BOOL CYRService::StartService()
{
    // Define the Service Main - the Callback function
    SERVICE_TABLE_ENTRY st[] = {
        {m_szServiceName, ServiceMain},
        {NULL, NULL}
    };

    BOOL b = ::StartServiceCtrlDispatcher(st);
    return b;
}

//
// ServiceMain
// ===========
//
// static member function (callback)
void CYRService::ServiceMain(DWORD /*dwArgc*/, LPTSTR* /*lpszArgv*/)
{
    // Get a pointer to the C++ object
    CYRService* pService = m_pThis;

    // Register the control request handler
    pService->m_Status.dwCurrentState = SERVICE_START_PENDING;
    pService->m_hServiceStatus = RegisterServiceCtrlHandler(pService->m_szServiceName,
                                 Handler);
    if (pService->m_hServiceStatus == NULL)
    {
        return;
    }

    // Start the initialisation
    if (pService->Initialize())
    {
        // Do the real work.
        // When the Run function returns, the service has stopped.
        pService->m_bIsRunning = TRUE;
        pService->m_Status.dwWin32ExitCode = 0;
        pService->m_Status.dwCheckPoint = 0;
        pService->m_Status.dwWaitHint = 0;
        pService->Run();
    }

    // Tell the service manager we are stopped
    pService->SetStatus(SERVICE_STOPPED);
}

//
// SetStatus
// =========
//
// Set the status of the service
void CYRService::SetStatus(DWORD dwState)
{
    // Set the current status
    m_Status.dwCurrentState = dwState;

    // And send it to the service
    ::SetServiceStatus(m_hServiceStatus, &m_Status);
}


//
// Initialize
// ==========
//
// Service initialization
BOOL CYRService::Initialize()
{
    // Start the initialization
    SetStatus(SERVICE_START_PENDING);

    // Perform the actual initialization
    BOOL bResult = OnInit();

    // Set final state
    m_Status.dwWin32ExitCode = GetLastError();
    m_Status.dwCheckPoint = 0;
    m_Status.dwWaitHint = 0;

    if (!bResult)
    {
        SetStatus(SERVICE_STOPPED);
        return FALSE;
    }

    SetStatus(SERVICE_RUNNING);

    return TRUE;
}

//
// Run
// ===
//
// Run function to do the real work of the service

// This function performs the main work of the service.
// When this function returns the service has stopped.
void CYRService::Run()
{
    while (m_bIsRunning)
    {
        Sleep(5000);
    }
}

//
// Handler
// =======
//
// Control request handlers
//
// static member function (callback) to handle commands from the
// service control manager
void CYRService::Handler(DWORD dwOpcode)
{
    // Get a pointer to the object
    CYRService* pService = m_pThis;

    switch (dwOpcode)
    {
    case SERVICE_CONTROL_STOP: // 1
        {
            pService->SetStatus(SERVICE_STOP_PENDING);
            pService->OnStop();
            pService->m_bIsRunning = FALSE;
            break;
        }
    case SERVICE_CONTROL_PAUSE: // 2
        {
            pService->OnPause();
            break;
        }
    case SERVICE_CONTROL_CONTINUE: // 3
        {
            pService->OnContinue();
            break;
        }
    case SERVICE_CONTROL_INTERROGATE: // 4
        {
            pService->OnInterrogate();
            break;
        }
    case SERVICE_CONTROL_SHUTDOWN: // 5
        {
            pService->OnShutdown();
            break;
        }
    default:
        {
            break;
        }
    }

    // Set new status
    ::SetServiceStatus(pService->m_hServiceStatus, &pService->m_Status);
}

// Called when the service is first initialized
BOOL CYRService::OnInit()
{
    return TRUE;
}

// Called when the service control manager wants to stop the service
void CYRService::OnStop()
{}

// called when the service is interrogated

void CYRService::OnInterrogate()
{}

// called when the service is paused

void CYRService::OnPause()
{}

// called when the service is continued

void CYRService::OnContinue()
{}

// called when the service is shut down

void CYRService::OnShutdown()
{}

// called when the service gets a user control message

BOOL CYRService::OnUserControl(DWORD /*dwOpcode*/)
{
    return FALSE; // say not handled
}


