/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __PHLEX__
#define __PHLEX__


#include <iostream>
#include "DomStructure.h"
#include "Attributes.hpp"



// ---------------------------------------------------------------------------------------
// THREAD SAFETY: this class is NOT thread safe, but has no non-const static data members.
// =============  Therefore, thread safety is guaranteed only when an object of this class
//                is only accessed by it's creating thread.
// ---------------------------------------------------------------------------------------



template <typename U>
class PhLexer
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    PhLexer(CDOMStructure *pDS, U zInput, std::ostream& os = std::cout);


    //// --------------------------------------------------------------------------------
    //// Scanning function.
    //// --------------------------------------------------------------------------------

    // Get the next token and perform associated action.
    // The caller not interested in retrieving the token attributes may pass NULL for
    // pTokenAttributes.
    // The coordinates of the beginning of the token within the source buffer are
    // returned in the ulLine and ulCol arguments.
    long FindToken(unsigned long& ulLine, unsigned long& ulCol, CAttributes<U>* pTokenAttributes = NULL);


    //// --------------------------------------------------------------------------------
    //// Semantic functions.
    //// --------------------------------------------------------------------------------

    // Collect to end of the comment: -->
    void collectComment(void);

    // Collect to end of the script: </script
    void collectScript(void);

    // Does the token just found begin with <script ?
    BOOL InScript();

    // Collect to end of the text: opening tag.
    void collectText(void);

    // Inserting a text into the DOM.
    void InsertTextIntoDom(U pcText, U pcTextEnd);
    void InsertTextIntoDom(char *pcText);

#ifdef PHTRACE
    //// --------------------------------------------------------------------------------
    //// Output the lexeme.
    //// --------------------------------------------------------------------------------
    void DumpLexeme(std::ostream& os);
#endif

    //// --------------------------------------------------------------------------------
    //// Constants.
    //// --------------------------------------------------------------------------------
    enum
    {
        ENDMARKER = 0
    };

    //---------------------------------------------------------------------------

protected:

    //// --------------------------------------------------------------------------------
    //// Scanning functions.
    //// --------------------------------------------------------------------------------

    // Implement to set the token attributes member variable's common fields with the
    // proper values.
    // This method is called as soon as a token has been recognized, and for
    // all tokens.
    // After this method, one of SetDefaultTokenAttributes or OnToken is responsible
    // for setting the other fields of the attributes member variable.
    void SetCommonTokenAttributes(CAttributes<U>& tokenAttributes)
    {
        tokenAttributes.pLexeme_      = pcFirstChar_;
        tokenAttributes.ulLexemeSize_ = pcCurChar_ - pcFirstChar_;
    }

    // Implement to set the token attributes member variable with the proper value.
    // This method is called as soon as a token has been recognized, and ONLY for
    // tokens having NO associated semantic action (thus, returning a fixed token
    // value to the parser).
    // The responsibility of setting the token attributes member variable's value
    // in the case of a token having an associated semantic action is left to the
    // implementor of the semantic action.
    void SetDefaultTokenAttributes(CAttributes<U>& tokenAttributes)
    {}

    // Called when a token is found. If pTokenAttributes is NULL, the token
    // attributes are not retrieved.
    // This method is reponsible for setting the token attribute variable's
    // value.
    // Return the token, or -2 to continue scanning.
    // You must override this function.
    // Sample:
    // long OnToken(long lToken, T* pTokenAttributes = NULL)
    // {
    //      switch(lToken)
    //      {
    //          #include "wslexac.h"
    //      }
    //      return -2;
    // }

    long OnToken(long lToken, CAttributes<U>* pTokenAttributes = NULL);

    // Called when a lexical error is detected.
    BOOL OnError(long lLine, long lCol, BOOL bDuringRecovery);

    // Called on an unrecovered lexical error.
    void OnFatalError(BOOL bDuringRecovery);

    // Called after a lexical error, when attempting error recovery.
    void OnEnterErrorRecovery(BOOL bDuringRecovery);

    // Called when recovered from a lexical error.
    void OnErrorRecovered()
    {}

    // Automaton transations.

    unsigned long PackedNextState(unsigned long ulS, unsigned long ulA);

    // Update the line and column counters.
    void UpdateLineCol(unsigned char uc)
    {
        if(uc == '\n')
        {
            lLine_++;
            lColumn_ = 1;
        }
        else if(uc != '\r')
            lColumn_++;
    }

    //// --------------------------------------------------------------------------------
    //// Scanner Data.
    //// --------------------------------------------------------------------------------
    std::ostream&   os_;                    // The output stream
    U               pcInput_;               // The input buffer
    long*  plGetToken_;            // GetToken[state] table
    U               pcCurChar_;             // Pointer to the current character in the input buffer.
    long   lColumn_, lLine_;       // Current column and line in input buffer
    long   lCurCol_, lCurLine_;    // Current column and line, passed to the parser. Normally aligned
    // on token boundaries.
    U               pcFirstChar_;           // Pointer to the current character in the input buffer.
    long   lLexErrorType_;         // Variable to return to the parser the error type.
    // 0 means no error type, -1 means memory alloc error,
    // other values are user values
    //// --------------------------------------------------------------------------------
    //// Semantic Data.
    //// --------------------------------------------------------------------------------
    CDOMStructure *pDomStructure_;
    BOOL  bInTag_;                          // TRUE when we are inside a tag.
    BOOL  bInScript_;                       // TRUE when we are inside a script.
    U     pcText_;                          // Pointer to the text preceding the token (if any).

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const PhLexer&);
};

/************* USEFUL MACROS **********************/

// Continue reading the token at the current input character.
// This is useful when, after matching a lexical rule, some
// extra characters where ignored or skipped. PH_RESUME
// should be the last instruction executed since it breaks
// out of the switch.
#define PH_RESUME   {                                                   \
bMore = TRUE;                                   \
ulState = ulSavedState;                         \
return -2;                                      \
}

// Skip (ignore) the current character, and read the next one
// immediately. Does not change the token already recognized.
#define PH_SKIP     {                                                   \
unsigned char uc = *pcCurChar_;                 \
if(uc)                                          \
{                                               \
UpdateLineCol(uc);                          \
++pcCurChar_;                               \
}                                               \
}

// Ignore the current character in input buffer, and continue
// reading the token. PH_MORE should be the last instruction
// executed since it breaks out of the switch.
#define PH_MORE     {                                                   \
unsigned char uc = *pcCurChar_;                 \
if(uc)                                          \
{                                               \
UpdateLineCol(uc)                           \
bMore = TRUE;                               \
++pcCurChar_;                               \
ulState = ulSavedState;                     \
return -2;                                  \
}                                               \
}

/************* END OF USEFUL MACROS **********************/


#include "phleximpl.hpp"


#endif // defined(__PHLEX__)
