/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __PHPARSE__
#define __PHPARSE__


#include "phlex.hpp"
#include<vector>




// ---------------------------------------------------------------------------------------
// THREAD SAFETY: this class is NOT thread safe, but has no non-const static data members.
// =============  Therefore, thread safety is guaranteed only when an object of this class
//                is only accessed by it's creating thread.
// ---------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------


template <typename U>
class PhParser
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    PhParser(CDOMStructure *pDS, PhLexer<U>& lexer, std::ostream& os = std::cout)
: lexer_(lexer), os_(os), pDomStructure_(pDS), pczValue_(NULL),
    pczAttribute_(NULL), bDocType_(FALSE), iAttribute_(0),
    cLastData_(' ')
    {}

    ~PhParser()
    {
        if(pczValue_)
            delete pczValue_;
        if(pczAttribute_)
            delete pczAttribute_;
    }


    enum AfterErrorAction
    {
        RAISE_FATAL_ERROR,      // Emit a fatal error
        TRY_TO_RECOVER,         // Attempt to recover from error
        STOP_PARSER             // Stop the parser & go to the accepting state (successful termination).
    };


    //// --------------------------------------------------------------------------------
    //// Parsing Functions
    //// --------------------------------------------------------------------------------

    // Parsing algorithm.
    BOOL Parse(unsigned long* pulErrorLine = NULL, unsigned long* pulErrorCol = NULL);


    //// --------------------------------------------------------------------------------
    //// Accessors
    //// --------------------------------------------------------------------------------
    const PhLexer<U>& getLexer() const
    {
        return lexer_;
    }
    PhLexer<U>& getLexer()
    {
        return lexer_;
    }

    //---------------------------------------------------------------------------

protected:

    enum ParsingPhase
    {
        PHASE_NORMAL,
        PHASE_ERROR
    };

    struct StackEntry
    {
        StackEntry() : ulSymbol_(0)
        {}

        StackEntry(unsigned long ulSymbol, const CAttributes<U>& t = CAttributes<U>()) : ulSymbol_(ulSymbol), symbolAttributes_(t)
        {}


        unsigned long   ulSymbol_;
        CAttributes<U>  symbolAttributes_;
    };

    //// --------------------------------------------------------------------------------
    //// Parsing helpers
    //// --------------------------------------------------------------------------------

    // Parsing table.
    long ActionGoto(unsigned long ulState, unsigned long ulSymbol);

    // Pop n entries from the stack.
    void Pop(unsigned long ulLen);


    //// --------------------------------------------------------------------------------
    //// Parsing events. To be overriden.
    //// --------------------------------------------------------------------------------

    // Called when the parser detects a syntax error.
    AfterErrorAction OnError(long lLine, long lCol, BOOL bDuringRecovery);

    // Called on a syntax error, when the parser enters the recovery phase.
    void OnEnterErrorRecovery(BOOL /*bDuringRecovery*/)
    {}

    // Called on an unrecovered lexical error.

    void OnFatalError(BOOL bDuringRecovery);

    // Called when recovered from a lexical error.
    void OnErrorRecovered()
    {}

    // Called after the parser reaches the ACCEPT state.

    void OnAccept()
    {}

    // Called when the parser is stopped by its caller

    void OnStopParser()
    {}

    // Called to execute a semantic action. Must be overriden.
    // Return TRUE to continue parsing, FALSE to stop.

    BOOL PerformSemanticAction(CAttributes<U>& leftSymbolAttributes, long /*lProduction*/,
                               StackEntry* /*pProdAttr*/, ParsingPhase& /*ulPhase*/);


    //// --------------------------------------------------------------------------------
    //// Stack manipulation from within a semantic action.
    //// --------------------------------------------------------------------------------
    CAttributes<U>& GetATT(long l, StackEntry* pEntry)
    {
        return pEntry[l].symbolAttributes_;
    }
    const CAttributes<U>& GetATT(long l, StackEntry* pEntry) const
    {
        return pEntry[l].symbolAttributes_;
    }

    //// --------------------------------------------------------------------------------
    //// Semantic helpers.
    //// --------------------------------------------------------------------------------
    void freeBuffers(void);
    void endDocType(void);
    void beginOpenTag(U tagName, unsigned long tagLength);
    void endOpenTag(void);
    void beginOpenDoctype(U tagName, unsigned long tagLength);
    void endTagEmpty(void);
    void beginEndTag(U tagName, unsigned long tagLength);
    void insertAttributeWithValue(void);
    void insertAttributeWithoutValue(void);
    void setAttributes(int);
    void setAttributeName(U tagName, unsigned long tagLength);
    void setAttributeValue(U tagName, unsigned long tagLength);
    void CallBackStartTag(const char *);
    void CallBackEndTag(const char *);
    void CallBackStartAttribute(const char *, const char *);
    void InsertValueIntoDom(const char *);

    //// --------------------------------------------------------------------------------
    //// Parser Data
    //// --------------------------------------------------------------------------------
    typedef std::vector<StackEntry> StackType;
    std::ostream&   os_;            // The output stream
    StackType       stack_;         // Dynamic parser stack
    PhLexer<U>&     lexer_;         // Lexical Analyzer

    //// --------------------------------------------------------------------------------
    //// Semantic data.
    //// --------------------------------------------------------------------------------
    CDOMStructure *pDomStructure_;
    char          *pczValue_;
    char          *pczAttribute_;
    BOOL           bDocType_;
    int            iAttribute_;
    char           cLastData_;

    //// --------------------------------------------------------------------------------
    //// Constants.
    //// --------------------------------------------------------------------------------
    enum
    {
        BLANK       = -32768,   // Value of a parsing table entry that indicates a syntax error.
        ACCEPT      =  32767,   // "Accept" value of a parsing table entry.
        ERROR_TOKEN =      1,   // The error token. For error productions (recovery).
        END_OF_LIST =     -1    // Marker for end of list in AGPool array.
    };

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const PhParser&);
};


//// --------------------------------------------------------------------------------
//// Semantic action macros.
//// --------------------------------------------------------------------------------

// Retrieving a symbol's attributes (right side of the production).
#define ATT(l) GetATT(l, pProdAttr)

// Exiting the error recovery mode, simulating successful error recovery.
#define ExitErrorRecovery() (ulPhase = PHASE_NORMAL, onErrorRecovered())


#include "phparseimpl.hpp"


#endif  // defined(__PHPARSE__)
