/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


/**********************************
*              PHOENIX            *
* Base class for parser.          *
* Class definition.               *
**********************************/


#include "poptions.h"                   // Compilation options. Generated by parser generator or lexical generator.
// Includes choice of target platform and array packing
#include "wsparrt.h"                    // User code and support routines, Part 1 and Part 3 of the parser source program
#include "wsparseref.h"                 // Static production length and left side arrays





// Lookup in the parsing table.
// Arguments :  ulState (IN) : the current state.
//              ulSymbol (IN): the current input symbol (terminal or nonterminal).
// Return    :  parsing table entry value: an action or a goto state.
template <typename U>
inline long PhParser<U>::ActionGoto(unsigned long ulState, unsigned long ulSymbol)
{
    unsigned int uiAGPoolIndex, uiAGs;

    if(ulSymbol < NONTERMINAL_START)
    {
        /* Action */
        if(ActionState[ulState] < 0)
            return ActionState[ulState];   // Simple reduction!
        uiAGPoolIndex = ActionState[ulState] * 2;
        while((uiAGs = AGPool[uiAGPoolIndex]) != ulSymbol && uiAGs != END_OF_LIST)
            uiAGPoolIndex += 2;
    }
    else
    {
        /* Goto */
        uiAGPoolIndex = GotoSymbol[ulSymbol - NONTERMINAL_START] * 2;
        while((uiAGs = AGPool[uiAGPoolIndex]) != ulState && uiAGs != END_OF_LIST)
            uiAGPoolIndex += 2;
    }

    return AGPool[uiAGPoolIndex + 1];
}


// Parsing algorithm.
// Return:  TRUE if success, FALSE if fatal parser error.
template <typename U>
inline BOOL PhParser<U>::Parse(unsigned long* pulErrorLine, unsigned long* pulErrorCol)
{
    BOOL bError = FALSE;
    BOOL bContinueParsing = TRUE;
    ParsingPhase ulPhase = PHASE_NORMAL;
    long lToken = 0, lAction = 0, lProduction = 0, lState = 0, lPlen = 0;
    unsigned long ulTokenLine = 0, ulTokenCol = 0;
    unsigned long ulShiftError = 0;
    BOOL bNeedNextToken = TRUE;
    CAttributes<U> tokenAttributes = CAttributes<U>();
    CAttributes<U> leftSymbolAttributes = CAttributes<U>();

    // Push state 0 on parser stack.
    stack_.push_back(StackEntry());

    do
    {
        if(bNeedNextToken)
        {
            // Read next token.
            lToken = lexer_.FindToken(ulTokenLine, ulTokenCol, &tokenAttributes);
            if(lToken == -1 || (bError && lToken == PhLexer<U>::ENDMARKER))
            {
                OnFatalError(ulPhase == PHASE_ERROR);
                return FALSE;
            }
        }

        if(pulErrorLine)
            *pulErrorLine = ulTokenLine;
        if(pulErrorCol)
            *pulErrorCol  = ulTokenCol;

        AfterErrorAction afterErrorAction;
        lAction = ActionGoto(stack_.back().ulSymbol_, lToken);
        switch(lAction)
        {
        case BLANK:           // Syntax error
            bError = TRUE;

            // Signal the error.
            afterErrorAction = OnError(ulTokenLine, ulTokenCol, ulPhase == PHASE_ERROR);
            switch(afterErrorAction)
            {
            case RAISE_FATAL_ERROR:
                OnFatalError(TRUE);
                bContinueParsing = FALSE;
                break;

            case STOP_PARSER:
                OnAccept();
                bContinueParsing = FALSE;
                bError = FALSE;
                break;

            case TRY_TO_RECOVER:
            default:
                break;
            }
            if(!bContinueParsing)
                break;

            if(ulPhase == PHASE_ERROR)
            {
                // Pop states from stack to return to error state
                while(ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN) < 0)
                    Pop(1);

                // Shift the error virtual token
                ulShiftError = ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN);
                stack_.push_back(StackEntry(ulShiftError));
            }
            else
            {
                // Pop states from stack to try to find error state
                while(stack_.size() > 0 && ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN) < 0)
                    Pop(1);

                if(ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN) < 0)
                {
                    // Error state not found: Fatal error.
                    OnFatalError(FALSE);
                    bContinueParsing = FALSE;
                }
                else
                {
                    //// Found error state

                    // Signal beginning of recovery phase.
                    OnEnterErrorRecovery(ulPhase == PHASE_ERROR);

                    // Enter error recovery phase.
                    ulPhase = PHASE_ERROR;

                    // Shift the error virtual token.
                    ulShiftError = ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN);
                    stack_.push_back(StackEntry(ulShiftError));
                }
            }
            bNeedNextToken = TRUE;
            break;

        case ACCEPT:
#ifdef PHTRACE
            os_ << "@@@ ACCEPT" << std::endl;
#endif
            OnAccept();
            bContinueParsing = FALSE;
            break;

        default:    // Shift or reduce
            if(lAction > 0) // SHIFT
            {
                stack_.push_back(StackEntry(lAction, tokenAttributes));

#ifdef PHTRACE
                os_ << "@@@ SHIFT " << lAction << "(";
                lexer_.DumpLexeme(os_);
                os_ << ")" << std::endl;
#endif
                bNeedNextToken = TRUE;
            }
            else           // REDUCE
            {
                lProduction = -lAction;
                lPlen = ProdLength[lProduction];

                // Define an entry pointing one below the leftmost symbol's
                // entry in the stack.
                StackEntry* pProdAttr = stack_.end() - lPlen - 1;

#ifdef PHTRACE
                os_ << "@@@ REDUCE " << lProduction << std::endl;
#endif

                // Perform the semantic action associated with the production.
                bContinueParsing = PerformSemanticAction(leftSymbolAttributes, lProduction, pProdAttr, ulPhase);

                // Did the semantic action ask the parser to stop?
                if(!bContinueParsing)
                {
                    if(!bError)
                        OnStopParser();
                    else
                        OnFatalError(ulPhase == PHASE_ERROR);
                    break;
                }

                // Pop the production's symbols off the stack.
                Pop(lPlen);

                // Push the nonterminal on the left side of the production on the stack.
                lState = ActionGoto(stack_.back().ulSymbol_, LeftSide[lProduction]);
                stack_.push_back(StackEntry(lState, leftSymbolAttributes));
                bNeedNextToken = FALSE;
            }
            break;
        }
    }
    while(bContinueParsing);

    return !bError;
}

// Called after the parser reaches a syntax error.
// lLine, lCol: the line & column of the beginning of the token where the error appeared.
// bDuringRecovery: TRUE error was detected while attempting to recover from a previous
// error.
// Return:  RAISE_FATAL_ERROR:  Emit a fatal error
//          TRY_TO_RECOVER:     Attempt to recover from error
//          STOP_PARSER         Stop the parser go to the accepting state
//                                  (successful termination).
template <typename U>
inline PhParser<U>::AfterErrorAction PhParser<U>::OnError(long lLine, long lCol, BOOL /*bDuringRecovery*/)
{
#ifdef PHTRACE
    os_ << "Syntax error line " << lLine << ", column " << lCol << std::endl;
#else
    lLine;
    lCol;    // To avoid "unreferenced formal parameter" warning.
#endif
    return TRY_TO_RECOVER;    // By default, attempt to recover from error.
}

// Called on an unrecovered syntax error.
// bDuringRecovery: TRUE if function was called while attempting to recover
// from a previous error.

template <typename U>
inline void PhParser<U>::OnFatalError(BOOL /*bDuringRecovery*/)
{
#ifdef PHPARSE
    os_ << " FATAL ERROR" << std::endl;
#endif
}

template <typename U>
inline void PhParser<U>::Pop(unsigned long ulLen)
{
    while(ulLen)
    {
        stack_.pop_back();
        ulLen--;
    }
}

template <typename U>
void PhParser<U>::freeBuffers(void)
{
    if( pczValue_ )
    {
        delete(pczValue_ ) ;
        pczValue_ = NULL ;
    }
    if( pczAttribute_ )
    {
        delete(pczAttribute_ ) ;
        pczAttribute_ = NULL ;
    }
}

template <typename U>
void PhParser<U>::endDocType(void)
{
    // end tag for doctype ->bDocType = true
    bDocType_ = FALSE;
    CallBackEndTag( "" );
}

template <typename U>
void PhParser<U>::beginOpenTag(U tagName, unsigned long tagLength)
{
    //-> inject tag in the DOM Structure
    char *yytext = new char[tagLength + 1];
    for(unsigned long ul = 0; ul < tagLength; ul++)
    {
        yytext[ul] = (char)*tagName;
        ++tagName;
    }
    yytext[tagLength] = '\0';

    CallBackStartTag( yytext );

    delete yytext;
}

template <typename U>
void PhParser<U>::endOpenTag(void)
{}


template <typename U>
void PhParser<U>::beginOpenDoctype(U tagName, unsigned long tagLength)
{
    // doc type ->bDocType = true
    bDocType_ = TRUE;
    // start tag
    // -> inject tag in the DOM Structure
    char *yytext = new char[tagLength + 1];
    for(unsigned long ul = 0; ul < tagLength; ul++)
    {
        yytext[ul] = (char)*tagName;
        ++tagName;
    }
    yytext[tagLength] = '\0';

    CallBackStartTag( yytext );

    delete yytext;
}

template <typename U>
void PhParser<U>::endTagEmpty(void)
{
    // end tag
    CallBackEndTag( "" );
}

template <typename U>
void PhParser<U>::beginEndTag(U tagName, unsigned long tagLength)
{
    // end tag
    char *yytext = new char[tagLength + 1];
    for(unsigned long ul = 0; ul < tagLength; ul++)
    {
        yytext[ul] = (char)*tagName;
        ++tagName;
    }
    yytext[tagLength] = '\0';

    CallBackEndTag( yytext );

    delete yytext;
}

template <typename U>
void PhParser<U>::insertAttributeWithValue(void)
{
    // attribute = value : if attribute and value exist -> inject attribute in the DOM Structure
    if( pczAttribute_ && pczValue_ )
    {
        CallBackStartAttribute( pczAttribute_ , pczValue_ );
        delete( pczValue_ );
        pczValue_ = NULL;
        delete( pczAttribute_ );
        pczAttribute_ = NULL;
    }
    // deals no more with attributes -> bAttribute = 0
    iAttribute_ = 0 ;
}

template <typename U>
void PhParser<U>::insertAttributeWithoutValue(void)
{
    // attribute (alone) : if attribute exists -> inject attribute in the DOM Structure
    if( pczAttribute_ && iAttribute_ == 0 )
    {
        CallBackStartAttribute( pczAttribute_ , "" );
        delete( pczAttribute_ ) ;
        pczAttribute_ = NULL ;
    }
}

template <typename U>
void PhParser<U>::setAttributes(int i)
{
    // deals with attributes -> bAttribute = 1
    iAttribute_ = i;
}

template <typename U>
void PhParser<U>::setAttributeName(U tagName, unsigned long tagLength)
{
    char *yytext = new char[tagLength + 1];
    for(unsigned long ul = 0; ul < tagLength; ul++)
    {
        yytext[ul] = (char)*tagName;
        ++tagName;
    }
    yytext[tagLength] = '\0';


    // if we deal with text -> initialize the text by appending strings.
    if( iAttribute_ == 1 )
    {
        if( pczValue_ )
            delete( pczValue_ ) ;
        pczValue_ = new char[strlen( yytext ) + 1]; // malloc( (strlen( yytext ) + 1) * sizeof(char) ) ;
        strcpy( pczValue_ , yytext ) ;
    }
    // The attribute is alone -> initialize the attribute
    else
    {
        if( pczAttribute_ )
            delete( pczAttribute_ ) ;
        pczAttribute_ = new char[strlen( yytext ) + 1]; //malloc( (strlen( yytext ) + 1) * sizeof(char) ) ;
        strcpy( pczAttribute_ , yytext ) ;
    }

    delete yytext;
}

template <typename U>
void PhParser<U>::setAttributeValue(U tagName, unsigned long tagLength)
{
    char *yytext = new char[tagLength + 1];
    for(unsigned long ul = 0; ul < tagLength; ul++)
    {
        yytext[ul] = (char)*tagName;
        ++tagName;
    }
    yytext[tagLength] = '\0';

    // initialize the attribute value
    if(pczValue_)
        delete(pczValue_);
    pczValue_ = new char[strlen( yytext ) + 1]; // malloc( (strlen( yytext ) + 1) * sizeof(char) ) ;
    strcpy(pczValue_, yytext);

    // if we deal with DocType -> initialize the text
    if(bDocType_)
    {
        if(pczValue_)
        {
            lexer_.InsertTextIntoDom(pczValue_);
            delete(pczValue_);
            pczValue_ = NULL;
        }
    }

    delete yytext;
}

// Callback : found a tag
template <typename U>
void PhParser<U>::CallBackStartTag( const char *stoken )
{
    char *pczBuffer = new char[ strlen( stoken ) + 1 ] ;
    strcpy( pczBuffer , stoken  ) ;
    pDomStructure_->InsertTagChild( pczBuffer + 1 ) ;
    if( pczBuffer )
        delete pczBuffer ;
}

// Callback : found the end of a tag
template <typename U>
void PhParser<U>::CallBackEndTag( const char *stoken )
{
    if( stricmp( stoken , "" ) == 0 )
        return ;
    char* pczEndTag = new char[ strlen( stoken ) + 1] ;
    strcpy( pczEndTag , stoken ) ;
    pDomStructure_->EndTag( pczEndTag + 2 ) ;
    if( pczEndTag )
        delete pczEndTag ;

}

// Callback : found an attribute
template <typename U>
void PhParser<U>::CallBackStartAttribute( const char *pczAttribute, const char *pczValue)
{
    pDomStructure_->InsertAttributeChild( (char*)pczAttribute , (char*)pczValue ) ;
}

template <typename U>
void PhParser<U>::InsertValueIntoDom(const char *pczText)
{
    pDomStructure_->InsertTextChild( (char*)pczText ) ;
}

template <typename U>
inline
BOOL PhParser<U>::PerformSemanticAction(CAttributes<U>& leftSymbolAttributes, long lProduction, StackEntry* pProdAttr, ParsingPhase& ulPhase)
{
    BOOL bRet = TRUE;           // By default, continue parsing.
    switch(lProduction)
    {
#include "wsparac.h"    // Phoenix - generated parser actions.
    default:
        // By default, the left nonterminal's attributes are copied from the
        // leftmost symbol's in the right side of the production.
        leftSymbolAttributes = GetATT(1, pProdAttr);
        break;
    }
    return bRet;
}
