/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_ASCBIN                  */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains the routine for performing conversions      */
/*  from a numeric ASCII string to Binary.                   */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <ctype.h>
#include <limits.h>
#include <stddef.h>
#include <string.h>

/*
 *  Function prototypes
 */
#include psl_common_file (psl_rtndef.h) /* PSL Routines */

static unsigned long PSL_ASCBIN_NEW (long numchr,
                                     char *ascstr_ptr,
                                     long radix,
                                     void *number_ptr,
                                     long *chrcnt_ptr,
                                     unsigned long flags);

static unsigned long CNV_NEGATIVE (long numchr,
                                   char *ascstr_ptr,
                                   long radix,
                                   void *number_ptr,
                                   long *chrcnt_ptr);

static unsigned long CNV_POSITIVE (long numchr,
                                   char *ascstr_ptr,
                                   long radix,
                                   void *number_ptr,
                                   long *chrcnt_ptr);

static unsigned long CNV_UNSIGNED (long numchr,
                                   char *ascstr_ptr,
                                   long radix,
                                   void *number_ptr,
                                   long *chrcnt_ptr);

/*
 *  Structure definitions.
 */
#include psl_common_file (shrmsg.h) /* PSL messages */
#include psl_common_file (ascbin_prmdef.h)
/* PSL_ASCBIN Symbols */

/*
 *  Define list of valid characters.
 */
static char valchr [] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";


/************************************************************************/
/*                                  */
/*  PSL_ASCBIN - Convert a numeric string to Binary         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*    Iret = PSL_ASCBIN (numchr_ptr, ascstr_ptr, radix_ptr,     */
/*               number_ptr, chrcnt_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*    Inputs:                           */
/*      numchr_ptr = Pointer to number of characters to be      */
/*           converted.                 */
/*      ascstr_ptr = Pointer to String containing number.       */
/*      radix_ptr  = Pointer to the conversion radix.       */
/*                                  */
/*    Outputs:                          */
/*      number_ptr = Pointer to the binary Value equivalent.    */
/*      chrcnt_ptr = Pointer to the number of unconverted       */
/*           characters.                    */
/*                                  */
/*    Returns:                          */
/*     iret     = Unsigned Long PSL Routine Return Status       */
/*                                  */
/*  This routine converts a string of numeric digit into its    */
/*  Binary equivalent.                      */
/*                                  */
/*                                  */
/*  NOTE: The (binary) number returned is treated as "unsigned" */
/*        or "signed" depending on the conversion radix; hence  */
/*        its declaration as type VOID (see comment later).     */
/*                                  */
/************************************************************************/

unsigned long PSL_ASCBIN (unsigned char *numchr_ptr,
                          char *ascstr_ptr,
                          unsigned char *radix_ptr,
                          void *number_ptr,
                          unsigned short *chrcnt_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* General status return */
    long       numchr;     /* Number of characters to convert */
    long       radix;      /* Conversion Radix */
    long       chrcnt;     /* Unconverted characters */
    unsigned long       flags;      /* Flags */

    /*
     *  Set up parameters.
     */
    numchr = *numchr_ptr;
    radix = *radix_ptr;

    if (radix == 10)
    {
        flags = 0;
    }
    else
    {
        flags = ASCBIN_UNSIGNED;
    }

    /*
     *  Simply call the new version of the routine with new parameters.
     */
    iret = PSL_ASCBIN_NEW (
               numchr,
               ascstr_ptr,
               radix,
               number_ptr,
               &chrcnt,
               flags);

    /*
     *  Return information to the caller.
     */
    *chrcnt_ptr = (short) chrcnt;

    return iret;
}


/************************************************************************/
/*                                  */
/*  PSL_ASCBIN_NEW - Convert a numeric string to Binary     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_ASCBIN_NEW (numchr, ascstr_ptr, radix,        */
/*                number_ptr, chrcnt_ptr, flags);   */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Routine Return Status     */
/*     numchr     = Number of characters to be converted        */
/*     ascstr_ptr = Pointer to String containing number     */
/*     radix      = Conversion radix in the range 2-36      */
/*     number_ptr = Pointer to the binary Value equivalent      */
/*     chrcnt_ptr = Pointer to the number of unconverted characters */
/*     flags      = Conversion Flags                */
/*          ASCBIN_UNSIGNED - perform unsigned conversion   */
/*                                  */
/*  This routine converts a string of numeric digits into its   */
/*  Binary equivalent.                      */
/*                                  */
/************************************************************************/

static unsigned long PSL_ASCBIN_NEW (long numchr,
                                     char *ascstr_ptr,
                                     long radix,
                                     void *number_ptr,
                                     long *chrcnt_ptr,
                                     unsigned long flags)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* General status return */

    /*
     *  Check that the radix supplied is valid.
     */
    if ((radix < 2) || (radix > 36))    /* Invalid radix supplied? */
    {
        iret = PSL_BADRADIX;        /* Return 'Invalid Radix' */
        goto end;           /* And exit */
    }

    /*
     *  Handle the various types of conversion in lower level routines.
     */
    if ((flags & ASCBIN_UNSIGNED) == 0) /* If signed conversion required */
    {
        if ((numchr > 1)        /* If > 1 character */
                && (*ascstr_ptr == '-'))   /* and leading '-' */
        {
            iret = CNV_NEGATIVE (   /* Convert as negative number */
                       numchr-1,   /* Adjusted characters to convert */
                       ascstr_ptr+1,   /* Adjusted string pointer */
                       radix,      /* Specified Radix */
                       number_ptr, /* Returning value here */
                       chrcnt_ptr);    /* Unconverted character count */
        }
        else if ((numchr > 1)       /* If > 1 character */
                 && (*ascstr_ptr == '+'))  /* and leading '+' */
        {
            iret = CNV_POSITIVE (   /* Convert as positive number */
                       numchr-1,   /* Adjusted characters to convert */
                       ascstr_ptr+1,   /* Adjusted string pointer */
                       radix,      /* Specified Radix */
                       number_ptr, /* Returning value here */
                       chrcnt_ptr);    /* Unconverted character count */
        }
        else
        {
            iret = CNV_POSITIVE (   /* Convert as positive number */
                       numchr,     /* Specified characters to convert */
                       ascstr_ptr, /* Specified string pointer */
                       radix,      /* Specified Radix */
                       number_ptr, /* Returning value here */
                       chrcnt_ptr);    /* Unconverted character count */
        }
    }
    else
    {
        iret = CNV_UNSIGNED (       /* Convert as unsigned number */
                   numchr,     /* Specified characters to convert */
                   ascstr_ptr, /* Specified string pointer */
                   radix,      /* Specified Radix */
                   number_ptr, /* Returning value here */
                   chrcnt_ptr);    /* Unconverted character count */
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  CNV_NEGATIVE - Convert a numeric string to Binary       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = CNV_NEGATIVE (numchr, ascstr_ptr, radix,      */
/*              number_ptr, chrcnt_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Routine Return Status     */
/*     numchr     = Number of characters to be converted        */
/*     ascstr_ptr = Pointer to String containing number     */
/*     radix      = Conversion radix                */
/*     number_ptr = Pointer to the binary Value equivalent      */
/*     chrcnt_ptr = Pointer to the number of unconverted characters */
/*                                  */
/*  This routine converts a string of numeric digits into its   */
/*  negative signed value in decimal radix.             */
/*                                  */
/************************************************************************/

static unsigned long CNV_NEGATIVE (long numchr,
                                   char *ascstr_ptr,
                                   long radix,
                                   void *number_ptr,
                                   long *chrcnt_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* General status return */
    long       jsub;       /* Loop Control */
    char       *char_ptr;      /* Pointer to current character */
    char       *wrk_ptr;       /* Work Pointer to char */
    long       chrval;     /* Character 'value' */
    long       number;     /* Running total */

    /*
     *  Assume Success.
     */
    iret = PSL_NORMAL;

    /*
     *  Loop through the characters in the string converting each.
     */
    for (jsub = numchr, char_ptr = ascstr_ptr, number = 0;
            jsub != 0;
            jsub--, char_ptr++)
    {

        /*
         *  Get the 'value' of the current character.
         */
        wrk_ptr = memchr (valchr, toupper (*char_ptr), radix);
        if (wrk_ptr == NULL)
        {
            goto end_processing;
        }
        else
        {
            chrval = wrk_ptr - valchr;
        }

        /*
         *  Check for overflow.
         *  Return error if multiplying running total by radix exceeds max.
         */
        if (number < (LONG_MIN/radix))
        {
            iret = PSL_NUMTOOBIG;   /* Return 'Number too big' */
            goto end;           /* And exit */
        }

        /*
         *  Multiply by the radix to get the next significant digit.
         */
        number *= radix;

        /*
         *  Return error if adding next digit to total exceeds max.
         */
        if (number < (LONG_MIN + chrval))
        {
            iret = PSL_NUMTOOBIG;   /* Return 'Number too big' */
            goto end;           /* And exit */
        }

        number -= chrval;       /* Update total */

    }                   /* End of (for loop) */

    /*
     *  END_PROCESSING:
     *
     *  Perform processing required when we reach the end of the (valid)
     *  characters.
     *  Test for unconverted characters, and report if there are any.
     */
end_processing:

    if (jsub != 0)          /* 'Unconverted chars' != 0 ? */
    {
        iret = PSL_UNCONVCHARS;     /* Set Unconverted Characters error */
    }
    *chrcnt_ptr = jsub;         /* Return 'Unconverted chars' */

    *(long *)number_ptr = number;   /* Return value */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  CNV_POSITIVE - Convert a numeric string to Binary       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = CNV_POSITIVE (numchr, ascstr_ptr, radix,      */
/*              number_ptr, chrcnt_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Routine Return Status     */
/*     numchr     = Number of characters to be converted        */
/*     ascstr_ptr = Pointer to String containing number     */
/*     radix      = Conversion radix                */
/*     number_ptr = Pointer to the binary Value equivalent      */
/*     chrcnt_ptr = Pointer to the number of unconverted characters */
/*                                  */
/*  This routine converts a string of numeric digits into its   */
/*  positive signed value in decimal radix.             */
/*                                  */
/************************************************************************/

static unsigned long CNV_POSITIVE (long numchr,
                                   char *ascstr_ptr,
                                   long radix,
                                   void *number_ptr,
                                   long *chrcnt_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* General status return */
    long       jsub;       /* Loop Control */
    char       *char_ptr;      /* Pointer to current character */
    char       *wrk_ptr;       /* Work Pointer to char */
    long       chrval;     /* Character 'value' */
    long       number;     /* Running total */

    /*
     *  Assume Success.
     */
    iret = PSL_NORMAL;

    /*
     *  Loop through the characters in the string converting each.
     */
    for (jsub = numchr, char_ptr = ascstr_ptr, number = 0;
            jsub != 0;
            jsub--, char_ptr++)
    {

        /*
         *  Get the 'value' of the current character.
         */
        wrk_ptr = memchr (valchr, toupper (*char_ptr), radix);
        if (wrk_ptr == NULL)
        {
            goto end_processing;
        }
        else
        {
            chrval = wrk_ptr - valchr;
        }

        /*
         *  Check for overflow.
         *  Return error if multiplying running total by radix exceeds max.
         */
        if (number > (LONG_MAX/10))
        {
            iret = PSL_NUMTOOBIG;   /* Return 'Number too big' */
            goto end;           /* And exit */
        }

        /*
         *  Multiply by the radix to get the next significant digit.
         */
        number *= radix;

        /*
         *  Return error if adding next digit to total exceeds max.
         */
        if (number > (LONG_MAX - chrval))
        {
            iret = PSL_NUMTOOBIG;   /* Return 'Number too big' */
            goto end;           /* And exit */
        }

        number += chrval;       /* Update total */

    }                   /* End of (for loop) */

    /*
     *  END_PROCESSING:
     *
     *  Perform processing required when we reach the end of the (valid)
     *  characters.
     *  Test for unconverted characters, and report if there are any.
     */
end_processing:

    if (jsub != 0)          /* 'Unconverted chars' != 0 ? */
    {
        iret = PSL_UNCONVCHARS;     /* Set Unconverted Characters error */
    }
    *chrcnt_ptr = jsub;         /* Return 'Unconverted chars' */

    *(long *)number_ptr = number;   /* Return value */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  CNV_UNSIGNED - Convert a numeric string to Binary       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = CNV_UNSIGNED (numchr, ascstr_ptr, radix,      */
/*              number_ptr, chrcnt_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Routine Return Status     */
/*     numchr     = Number of characters to be converted        */
/*     ascstr_ptr = Pointer to String containing number     */
/*     radix      = Conversion radix                */
/*     number_ptr = Pointer to the binary Value equivalent      */
/*     chrcnt_ptr = Pointer to the number of unconverted characters */
/*                                  */
/*  This routine converts a string of numeric digits into its   */
/*  unsigned value in any radix.                    */
/*                                  */
/************************************************************************/

static unsigned long CNV_UNSIGNED (long numchr,
                                   char *ascstr_ptr,
                                   long radix,
                                   void *number_ptr,
                                   long *chrcnt_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* General status return */
    long       jsub;       /* Loop Control */
    char       *char_ptr;      /* Pointer to current character */
    char       *wrk_ptr;       /* Work Pointer to char */
    unsigned long       chrval;     /* Character 'value' */
    unsigned long       number;     /* Running total */

    /*
     *  Assume Success.
     */
    iret = PSL_NORMAL;

    /*
     *  Loop through the characters in the string converting each.
     */
    for (jsub = numchr, char_ptr = ascstr_ptr, number = 0;
            jsub != 0;
            jsub--, char_ptr++)
    {

        /*
         *  Get the 'value' of the current character.
         */
        wrk_ptr = memchr (valchr, toupper (*char_ptr), radix);
        if (wrk_ptr == NULL)
        {
            goto end_processing;
        }
        else
        {
            chrval = wrk_ptr - valchr;
        }

        /*
         *  Check for overflow.
         *  Return error if multiplying running total by radix exceeds max.
         */
        if (number > (ULONG_MAX/radix))
        {
            iret = PSL_NUMTOOBIG;   /* Return 'Number too big' */
            goto end;           /* And exit */
        }

        /*
         *  Multiply by the radix to get the next significant digit.
         */
        number *= radix;

        /*
         *  Return error if adding next digit to total exceeds max.
         */
        if (number > (ULONG_MAX - chrval))
        {
            iret = PSL_NUMTOOBIG;   /* Return 'Number too big' */
            goto end;           /* And exit */
        }

        number += chrval;       /* Update total */

    }                   /* End of (for loop) */

    /*
     *  END_PROCESSING:
     *
     *  Perform processing required when we reach the end of the (valid)
     *  characters.
     *  Test for unconverted characters, and report if there are any.
     */
end_processing:

    if (jsub != 0)          /* 'Unconverted chars' != 0 ? */
    {
        iret = PSL_UNCONVCHARS;     /* Set Unconverted Characters error */
    }
    *chrcnt_ptr = jsub;         /* Return 'Unconverted chars' */

    *(unsigned long *)number_ptr = number;
    /* Return value */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}
