/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_FAO                     */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This routine formats parameters into a message buffer using an       */
/*  approach similar to SYS$FAO.                         */
/*                                       */
/*===========================================================================*/


#include  <psl_macros.h>        /* PSL Standard Macros */

/*    Define useful Include structures   */

#include  <stdarg.h>            /* Variable Argument List handling */
#include  <string.h>            /* String functions */

/*    Define Function Prototypes for all external routines */
/*    used by the module (and for ourselves) */

#include  psl_common_file(psl_rtndef.h) /* Define ourselves */

/*   Define Data Structures and such-like */

#include  psl_common_file(binasc_prmdef.h) /* BINASC statuses */
#include  psl_common_file(shrmsg.h) /* Message Statuses */


/************************************************************************/
/*                                  */
/*  PSL_FAO - Format Parameters into a String           */
/*                                  */
/*  This routine emulates a small part of the VMS SYS$FAO       */
/*  system service and formats a string from a skeleton string  */
/*  containing a set of !xx directives and the appropriate      */
/*  parameters.                         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_FAO (out_ptr, outlen, inp_ptr, prm1, ...);       */
/*                                  */
/*  Where:                              */
/*                                  */
/*     out_ptr = Pointer to Output Buffer               */
/*     outlen  = Size of Output Buffer              */
/*     inp_ptr = Pointer to Skeleton String             */
/*     prmi    = ith parameter to be formatted.         */
/*                                  */
/*  The number and type of the parameters for the string are    */
/*  determined by the insertion directives in the string itself.    */
/*  These directives are currently as follows:          */
/*                                  */
/*     !AS  - Pointer to Null-Terminated String         */
/*     !AF  - Pointer to Null-Terminated String (with non-printing) */
/*     !nUL - Unsigned Long (in Decimal)                */
/*     !nUW - Unsigned Short (in Decimal)               */
/*     !nUB - Unsigned Char (in Decimal)                */
/*     !nSL - Signed Long (in Decimal)              */
/*     !nSW - Signed Short (in Decimal)             */
/*     !nSB - Signed Char (in Decimal)              */
/*     !nXL - Unsigned Long (in Hexadecimal)            */
/*     !nXW - Unsigned Short (in Hexadecimal)           */
/*     !nXB - Unsigned Char (in Hexadecimal)            */
/*                                  */
/*  For the numeric conversions the 'n', if specified, indicates    */
/*  a field width within which the output is displayed, right-  */
/*  justified.  If 'n' is omitted the output is left-justified  */
/*  and occupies only as much space as is necessary.        */
/*                                  */
/*  There is also one special insertion directive that does not */
/*  take a parameter.                       */
/*                                  */
/*     !%S - Insert s iff last numeric parameter was not 1      */
/*                                  */
/*  Note that this routine formats a single control string into */
/*  a single message buffer and, as such, deliberately does NOT */
/*  support the !/ directive.                   */
/*                                  */
/************************************************************************/

void PSL_FAO (char *out_ptr, unsigned long outlen, char *inp_ptr, ...)
{

    /*   Local Data   */

    va_list    args;       /* Argument List Pointer */

    /*   Set up the Variable Argument List pointer and call PSL_FAO_L */

    va_start (args, inp_ptr);   /* Initialize Pointer */
    PSL_FAO_L (out_ptr, outlen, inp_ptr, &args);
    /* Call PSL_FAO_L */

    /*   All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}
/************************************************************************/
/*                                  */
/*  PSL_FAO_L - Format Parameters into a String         */
/*                                  */
/*  This routine does all the hard work of PSL_FAO and is       */
/*  called thereby.  It is also called by other routines such   */
/*  as PSL_PUTDIAG, but is not available to the user.       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_FAO_L (out_ptr, outlen, inp_ptr, va_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     out_ptr = Pointer to Output Buffer               */
/*     outlen  = Size of Output Buffer              */
/*     inp_ptr = Pointer to Skeleton String             */
/*     va_ptr  = Pointer to va_list structure for extra parameters  */
/*                                  */
/************************************************************************/

void PSL_FAO_L (char *out_ptr, unsigned long outlen, char *inp_ptr,
                va_list *va_ptr)
{

    /*   Local Data   */

    char       *fmt_ptr;       /* Pointer to next character in format */
    char       *msg_ptr;       /* Pointer to next character in message */
    Boolean    last_single;    /* Last number single flag */
    char       *chr_ptr;       /* Pointer to character parameter */
    unsigned long       unsigned_number;    /* Unsigned numeric parameter */
    long       signed_number;  /* signed numeric parameter */
    unsigned char       numlen;     /* Numeric buffer length */
    unsigned char       radix;      /* Radix for conversion */
    unsigned short      flags;      /* Conversion Flags */
    unsigned short      chrcnt;     /* Unconverted character count */
    unsigned long       buflen;     /* Buffer Length */
    unsigned long       bufsiz;     /* Room left in buffer */
    long       field_size;     /* Field Size for !nUW et al. */
    unsigned long       iret;       /* Routine status return */

    /*   Initialize various local variables */

    last_single = PFALSE;       /* Last number was not 1 */
    msg_ptr = out_ptr;      /* Current position in output */
    bufsiz = outlen - 1;        /* Size of buffer (excl null at end) */

    /*   Loop through the format string looking for the next        */
    /*   directive, and format the next parameter as required when      */
    /*   found.                             */

    for  (fmt_ptr = inp_ptr; *fmt_ptr != '\0'; fmt_ptr++)
        /* For each character in format string */
    {
        if  (*fmt_ptr == '!')   /* If a directive character */
        {



            /*   First check to see if we have a numeric value and, if so,      */
            /*   convert it.                            */

            numlen = strlen(fmt_ptr+1);
            /* Get length of rest of string */
            radix = 10;     /* Radix is decimal */
            iret = PSL_ASCBIN ( /* Convert from ASCII */
                       &numlen,  /* with length of buffer */
                       fmt_ptr+1,    /* buffer address */
                       &radix,   /* radix */
                       &field_size,
                       /* into signed number */
                       &chrcnt); /* with count of unconv chars */
            if  ((!psl_test_status(iret))
                    /* If an error returned */
                    &&  (iret != PSL_UNCONVCHARS))
                /* other than expected one */
            {
                goto not_directive; /* Can't be directive */
            }
            if  (chrcnt == numlen)  /* If no characters converted */
            {
                chrcnt = 1;     /* Show no characters absorbed */
                field_size = 0; /* And default conversion */
            }
            else            /* Else (some characters converted) */
            {
                if  (field_size <= 0) goto not_directive;
                /* If number <= 0, can't be directive */
                chrcnt = 1 + (numlen - chrcnt);
                /* Calculate number absorbed */
            }



            /*   Now switch on the next format character to see what type it is */

            switch (*(fmt_ptr+chrcnt))
                /* Case on next character */
            {

                /*   First handle the unsigned decimal types (!UL, !UW and !UB).    */
                /*   Although these are basically the same they must use        */
                /*   size-specific calls to VA_ARG.                 */

            case 'U':       /* If unsigned decimal */
                {
                    if  (*(fmt_ptr+chrcnt+1) == 'L')
                        /* If long */
                    {
                        unsigned_number = va_arg (*va_ptr, unsigned long);
                        /* get the parameter */
                        numlen = 20; /* and set field width */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'W')
                        /* If short */
                    {
                        unsigned_number = va_arg (*va_ptr, unsigned short);
                        /* get the parameter */
                        numlen = 10; /* and set field width */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'B')
                        /* If char */
                    {
                        unsigned_number = va_arg (*va_ptr, unsigned char);
                        /* get the parameter */
                        numlen = 5; /* and set field width */
                    }
                    else        /* Else (unrecognized type) */
                    {
                        goto not_directive;
                        /* just copy it over */
                    }
                    radix = 10; /* Set the radix */

                    if  (field_size == 0)
                        /* If default directive */
                    {
                        flags = BINASC_JUSTIFY;
                        /* Set for left-justified unsigned*/
                    }
                    else        /* Else (directive with size) */
                    {
                        numlen = (char) field_size;
                        /* Reset field width */
                        flags = 0;  /* And set for right-justified */
                    }

                    if  (numlen > bufsiz) numlen = bufsiz;
                    /* Constrain to size of buffer */
                    (void) PSL_BINASC (
                        /* Convert the number */
                        &unsigned_number,
                        &numlen,
                        &radix,
                        msg_ptr,
                        &flags);
                    if  (unsigned_number == 1)
                        /* If it was 1 */
                    {
                        last_single = PTRUE;
                        /* say so */
                    }
                    else        /* Else */
                    {
                        last_single = PFALSE;
                        /* say so */
                    }
                    goto end_number; /* and finish off */
                }

                /*   Next handle the signed decimal types (!SL, !SW and !SB).       */
                /*   These are much the same as before, but need to use a signed    */
                /*   variable and need different calls to VA_ARG.           */

            case 'S':       /* If signed decimal */
                {
                    if  (*(fmt_ptr+chrcnt+1) == 'L')
                        /* If long */
                    {
                        signed_number = va_arg (*va_ptr, long);
                        /* get the parameter */
                        numlen = 20; /* and set field width */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'W')
                        /* If short */
                    {
                        signed_number = va_arg (*va_ptr, short);
                        /* get the parameter */
                        numlen = 10; /* and set field width */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'B')
                        /* If char */
                    {
                        signed_number = va_arg (*va_ptr, char);
                        /* get the parameter */
                        numlen = 5; /* and set field width */
                    }
                    else        /* Else (unrecognized type) */
                    {
                        goto not_directive;
                        /* just copy it over */
                    }
                    radix = 10; /* Set the radix */

                    if  (field_size == 0)
                        /* If default directive */
                    {
                        flags = BINASC_SIGN | BINASC_JUSTIFY;
                        /* Set for left-justified signed */
                    }
                    else        /* Else (directive with size) */
                    {
                        numlen = (char) field_size;
                        /* Reset field width */
                        flags = BINASC_SIGN;
                        /* And set for right-justified signed */
                    }

                    if  (numlen > bufsiz) numlen = bufsiz;
                    /* Constrain to size of buffer */
                    (void) PSL_BINASC (
                        /* Convert the number */
                        &signed_number,
                        &numlen,
                        &radix,
                        msg_ptr,
                        &flags);
                    if  (signed_number == 1)
                        /* If it was 1 */
                    {
                        last_single = PTRUE;
                        /* say so */
                    }
                    else        /* Else */
                    {
                        last_single = PFALSE;
                        /* say so */
                    }
                    goto end_number; /* and finish off */
                }

                /*   Next handle the hexadecimal types (!XL, !XW and !XB).  These   */
                /*   are much the same as before.                   */

            case 'X':       /* If unsigned hexadecimal */
                {
                    if  (*(fmt_ptr+chrcnt+1) == 'L')
                        /* If long */
                    {
                        unsigned_number = va_arg (*va_ptr, unsigned long);
                        /* get the parameter */
                        numlen = 8; /* and set field width */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'W')
                        /* If short */
                    {
                        unsigned_number = va_arg (*va_ptr, unsigned short);
                        /* get the parameter */
                        numlen = 4; /* and set field width */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'B')
                        /* If char */
                    {
                        unsigned_number = va_arg (*va_ptr, unsigned char);
                        /* get the parameter */
                        numlen = 2; /* and set field width */
                    }
                    else        /* Else (unrecognized type) */
                    {
                        goto not_directive;
                        /* just copy it over */
                    }
                    radix = 16; /* Set the radix */

                    if  (field_size == 0)
                        /* If default directive */
                    {
                        flags = BINASC_FILL | BINASC_JUSTIFY;
                        /* and set for left-justified unsigned*/
                        /* with leading zeros */
                    }
                    else        /* Else (directive with size) */
                    {
                        numlen = (char)field_size;
                        /* Reset field width */
                        flags = BINASC_FILL;
                        /* and set for right-justified unsigned*/
                        /* with leading zeros */
                    }

                    if  (numlen > bufsiz) numlen = bufsiz;
                    /* Constrain to size of buffer */
                    (void) PSL_BINASC (
                        /* Convert the number */
                        &unsigned_number,
                        &numlen,
                        &radix,
                        msg_ptr,
                        &flags);
                    if  (unsigned_number == 1)
                        /* If it was 1 */
                    {
                        last_single = PTRUE;
                        /* say so */
                    }
                    else        /* Else */
                    {
                        last_single = PFALSE;
                        /* say so */
                    }
                    goto end_number; /* and finish off */
                }

                /*   Next handle the ASCII types (!AS and !AF).  These are      */
                /*   identical except that we want to replace non-printing      */
                /*   characters in the latter with dots.  Note that the parameters  */
                /*   for these are pointers to null-terminated strings which is NOT */
                /*   the same as for SYS$FAO.                       */

            case 'A':       /* If ASCII string */
                {
                    if  (field_size != 0) goto not_directive;
                    /* Preceding numbers not allowed */
                    if  (*(fmt_ptr+chrcnt+1) == 'S')
                        /* If normal */
                    {
                        chr_ptr = va_arg (*va_ptr, char *);
                        /* get pointer to string */
                        buflen = strlen(chr_ptr);
                        /* Get length of segment */
                        if  (buflen > bufsiz) buflen = bufsiz;
                        /* Constrain to size of buffer */
                        (void) memcpy (msg_ptr, chr_ptr, buflen);
                        /* Copy to buffer */
                    }
                    else if  (*(fmt_ptr+chrcnt+1) == 'F')
                        /* If with non-printing */
                    {
                        chr_ptr = va_arg (*va_ptr, char *);
                        /* get pointer to string */
                        buflen = strlen(chr_ptr);
                        /* Get length of segment */
                        if  (buflen > bufsiz) buflen = bufsiz;
                        /* Constrain to size of buffer */
                        (void) memcpy (msg_ptr, chr_ptr, buflen);
                        /* Copy to buffer */
                        (void) STR_REPLACE_NONPRINTING (
                            /* And replace non-printing characters */
                            msg_ptr,
                            buflen,
                            '.');
                    }
                    else        /* Else (unrecognized type) */
                    {
                        goto not_directive;
                        /* just copy it over */
                    }
                    msg_ptr += buflen;
                    /* And update pointer */
                    bufsiz -= buflen;
                    /* And reset buffer size */
                    fmt_ptr += chrcnt+1;
                    /* And step over directive */
                    goto get_next;  /* And get next one */
                }

                /*   Lastly the special (!%S).  We just insert an 's' if the last   */
                /*   number was not 1 and omit it otherwise.                */

            case '%':       /* If special */
                {
                    if  (field_size != 0) goto not_directive;
                    /* Preceding numbers not allowed */
                    if  (*(fmt_ptr+chrcnt+1) != 'S') goto not_directive;
                    /* Ignore if not normal */
                    if  ((!last_single) && (bufsiz != 0))
                        /* If last number was multiple */
                        /* and buffer is not full */
                    {
                        *msg_ptr = 's';
                        /* Add an 's' */
                        msg_ptr += 1;
                        /* And update pointer */
                        bufsiz -= 1; /* And buffer size */
                    }
                    fmt_ptr += chrcnt+1;
                    /* Step over directive */
                    goto get_next;  /* And get next one */
                }
            default:        /* Owt else */
                {
                    goto not_directive;
                    /* Just copy it over */
                }
            }           /* End of (switch on directive type) */
        }               /* End of (if directive) */
        else            /* Else (not directive) */
        {
            goto not_directive; /* Just copy it over */
        }



        /*   END_NUMBER: branch here to do the common tidying up after      */
        /*   formatting one of the numeric types.               */

end_number:
        if  (field_size == 0)   /* If default directive */
        {
            *(msg_ptr + numlen) = ' ';
            /* Ensure at least one space */
            buflen = (strchr(msg_ptr, ' ') - msg_ptr);
            /* Hence buffer length */
        }
        else            /* Else */
        {
            buflen = field_size;    /* Buffer length is what he said */
        }
        msg_ptr += buflen;      /* Update pointer */
        bufsiz -= buflen;       /* And reset buffer size */
        fmt_ptr += chrcnt+1;    /* And step over directive */
        goto get_next;      /* And get next one */



        /*   NOT_DIRECTIVE: branch here if not a directive to just copy     */
        /*   over the character and then check the next.            */

not_directive:
        if  (bufsiz != 0)       /* If buffer is not full */
        {
            *msg_ptr = *fmt_ptr;    /* Copy it over */
            msg_ptr += 1;       /* Increment pointer */
            bufsiz -= 1;        /* Decrement buffer size */
        }
        goto get_next;      /* And get next character */

        /*   GET_NEXT: branch here to get next format character */

get_next:
        ;
    }               /* End of (for each format character) */

    /*   End of string - return to caller */

    *msg_ptr = '\0';        /* null-terminate buffer */
    return;             /* Return to caller */
}
