/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// CECStatusBar.cpp
// ----------------
//
//    This is the implementation file for the CECStatusBar class
//
//=========================================================================

#include "stdafx.h"
#include "CECStatusBar.h"
#include "resource.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CECStatusBar

CECStatusBar::CECStatusBar()
{
    m_nFrame = 0;
    m_bAnimationRunning = FALSE;
}

CECStatusBar::~CECStatusBar()
{}



BEGIN_MESSAGE_MAP(CECStatusBar, CStatusBar)
//{{AFX_MSG_MAP(CECStatusBar)
ON_WM_DRAWITEM()
//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CECStatusBar message handlers

void CECStatusBar::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
    switch(lpDrawItemStruct->itemID)
    {
    case 1:
        {
            // Attach to a CDC object
            CDC dc;
            dc.Attach(lpDrawItemStruct->hDC);

            // Get the pane rectangle and calculate text coordinates
            CRect rect(&lpDrawItemStruct->rcItem);

            //Draw the frame
            DrawCurrentFrame(&dc,lpDrawItemStruct->rcItem);

            // Detach from the CDC object, otherwise the hDC will be
            // destroyed when the CDC object goes out of scope
            dc.Detach();
            return;
        }
    }
    CStatusBar::DrawItem(lpDrawItemStruct);
}

UINT CECStatusBar::MultiThreadedPlayFunction( LPVOID pParam )
{
    // Grab hold of the status bar ...
    CECStatusBar *pWorkerObject=(CECStatusBar*)pParam;
    ASSERT (pWorkerObject != NULL);

    // Reset the animation
    pWorkerObject->Reset();

    // And start the ball rolling ...
    pWorkerObject->AdvanceFrame();
    pWorkerObject->DrawCurrentFrame();

    //Since the main application thread is busy timers don't work
    //So create an event and wait on it for the duration of a frame
    //Since the event is never signalled the wait times out and
    //we can draw the next frame.Keep doing this till the animation is stopped

    //Create an unsignalled event
    HANDLE hEvent = CreateEvent(NULL,FALSE,FALSE,NULL);

    // And keep looping until the timer stops
    while (pWorkerObject->IsAnimationRunning())
    {
        //Wait on this event, it is never signalled, so it times out
        //after the animation delay
        if(WaitForSingleObject(hEvent,pWorkerObject->GetAnimationFrameDelay())==WAIT_TIMEOUT)
        {
            //Draw the next frame
            pWorkerObject->AdvanceFrame();
            pWorkerObject->DrawCurrentFrame();
        }
    }

    //Don't forget to close the handle or we get resource leaks
    CloseHandle(hEvent);

    // And redraw the window to show it
    pWorkerObject->RedrawWindow();

    // And quit the Worker Thread
    return 0;
}

//
// Play
// ====
//
// Function to start playing the bitmap sequence
//
void CECStatusBar::Play()
{
    //If we're not already playing
    if(!m_bAnimationRunning)
    {
        m_bAnimationRunning=true;
        // Kick the animation in a worker thread
        AfxBeginThread(MultiThreadedPlayFunction, this);
    }
}


//
// Stop
// ====
//
// Function to stop playing the bitmap sequence
//
void CECStatusBar::Stop()
{
    //Are we actually playing?
    if(!m_bAnimationRunning)
        return;

    m_bAnimationRunning=false;

    // Allow the current frame to be updated
    ::Sleep(GetAnimationFrameDelay());

    // Reset the animmation
    Reset();

    // And redraw the window once more
    RedrawWindow();
}

//
// GetFrame
// ========
//
// Function to retrieve the bitmap next in line
//
CBitmap* CECStatusBar::GetFrame()
{
    // Translate the integer into a bitmap ID
    UINT nID = GetBitmapID();

    CBitmap* pBitmap = new CBitmap();
    pBitmap->LoadBitmap(nID);

    return pBitmap;
}

//
// GetBitmapID
// ===========
//
// Function to retrieve the bitmap next in line
//
UINT CECStatusBar::GetBitmapID()
{
    // Just find the bitmap associated with this frame
    for (UINT nID = IDB_STATUS_BAR_FRM0; nID <= IDB_STATUS_BAR_FRM0 + m_nFrame; nID++);

    // And return its ID value
    return nID-1;
}

//
// AdvanceFrame
// ============
//
// Function to advance the bitmap frame
//
void CECStatusBar::AdvanceFrame()
{
    // Advance to the next frame
    m_nFrame++;

    // Reset if we have looped round
    if (m_nFrame >= 32)
        Reset();
}

void CECStatusBar::DrawCurrentFrame()
{
    //
    //Get the panes rectangle
    //
    //Use the window proc direct, because
    //a) Using GetItemRect(index,rect) will not work if this is in second thread
    //b) SendMessage(SB_GETRECT,1, (LPARAM)&rcPane); will not work because
    //  main thread is busy and messages are not being processed
    CRect rcPane;
    DefWindowProc(SB_GETRECT,1, (LPARAM)&rcPane);

    CClientDC dc(this);

    //Draw the currently selected frame
    DrawCurrentFrame(&dc,rcPane);
}

void CECStatusBar::DrawCurrentFrame(CDC *pDC,const CRect& rcFrame)
{
    CDC srcDC; // select current bitmap into a compatible CDC
    srcDC.CreateCompatibleDC(NULL);

    CBitmap* pBitmap = GetFrame();

    CBitmap* pOldBitmap = srcDC.SelectObject(pBitmap);
    pDC->BitBlt(rcFrame.left, rcFrame.top, rcFrame.Width(), rcFrame.Height(),
                &srcDC, 0, 0, SRCCOPY); // BitBlt to pane rect

    srcDC.SelectObject(pOldBitmap);

    pBitmap->DeleteObject();
    delete pBitmap;
}
