/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// CECTargetEnvironment.cpp: implementation of the CECTargetEnvironment class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CECTargetEnvironment.h"
#include "..\Interfaces\CyrPlugIns.h"
#include "CECRegistryTools.h"
#include "CECCyranoHome.h"
#include "Solutions.h"
#include "server.h"

#pragma warning(disable:4786)

//
//    Define the current context key and value.
//
#define CONTEXT_VALUE   "Context"
#define CONTEXT_COMMAND "COMMAND"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CECTargetEnvironment::CECTargetEnvironment()
{
}


CECTargetEnvironment::~CECTargetEnvironment()
{
}


void    CECTargetEnvironment::GetTargets(CComboBox &ComboBox) const
{
    //
    //    Get a list of the licenced target environments or contexts
    //
    CStringArray TargetArray;
    GetLicencedTargets(TargetArray);

    //
    //    COMMAND is always supported targets so add it .
    //
    char CommandTarget[32];
    Tgt_GetTarget (TARGET_COMMAND ,CommandTarget ,sizeof(CommandTarget) ,FALSE);
    ComboBox.AddString(CommandTarget);

    //
    //    Now loop through each of the targets and add them to the list (we have already checked
    //    that we are licenced for the targets
    //
    for (int isub = 0; isub < TargetArray.GetSize(); isub++)
    {
        ComboBox.AddString(TargetArray[isub]);
    }

    TargetArray.RemoveAll();
}


//
//    GetTargets
//
//    Recover (as a string array) a list of the currently supported targets
//
void    CECTargetEnvironment::GetTargets    (CStringArray& TargetArray) const
{

    //
    //    COMMAND is always suported targets so add them now.
    //
    char CommandTarget[32];
    Tgt_GetTarget (TARGET_COMMAND ,CommandTarget ,sizeof(CommandTarget) ,FALSE);
    TargetArray.Add (CommandTarget);

    //
    //    Get a list of the licenced target environments or contexts
    //
    GetLicencedTargets(TargetArray);
}


void    CECTargetEnvironment::GetLicencedTargets    (CStringArray& TargetArray) const
{
    //
    //    Get a list of the INSTALLED targets
    //
    GetInstalledTargets (TargetArray);

    //
    //    Now loop through each of the targets, checking to see for which we are licenced and
    //    if we are then adding it to the list
    //
    int Count = TargetArray.GetSize();
    int ThisEntry = 0;

    while (ThisEntry < Count)
    {
        CString Target = TargetArray.GetAt(ThisEntry);

        enum TARGETID TargetID = Tgt_GetTargetID (Target.GetBuffer(0));

        if (!IsTargetSupported(TargetID ,TRUE))
        {
            TargetArray.RemoveAt(ThisEntry);
            Count--;
        }
        else
        {
            ThisEntry++;
        }
    }

    return;
}

//
//    GetCurrentContext
//
//    Return as a string the curren context held in the registry
//
void    CECTargetEnvironment::GetCurrentContext(CString& Context) const
{

    //
    //    Get the current target context from the registry
    //
    CString ContextKey;
    ContextKey.Format("%s\\Engines\\Context" ,SOLUTIONS_KEY);

    CECRegistryTools reg;
    Context = reg.GetStringValue(HKEY_CURRENT_USER ,ContextKey ,CONTEXT_VALUE);

    //
    //    ...and ensure that the value returned is a valid target and return a null context if it is not
    //
#pragma message ("CECTargetEnvironment::GetCurrentContext: Taken out target validation!!!")
    //  if ((Tgt_GetTargetID (Context.GetBuffer(0)) == -1))
    //      Context.Empty();

    return;
}


//
//    SetCurrentContext
//
//    Save the specified context back into the registry
//
int CECTargetEnvironment::SetCurrentContext (CString &Context)
{
    HKEY    hRegKey;
    LPDWORD lpdwDisposition = 0;
    CString  szRegKey = SOLUTIONS_KEY + CString("\\Engines\\Context");
    LPCTSTR lpszValue = Context;

    CString ContextKey;
    ContextKey.Format("%s\\Engines\\Context" ,SOLUTIONS_KEY);

    // First lets open the key
    if (RegCreateKeyEx (HKEY_CURRENT_USER, ContextKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hRegKey, lpdwDisposition) == ERROR_SUCCESS)
    {
        // And set the default to be the chosen database
        if (RegSetValueEx(hRegKey, CONTEXT_VALUE, NULL, REG_SZ, (BYTE * const)lpszValue, (lstrlen(lpszValue)+1)*sizeof(TCHAR)) != ERROR_SUCCESS)
        {
            AfxMessageBox("Failed to save the Current Target", MB_OK);
            return -1;
        }
    }
    return 0;
}


//
//  IsTargetSupported
//  =================
//
//  Wrapper function for the main IsTargetSupported function
//
BOOL CECTargetEnvironment::IsTargetSupported    (char* pTarget) const
{
    TARGETID TargetID = Tgt_GetTargetID (pTarget);
    return IsTargetSupported (TargetID);
}

//
// IsTargetSupported
// =================
//
// This function will determine whether any particular database is supported
// the authorisation flag is optional if this info is needed.
//
// Calling Format:
//  int IsTargetSupported   (TARGETID TargetID ,CheckAuthorized);
//
// Parameters:
//  TargetID - TARGETID for the target to check for.
//  CheckAuthorized - BOOL value, TRUE if we are to check not only that this is a valid target
//                        environemnt but also that we are authorized for it.
//
// Returns:
//  0   - The target is authorized
//  -1  - Target is not authorized
//
BOOL CECTargetEnvironment::IsTargetSupported    (enum TARGETID TargetID, BOOL bCheckAuthorized /*=TRUE*/) const
{
    BOOL    bSupported = TRUE;

    //
    //    ... Always authorize when running without any protection
    //
#if defined(NOPROTECTION)
    if (TargetID != TARGET_UNKNOWN)
    {
        return TRUE;
    }
#endif

    //
    //    ...then check to ensure that we have a valid target environment returning FALSE if we don't
    //
    char    TargetString[32];
    if (Tgt_GetTarget (TargetID ,TargetString ,sizeof (TargetString) ,TRUE) == -1) return FALSE;

    //
    //    If requested, then check the licence to see if we actually authorized for the specified
    //    target environment
    //
    if (bCheckAuthorized)
    {
        //      Status = Tgt_IsTargetAuthorised (TargetID);
        //      if (Status != 0) bSupported = FALSE;
    }

    else
    {
        bSupported = FALSE;
    }
    return bSupported;
}



//
//    GetInstalledTargets
//    ===================
//
//    Pick up and display the version information for the plug-ins.
//
void    CECTargetEnvironment::GetInstalledTargets (CStringArray& TargetArray) const
{

    //
    //    Loop through all of the sub-keys below the main product key in the registry looking for the
    //    snap-in components.
    //
    CECRegistryTools reg;
    CString ProductKey ,PluginKey;

    ProductKey = SOLUTIONS_KEY;
    ProductKey += "\\Engines";
    LONG lResult = reg.Open(HKEY_LOCAL_MACHINE, ProductKey);

    if (lResult != ERROR_SUCCESS) return;

    TCHAR   szBuffer[256];
    DWORD   dwSize = 256;
    DWORD   Index = 0;
    CString Version;

    HKEY Key = reg.GetHKey();

    while (RegEnumKeyEx(Key, Index, szBuffer, &dwSize, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
    {

        //
        //    For each sub-key we should determine if it begins with "for " as this indicates a target
        //
        CString Target = szBuffer;
        if (Target.Left(4) == "for ")
        {
            Target = Target.Mid(4);
            Target.MakeUpper();
            TargetArray.Add(Target);
        }

        Index++;
        dwSize = 256;
    }
}

void CECTargetEnvironment::GetTargets(CListBox &ListBox) const
{
    //
    //    Get a list of the licenced target environments or contexts
    //
    CStringArray TargetArray;
    GetLicencedTargets(TargetArray);

    //
    //    COMMAND is always supported targets so add them now.
    //
    char CommandTarget[32];
    Tgt_GetTarget (TARGET_COMMAND ,CommandTarget ,sizeof(CommandTarget) ,FALSE);
    ListBox.AddString(CommandTarget);

    //
    //    Now loop through each of the targets and add them to the list (we have already checked
    //    that we are licenced for the targets
    //
    for (int isub = 0; isub < TargetArray.GetSize(); isub++)
    {
        ListBox.AddString(TargetArray.GetAt (isub));
    }

    TargetArray.RemoveAll();
}

//
//    InvokeRecorder
//    ==============
//
//    Invoke the Test Recorder appropriate for the current default environment (if any)
//    optionally passing it the name of the project and TOF to work on
//
int CECTargetEnvironment::InvokeRecorder    (enum TARGETID TargetID ,const CString& ObjectName) const
{

    // Get the path to both the generic recorder and modeller which are both in the Impact directory
    CECCyranoHome Home;
    CString szHomeDir, szCommandLine;
    Home.GetCyranoHome (szHomeDir);

    // We need to pass the username, password, Project, TOF name as command line arguments to the executable
    // so get these now
    CYRServer *pServer = (CYRServer*)AfxGetMainWnd()->SendMessage(WMU_GET_SERVERPTR);

    // Now get the server login information from the framework
    CString szUsername = pServer->GetUserName();
    CString szPassword = pServer->GetPassword();
    CString szProjectName = pServer->GetFullOpenProjectName();

    // And pass these in along with the Project Name we are connecting to
    CString szArgs = "--user " + szUsername;

    if (!szPassword.IsEmpty())
    {
        szArgs += " --pwd " + szPassword;
    }

    szArgs += " --project \"" + szProjectName + "\"";

    // And don't pass in a tof as the modeller will do it for us
    if (!ObjectName.IsEmpty())
    {
        CString szFile = ObjectName;
        szFile.MakeUpper();
        szArgs += " --file " + szFile;
    }

    // If we have not been passed the context of the file to be recorded then pick up the default
    CString szCurrentContext = "";
    CString szImageToInvoke;

    if (TargetID == TARGET_UNKNOWN)
    {
        GetCurrentContext(szCurrentContext);
    }

    //
    //    If the current context (target) is COMMAND then we invoke the generic shell program.
    if (szCurrentContext.CompareNoCase(CONTEXT_COMMAND) == 0)
    {
        // Format the command line
        szCommandLine.Format("%sEngines\\TRecorder.exe %s", szHomeDir, szArgs);
    }
    else
    {
        // Must be in another Mode so ...
        char chExecutable[32];
        enum TARGETID TargetId = Tgt_GetTargetID (szCurrentContext.GetBuffer(0));

        Tgt_GetTargetExecutable (TargetId, WHERE, chExecutable, sizeof (chExecutable));
        szImageToInvoke = chExecutable;

        szCurrentContext += "\\";
        //
        // We need to then format the command line
        szCommandLine.Format("%sEngines\\%sModeller\\%s.exe %s",
                             szHomeDir, szCurrentContext, szImageToInvoke, szArgs);
    }

    // ...and run the image.
    InvokeImage (szCommandLine);
    return 0;
}

int CECTargetEnvironment::InvokeModeller(enum TARGETID TargetID ,const CString &ObjectName) const
{

    // Get the path to both the generic recorder and modeller which are both in the Impact directory
    CECCyranoHome Home;
    CString szHomeDir, szCommandLine;
    Home.GetCyranoHome (szHomeDir);

    //Get hold of the server
    CYRServer *pServer = (CYRServer*)AfxGetMainWnd()->SendMessage(WMU_GET_SERVERPTR);

    // Firstly we check to see if the file actually exists
    CString strDll(SOLUTIONS_GENERIC_DLL);
    CString szWorkingDir = pServer->GetFolderPath(CYRServer::enSourceFolder);

    // Now work out the full path to the local file
    CString szFilePath = szWorkingDir + "\\" + ObjectName;

    // And see if it exists - if it doesn't go straight into recording mode
    CFileStatus FileStatus;
    if (CFile::GetStatus(szFilePath, FileStatus) == FALSE)
    {
        InvokeRecorder(TargetID, ObjectName);
        return 0;
    }

    // We need to pass the username, password, Project, TOF name as command line arguments to the executable
    // so get these now

    // Now get the server login information from the framework
    CString szUsername = pServer->GetUserName();
    CString szPassword = pServer->GetPassword();
    CString szProjectName = pServer->GetFullOpenProjectName();

    // And pass these in along with the Project Name we are connecting to
    CString szArgs = "--user " + szUsername;

    if (!szPassword.IsEmpty())
    {
        szArgs += " --pwd " + szPassword;
    }

    szArgs += " --project \"" + szProjectName + "\"";

    // And pass in the object
    if (!ObjectName.IsEmpty())
    {
        CString szFile = ObjectName;
        szFile.MakeUpper();
        szArgs += " --file " + szFile;
    }

    // If we have not been passed the context of the file to be recorded then pick up the default
    CString szCurrentContext = "";
    CString szImageToInvoke;

    if (TargetID == TARGET_UNKNOWN)
    {
        GetCurrentContext(szCurrentContext);
    }
    //
    //    If the current context (target) is COMMAND then we invoke the generic shell program.
    if (szCurrentContext.CompareNoCase(CONTEXT_COMMAND) == 0)
    {
        //Format the command line
        szCommandLine.Format("%sEngines\\TModeller.exe %s", szHomeDir, szArgs);
    }
    else
    {
        // Must be in another Mode so ...
        char chExecutable[32];
        enum TARGETID TargetId = Tgt_GetTargetID (szCurrentContext.GetBuffer(0));

        Tgt_GetTargetExecutable (TargetId, WHERE, chExecutable, sizeof (chExecutable));
        szImageToInvoke = chExecutable;

        szCurrentContext += "\\";
        //
        // We need to then format the command line
        szCommandLine.Format("%sEngines\\%sModeller\\%s.exe %s",
                             szHomeDir, szCurrentContext, szImageToInvoke, szArgs);

    }

    // ...and run the image.
    InvokeImage (szCommandLine);
    return 0;

}


void    CECTargetEnvironment::InvokeImage (const CString& ImageName) const
{
    CString     ErrorBuffer;
    CString ImageToInvoke = ImageName;
    STARTUPINFO StartInfo = {sizeof(StartInfo)};
    char   *pszDirectory;

    STARTUPINFO si;
    ZeroMemory(&si,sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
    si.hStdOutput = NULL;
    si.hStdInput  = NULL;
    si.hStdError  = NULL;
    si.wShowWindow = SW_SHOW;

    // Set working directory:
    pszDirectory = NULL;

    // Create the process to execute the command
    PROCESS_INFORMATION pi;
    if (!CreateProcess(NULL,
                       ImageToInvoke.GetBuffer(0), // command line
                       NULL,                // process security attributes
                       NULL,                // primary thread security attributes
                       TRUE,                // handles are inherited
                       CREATE_NEW_CONSOLE,  // creation flags
                       NULL,                // use parent's environment
                       pszDirectory,        // Current directory
                       &si,                 // STARTUPINFO pointer
                       &pi))                // receives PROCESS_INFORMATION
    {
        ErrorBuffer.Format ("Error invoking the image %s" ,ImageToInvoke);
        LPTSTR lpMsgBuf;

        FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                       NULL,
                       GetLastError(),
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                       (LPTSTR) &lpMsgBuf,
                       0,
                       NULL);
        AfxMessageBox (lpMsgBuf);

        //
        //    Free the buffer.
        //
        LocalFree( lpMsgBuf );
    }

}

