/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
///////////////////////////////////////////////////////////////////////////////
// @doc     TestInitiator
// @module  TestInitiator | Wrapper class around processes needed for initiating a TestManager
//              Will Throw
///////////////////////////////////////////////////////////////////////////////
#include "CECNodeDefinition.h"
#include "TestInitiator.h"
#include "Factory/ParamList_i.h"

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor
// @parm    Name of Project where Test can be found
// @parm    Name of Test to be run
// @parm    Node on which test is to be run
// @parm    A CORBAServer/CORBAClient if one available, otherwise class creates a default one
TestInitiator::TestInitiator(const string& ProjectName,const string& TestName,const CECNodeDefinition& node,CORBAServer* CS)
        :m_ProjectName(ProjectName)
        ,m_TestName(TestName)
        ,m_Parent(TestLevel::_nil())
        ,m_ReposNode("")
        ,m_user("")
        ,m_password("")
        ,_Corbaptr(CS)
        ,m_NodeDefinition(node)
        ,m_TestManager(TestManager::_nil())
{}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor

TestInitiator::~TestInitiator()
{
    m_TestManager=TestManager::_nil();
    m_Parent=TestLevel::_nil();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Set the information required so that New test Manager can connect to a repository
// @parm    Node on which respoitory server can be found
// @parm    username
// @parm    password
void TestInitiator::SetRepositoryInformation(const string &ReposNode,const string &user,const string &pass)
{
    m_ReposNode=ReposNode;
    m_user=user;
    m_password=pass;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create an new TestManager and attach to it
// @parm    Parent TestLevel (if any) that this TestManagershould have
// @parm    pipe separated String defining the heirarchy of any parent TestRuns that the
//      new TestManager requires.
//@rdesc    -1 for failure or throws exceptions from methods called
//
long TestInitiator::CreateTestManager   (TestLevel_ptr Parent,string TestRunString/* ="" */)
{
    //
    //    Save the Parent test level and test run string if specified
    //
    m_Parent = Parent;
    m_TestRunString = TestRunString;

    //
    //    ...and construct the command line which we are going to execute.
    //
    char* c_tlevel=getenv(TP_TRACE_LEVEL);
    string tlevel="";
    if(c_tlevel!=NULL)
    {
        tlevel="-v "+string(c_tlevel);
    }
    else
    {
        tlevel="-v 20";
    }
    string params="-r \"%s.Tests.%s\" ";
    params+=tlevel;
    char ImageName[128];
    sprintf(ImageName,params.c_str(), m_ProjectName.c_str(), m_TestName.c_str());
    string Params = ImageName;
    char TestManagerName[128];
    sprintf (TestManagerName
             ,"%s.%s.Tests.%s"
             ,m_NodeDefinition.GetNodeIPAddrAsString().c_str()
             ,m_ProjectName.c_str()
             ,m_TestName.c_str());
    string ExpectedName = TestManagerName;
    //
    //    ...and spawn a process to invoke this command
    //
    CECSpawnObject SpawnObject (m_NodeDefinition
                                ,"Engines"
                                ,TMANAGER
                                ,TARGET_UNKNOWN
                                ,Params
                                ,TestManager::CLASS_NAME
                                ,ExpectedName
                                ,_Corbaptr);
    CORBA::Object_ptr obj = SpawnObject.RetrieveObject();
    m_TestManager = narrow_object((TestManager*)NULL ,obj);
    return 0;//success if no exceptions
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Execute a test

long  TestInitiator::ExecuteTest()
{
    try
    {
        CORBA::String_var cb_ReposNode=CORBA::string_dup(m_ReposNode.c_str());
        CORBA::String_var cb_user=CORBA::string_dup(m_user.c_str());
        CORBA::String_var cb_password=CORBA::string_dup(m_password.c_str());
        CORBA::String_var cb_TestRunString=CORBA::string_dup(m_TestRunString.c_str());
        CORBA::String_var cb_ProjectName=CORBA::string_dup(m_ProjectName.c_str());
        CORBA::String_var cb_TestName=CORBA::string_dup(m_TestName.c_str());
        m_TestManager->Initialise(cb_ReposNode.in(),cb_user.in(),cb_password.in(),cb_TestRunString.in());
        m_TestManager->ExecuteThread(cb_ProjectName.in(),cb_TestName.in(),m_Parent);
    }
    catch(CORBA::Exception&e)
    {
        string err=CorbaExceptionString(e);
        throw CyrException(err.c_str(),"TestInitiator::ExecuteTest",-1);
    }
    return 0;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Get The TestManager Object Created by the TestInitiator
TestManager_ptr TestInitiator::GetTestManager()
{
    return TestManager::_duplicate(m_TestManager);
}
