/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// Local includes
#include "AtvfInput.h"
#include "mdbexcept.h"

/**
 * Constructor
 **/
AtvfInput::AtvfInput(const char* fname, const bool ignore_comments)
        :   _ifP(NULL),
        _inP(NULL),
        _ignoreComments(ignore_comments)
{
    open(fname);
    _strbuf.reserve(BUFFLEN);
}

/**
 * Constructor
 **/
AtvfInput::AtvfInput(istream* inP, const bool ignore_comments)
        :   _ifP(NULL),
        _inP(inP),
        _ignoreComments(ignore_comments)
{
    _strbuf.reserve(BUFFLEN);
}

/**
 * Destructor
 **/
AtvfInput::~AtvfInput()
{
    if(_ifP != NULL)
    {
        _ifP->close();
        delete _ifP;
    }
}

/**
 * Find out if there is more information on a line to get.
 * Is true if started to read a line but not finished.
 **/
bool AtvfInput::moreOnLine() const
{
    return _moreOnLine;
}

/**
 * A sanity check to make sure the input stream within
 * AtvfInput is in a good state.
 **/
bool AtvfInput::ready() const
{
    return _inP->good();
}

/**
 * Check to see if we have come to the end of the input stream
 **/
bool AtvfInput::eof() const
{
    return _inP->eof();
}

/**
 * Open a file to use as input
 **/
void AtvfInput::open(const char* fname)
{
    if(_ifP != NULL)
    {
        _ifP->close();
        delete _ifP;
        _ifP = NULL;
    }
    _ifP = new ifstream(fname);
    if((_ifP == NULL) || !_ifP->good())
    {
        delete _ifP;
        _ifP = NULL;
        throw mdbIOException("Problem opening file", fname);
    }
    _inP = _ifP;
    return;
}

/**
 * Get a lines worth of information and return in a string
 **/
string* AtvfInput::getLine()
{
    _moreOnLine = false;

    // get a line from the input stream
    if(!_inP->getline(buff, BUFFLEN))
    {
        return NULL;
    }

    char* cP = buff;

    for (cP = buff; *cP != '\0'; cP++)
        // If _ignoreComments is on, and "//" is found
        // replace first '/' with'\0' and exit loop early. Hence causing the comment
        // to be removed from the returned string.
        if ( (_ignoreComments) && (*cP == '/') && (*(cP+1) == '/' ) )
        {
            // Set current position to NULL
            *cP = '\0';
            break;
        }

    // remove any trailing <CR>, on Windows will already have been done
    if((cP != buff) && (*(--cP) == '\r')) *cP = '\0';

    // return the contents of the buffer as a string
    return new string(buff);
}

/**
 * Remove a token from the input stream
 **/
string* AtvfInput::getToken()
{
    string data;
    char chr, tempChar;
    bool started = false;
    bool prevSlash = false;

    // Make token string all chars up to whitespace at start of input.
    // Whitespace at beginning of input doesn't terminate token.
    // If _ignoreComments is '//' will terminate a token.
    while(_inP->get(chr))
    {

        // First check for comment
        if ( _ignoreComments && ( chr == '/' ))
        {
            started = true; // In case at the start of the line
            if (prevSlash)
            {
                prevSlash = false;
                delete getLine(); // Trash rest of line
                break; // end of token
            }

            else
            {
                prevSlash = true;
                tempChar = chr;
                continue;
            }

        }

        if(started)
        { // If already have some non-whitespace
            if((chr == ' ') || (chr == '\t'))
            { // found whitespace
                _inP->putback(chr); // put whitespace back on input stream
                _moreOnLine = true;
                break; // end of token
            }

            else if (prevSlash)
            {
                data += tempChar;
                prevSlash = false;
            }
        }
        else if((chr != ' ') && (chr != '\t'))
        { // found non-whitespace
            started = true;
        }

        if(chr == '\n')
        { // EOL
            _moreOnLine = false;
            // remove any trailing <CR>
            if(data[data.length()-1] == '\r')
            {
                data = data.substr(0, data.length()-1);
            }
            break; // end of token
        }

        // Add the character to the token

        data += chr;
    }

    if (prevSlash) // In case last character in token is '/'
        data+=tempChar;

    // This effectively announces the end of file after the last record
    if((data.length() == 0) && !_inP->good())
    {
        return NULL;
    }

    // return with token we have collected
    return new string(data);
}

/**
 * Helper function to perform real getMatch(). If possible
 * return true if can, false otherwise.
 **/
bool AtvfInput::getMatchHelper(const string& match )
{
    char chr;

    // When an ODL contains comments discard leading white space.
    if (_ignoreComments)
    {
        do
        {
            if(!_inP->get(chr))
            {
                if(!_inP->good())
                    _inP->clear();
                // If problem with _inP, or no more character.
                return false;
            }
        }
        while ( (chr == ' ') || (chr == '\n') || (chr == '\t') || (chr == '\r') );
        // Put non-white space back in stream for next loop.
        _inP->putback(chr);
    }

    // read from input stream char by char
    for(unsigned int i = 0; i < match.length(); i++)
    {
        if((!_inP->get(chr)) || (chr != match[i]))
        {
            if(!_inP->good())
            {
                _inP->clear();
            }
            else
            {
                _inP->putback(chr);
            }
            while(i > 0) _inP->putback(match[--i]);
            return false;
        }
    }

    // successfully removed the match string from input
    return true;
}

/**
 * Remove matching string from input stream if possible
 * return true if can, false otherwise.
 * Note, real function is performed by getMatchHelper().
 * function has been split like this so that leading comments
 * may be discarded.
 **/
bool AtvfInput::getMatch(const string& match)
{

    // If _ignoreComments is on, discard leading comment lines
    while (_ignoreComments && getMatchHelper("//"))
        delete getLine();

    // Perform real get match
    return getMatchHelper(match);
}

/**
 * Remove from input stream until find the delimiting string.
 * the delimiting string and anything else until the end of
 * line is discarded.
 **/
string* AtvfInput::getUntil(const string& delim)
{
    _strbuf.reserve(_strbuf.capacity()); // Make sure we don't reduce capacity
    _strbuf.resize(0);

    do
    {
        _inP->get(buff, BUFFLEN, delim[0]);
        if(_inP->eof())
        { // Reached end of the file
            _strbuf.append(buff);
            break;
        }
        else if(_inP->fail() && (buff[0] == '\0'))
        { // Didn't read anything in.
            _inP->clear();
        }
        else
        {
            _strbuf.append(buff);
        }

        string rdDelim;
        do
        {
            _inP->get(buff, delim.size()-rdDelim.size()+1);
            rdDelim.append(buff);

            // do we have the delimiter
            if(rdDelim == delim)
            {
                delete getLine(); // Trash rest of line
                // return the result
                return new string(_strbuf.data(), _strbuf.size());
            }

            // Is this delimiter or any of it, worthwhile
            while((rdDelim.size() > 0) && (delim.find(rdDelim) != 0))
            {
                _strbuf.append(1, rdDelim[0]);
                rdDelim.erase(0, 1);
            }
        }
        while(_inP->good() && (rdDelim.size() > 0));

        if(_inP->fail())
        { // Didn't read anything in last time.
            if(rdDelim.size() > 0) _strbuf.append(rdDelim);
            _inP->clear();
        }
    }
    while(_inP->good());

    // Haven't found delimiter, but reached end of input
    return new string(_strbuf.data(), _strbuf.size());
}
