/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TRecCreator.cpp
/**
 * Define the Methods for the TRecCreator class
 **/

#ifdef WIN32
#pragma warning(disable:4786)
#endif

// Local includes
#include "TRecCreator.h"
/**
 * Constructor
 **/
TRecCreator::TRecCreator()
        :   _typRuleP(NULL)
{}

/**
 * Destructor
 **/
TRecCreator::~TRecCreator()
{
    delete _typRuleP;
    for(prulep_vect_t::const_iterator i = _params.begin();
            i != _params.end();
            i++)
    {
        delete *i;
    }
    _params.erase(_params.begin(), _params.end());
}

/**
 * Clear creator definition prior to the re-definition of creator.
 **/
void TRecCreator::clearDefinition()
{

    for(prulep_vect_t::const_iterator i = _params.begin();
            i != _params.end();
            i++)
    {
        delete *i;
    }
    _params.erase(_params.begin(), _params.end());

    // Delete the parameter rule for the record identifying token, and
    // set to NULL for safety.
    delete _typRuleP;
    _typRuleP = NULL;
}

/**
 * Accessor to gain access to an option of the Record
 * that this creator would make
 **/
string TRecCreator::option(const string& key) const
{
    return ((_typRuleP != NULL)?_typRuleP->options():_options).value(key);
}

/**
 * Accessor to report whether the record this creator would
 * make has a particular attribute.
 **/
bool TRecCreator::hasAttr(const string& attr) const
{
    return (_typRuleP != NULL) ? _typRuleP->hasAttr(attr) : false;
}

/**
 * Method to find specific parameter rule. 
 **/
const TParamRule* TRecCreator::getParamRule(const string& ruleName) const
{
    prulep_vect_t::const_iterator rule;
    for(rule = _params.begin();
            ((rule != _params.end()) &&
             (ruleName.compare((*rule)->token()) != 0));
            rule ++);
    return (rule != _params.end()) ? *rule :
#ifndef __VAX
           static_cast<TParamRule*>(NULL);
#else
(TParamRule*)(NULL);
#endif
}

/**
 * Mutator to set up the record type parameter rule 
 **/
void TRecCreator::addTypRule(TParamRule* prP)
{
    if (_typRuleP)
        delete _typRuleP;
    _typRuleP = prP;
    return;
}

/**
 * Mutator to give another parameter rule 
 **/
void TRecCreator::addPrmRule(TParamRule* prP)
{
    _params.push_back(prP);
    return;
}

/**
 * Mutator to set an option of the record type
 **/
void TRecCreator::addOption(const string& text)
{
    if(_typRuleP != NULL)
    {
        _typRuleP->addOption(text);
    }
    else
    {
        _options.add(text);
    }
    return;
}

/**
 * Mutator to set default value for this record type
 * doesn't mean anything if we haven't got a type level
 * parameter rule so we ignore.
 **/
void TRecCreator::addDefVal(const string& text)
{
    if(_typRuleP != NULL) _typRuleP->defStrVal(text);
    return;
}

/**
 * Mutator to set attributes for this record type
 * doesn't mean anything if we haven't got a type level
 * parameter rule so we ignore.
 **/
void TRecCreator::setAttr(const string& attr)
{
    if(_typRuleP != NULL) _typRuleP->setAttr(attr);
    return;
}

/**
 * Friend operator <<
 **/
ostream& operator<< ( ostream& os, const TRecCreator& tR)
{
    if (tR._typRuleP != NULL ) // In header _typRuleP is set to null
        os << "RECORD : " << *(tR._typRuleP) << endl;
    else
        os <<  tR._options << endl; // Options at this level for header

    os << endl << "Parameters:";
    if(!tR._params.empty())
    {
        for (prulep_vect_t::const_iterator paramIt1 = tR._params.begin();
                paramIt1 != tR._params.end();
                ++paramIt1)
        {
            if (paramIt1 != tR._params.begin())
                os << ",";
            os << " " << (*paramIt1)->token();
        }

        os << endl << "-----------";
        for (prulep_vect_t::const_iterator paramIt = tR._params.begin();
                paramIt != tR._params.end();
                ++paramIt)
        {
            if (paramIt != tR._params.begin())
                os << endl << "-";
            os << endl << *(*paramIt);
        }
    }
    else
        os <<" <NONE>" << endl << "-----------";

    return os;
}
