/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TScript.cpp
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include "TPLog.h"
#include "TContext.h"
#include "CECTofTxt.h"
// local includes
#include "TScript.h"

#include "mdbexcept.h"

#include "mdbparse.h"

const string TScript::_current_ver("6");
const string LabelKeySeparator="|";

/**
 * Default constructor, gives us empty Transaction Script
 **/
TScript::TScript(const TScriptRule* ruleP)
        :   _ruleP(ruleP)
        ,m_have_valid_testcase(false)
        ,m_cur_subroutine("")
{
    _headP = ruleP->newHead();
    _headP->setVer(new string(_current_ver));
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Consructor. Construct a transaction script object from a TOF file
//      and the associated rules.
//
// @parm    Pointer to TScriptRule defining the rules for this TOF file.
// @parm    Reference to ATVF format input stream.
//
TScript::TScript(const TScriptRule* ruleP, AtvfInput &in) :
        _ruleP(ruleP), m_have_valid_testcase(false), m_cur_subroutine("")
{
    _headP = _ruleP->extractHead(in);
    // Invalid tof file
    if(_headP == NULL)
        throw mdbParseException("Tof file format problem");
    // Loop through the records in the TOF file adding records to the Declarations
    // List (when declaration records are found) and Transaction Records List (all
    // other records).
    TRecord* recP;
    while ((recP = _ruleP->extractRecord(in)) != NULL)
    {
        if (recP->isDeclaration())
        {
            _decls.insert(recP);
        }
        else
        {
            add(recP);
        }
    }

    // Iterate through the Transaction Record List invoking the prepare method
    // for each record that has one. We do not do this in the above loop because
    // some of the prepare methods require data from the Declarations List that is
    // not completely set up until after all TOF records have been processed.
    m_endtestcase_list.clear();
    unsigned int scriptend = length();
    for (unsigned int isub = 0; isub < scriptend; isub++)
    {
        int i=rec_at_index(isub)->prepare(this, isub);
#ifdef CYRANO_TP_STANDALONE
        if(i == -1)
            throw mdbParseException("Tof file format problem");
#endif
    }
    return;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor.
//
TScript::~TScript()
{
    declarations(trecset_t());
    delete _headP;

    // Loop through the map of subroutine names/formal parameter details (TtValList*)
    // for this script deleting each formal parameter list object.
    for (subrtn_map_t::const_iterator subItr = _subroutines.begin();
            subItr != _subroutines.end();
            subItr++)
    {
        delete (*subItr).second;    // Delete formal parameters object
    }


    return;
}

/**
 * Accessor to the header, returned simplified as a transaction record
 **/
TRecord* TScript::header() const
{
    return _headP;
}

/**
 * Accessor to the options associated with this type of script
 **/
const TOptions& TScript::options() const
{
    return _headP->options();
}

/**
 * Accessor to the declarations associated with this script
 **/
const trecset_t& TScript::declarations() const
{
    return _decls;
}

/**
 * Accessor to save the transaction script out, to the provided stream,
 **/
void TScript::save(ostream& out) const
{
    _ruleP->saveHead(_headP->version(), out);
    _headP->save(out);
    _headP->saveDelim(out);
    TRecHolder::save(out);
    for(trecset_t::const_iterator i = _decls.begin(); i != _decls.end(); i++)
    {
        (*i)->saveDelim(out);
        (*i)->save(out);
    }

}

/**
 * Accessor to run through and perform all the prepare methods
 * of the records in this script.
 **/
int TScript::prepare(bool verbose)
{
    // Clean out any old labels, ensures recomputed indexes etc.
    _labels.erase(_labels.begin(), _labels.end());

    // Clean out any left over open objects
    while (_openObjects.empty() == false)
        _openObjects.pop();

    // Get record count for script. This must be done at least once
    // after file is loaded to initialize record holder iterator array.
    unsigned int scriptlen = length();
    int res = 0; // result code from recorc execution

    // loop through all our valid TRecords 'preparing'
    for(unsigned int i = 0; i < scriptlen; i++)
    {
        // Note: Codes returned by execute methods:-
        //  0 = Ok
        // -1 = Abort
        if( (res = rec_at_index(i)->prepare(this, i, verbose)) == -1 )
        {
            return -1;
        }
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Get pointer to specified variable's definition record (a TRecord).
//
// @rdesc   Pointer to variable definition TRecord; 'NULL' if variable not
//      defined.
// @parm    Name of variable for which definition record required.
//
TRecord* TScript::GetVardef(const string& varnam) const
{
    // Search the Declarations List for the specified variable declaration/definition.
    string strtkn;
    for (trecset_t::const_iterator it = _decls.begin();
            it != _decls.end();
            it++)
    {
        strtkn = (*it)->token();
        string_toupper(strtkn);
        if (strtkn == DECL_VARDEF)
        {
            const TParam* prm_ptr = (*it)->param("name");
            if (prm_ptr != NULL)
            {
                string strnam = prm_ptr->strVal();
                if (strnam == varnam) return *it; // Return pointer to variable's TRecord
            }

        }
    }

    // Here if variable not found.
    return NULL;
} // End GetVardef()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Add an entry to the subroutine name/parameters map.
//
// @rdesc   Zero (0) always - success.
// @parm    Subroutine name.
// @parm    Pointer to value list object describing subroutine parameters.
//
int TScript::AddSubroutineName(string subNam, TtValList* prmListP)
{
    _subroutines.insert(subrtn_map_t::value_type(subNam, prmListP));
    return 0;
}

/**
 * Accessor to run through and perform all the execute methods
 * of the records in this script.
 **/
int TScript::run(TContext& ctxt)
{
    // fill out the context with this scripts variable definitions
    ctxt.declarations(_decls);

    // Clean out the OpenObjects stack
    while (!ctxt.OpenObjects().empty()) ctxt.OpenObjects().pop();

    ctxt.ClearMask(~CTXT_DEFAULT);

    // Get record count for script. This must be done at least once
    // after file is loaded to initialize record holder iterator array.
    unsigned int scriptlen = length();
    int res = 0; // result code from record execution

    // loop through all our valid TRecords 'executing'
    for(unsigned int i = 0; i < scriptlen; i++)
    {
        if(ctxt.is_aborting())
            return 0;
        // Set current object (record) index within the script, used
        // for logging.
        ctxt.set_oi(i);
        // Note: Codes returned by execute methods:-
        // >0 = record number to continue execution at
        //  0 = Ok
        // -1 = Abort
        // -2 = Retry (not yet implemented)
        // -3 = Skip to next record (error has occurred)
        // -4 = End-of-Executable-Records, remaining records are Procedure definitions
        logr->AddTrace(SCRIPTCOMMAND,DEBUGLEVEL,rec_at_index(i)->token().c_str());

        try
        {
            res = rec_at_index(i)->execute(ctxt);
        }
        catch(DetachThreadException)//if thread detatching pass on up to top
        {
            throw;
        }
        catch(...)
        {
            //Catch any Exceptions and check on error goto statement
            //if no on error goto re-throw to be caught later
            int gotoindex=ctxt.GetErrorLabel();
            if(gotoindex==0)//no on error set
            {
                string msg="Error While Running Script, passing control to parent script";
                ctxt.log(msg,AUDITMSG);
                throw;
            }
            else
            {
                string msg="Error While Running Script, Jumping to record "+int_to_string(gotoindex);
                ctxt.log(msg,AUDITMSG);
                res=++gotoindex;
            }
            ctxt.CancelOnError();//have errored so turn off on error goto for this script
        }

        if(res == -1 )
        {
            // Remove any remaining commands/connections
            ctxt.cleanup(true);
            return -1;
        }
        else if ( res > 0 ) // Goto (res+1)
        {
            // Valid destination ?
            if (res <= scriptlen)
            {
                // Set next record index, adjust to 0 based, and allow for i++
                i = res-2;
            }
            else
            {
                // Destination beyond end of script!, Error.
                TRecord* recrd = rec_at_index(i);

                string msg("Flow Control error, bad destination: " +
                           int_to_string(res) + " from '" + recrd->token() +
                           "' at position: " + int_to_string(i));

                for (int pix = 0; pix < recrd->numParams(); pix++)
                {
                    const TParam* param = recrd->param(pix);
                    msg += "\n" + param->token() + "(" + param->strVal() + ")";
                }
                ctxt.log(msg, TRACEMSG);

                // Remove any remaining commands/connections
                ctxt.cleanup(true);
                return -1;
            }
        }
        else if (res == -4)
        {
            // End of excutable records reached, exit with success
            break;
        }
    }

    // Invalidate index when script done.
    ctxt.set_oi(-1);
    return 0;
}

/**
 * Method to replace the set of declarations associated with this 
 * script
 * Any TRecords effectively passed in through the set are seen to
 * be being given to this object and it must delete them later
 **/
void TScript::declarations(const trecset_t& decls)
{
    // First delete all the stuff in the old declaration set
    for(trecset_t::const_iterator i = _decls.begin();
            i != _decls.end();
            i++)
    {
        delete *i;
    }
    _decls.erase(_decls.begin(), _decls.end());

    // copy contents of given set into our set.
    for(trecset_t::const_iterator j = decls.begin();
            j != decls.end();
            j++)
    {
        _decls.insert(*j);
    }
    return;
}

////////////////////////////////////////////////////////////
// @mfunc   Add a label to the script to retrieve index later
//      the label is stored along with the current subroutine name
//      to allow for the duplication of label names
// @parm    label name
// @parm    index associated with label
void TScript::AddLabel(const string& inlabel, int recIndex)
{
    //set up label key (incluing current subroutine name)
    string label=inlabel+LabelKeySeparator+m_cur_subroutine;
    label_map_t::iterator iter;
    // If already an existing label for key, remove it.
    iter = _labels.find(label);
    if(iter != _labels.end())
    {
        _labels.erase(iter);
    }
    // Insert new label
    _labels[label] = recIndex;
}


int TScript::GetLabelIndex(const string& label,const string& subroutine) const
{
    string labelkey=label+LabelKeySeparator+subroutine;
    label_map_t::const_iterator iter = _labels.find(labelkey);
    if(iter == _labels.end()) return -1;
    return (*iter).second;
}
////////////////////////////////////////////////////////////
// @mfunc   Add a subroutine index to script
//      sets up the current subroutine name for associating with labels
// @parm    subroutine name
// @parm    index associated with subroutine
void TScript::AddSubroutineLabel(const string& label, int recIndex)
{
    m_cur_subroutine=label;//used in prepare as running through each record
    label_map_t::iterator iter;
    // If already an existing label for key, remove it.
    iter = _sublabels.find(label);
    if(iter != _sublabels.end())
    {
        _sublabels.erase(iter);
    }
    // Insert new label
    _sublabels[label] = recIndex;
}


int TScript::GetSubroutineLabelIndex(const string& label) const
{
    label_map_t::const_iterator iter = _sublabels.find(label);
    if(iter == _sublabels.end()) return -1;
    return (*iter).second;
}
void TScript::AddEndTestCase(int index)
{
    m_endtestcase_list.push_back(index);
    m_have_valid_testcase=false;
}
int TScript::GetNextEndTestCaseIndex(int _cur_index)
{
    int objind=-1;//fail
    vector<int>::iterator i;
    vector<int>::size_type sz=m_endtestcase_list.size();
    if(sz==0)return objind;
    for(i=m_endtestcase_list.begin();
            i!=m_endtestcase_list.end();
            i++)
    {
        if(_cur_index < *i)//if the current index is before the next end test case statement
        {
            objind=*i;
            break;
        }
    }
    return objind;//return object index within script for next end test case or -1 for failure to find one
}
