/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// ScriptCache.cpp
/**
 * Method definitions for the ScriptCache class
 **/

// local includes
#include "ScriptCache.h"

/**
 * Default constructor, create empty Script Cache.
 **/
ScriptCache::ScriptCache(TScriptRule* rule)
        :   _rule(*rule),
        _numScripts(10), // at least 10 scripts in cache
        _cacheSize(-1), // don't care about size of cache
        _sizeWghtng(-1)
{ // smallest get deleted first
}

/**
 * Destructor
 **/
ScriptCache::~ScriptCache()
{
    for(script_map_t::iterator i = _scripts.begin();
            i != _scripts.end();
            i++)
    {
        delete (*i).second;
    }
    _scripts.erase(_scripts.begin(), _scripts.end());
}

/**
 * Accessor to find out if we're using more than the amount of
 * space than would be desirable. 
 * Whole load of tuning params to check and more can be added
 **/
bool ScriptCache::spaceNeeded() const
{
    // This cleans up when we've got more than set num of scripts in cache
    if((_numScripts >= 0) &&
#ifndef __VAX
            (_scripts.size() > static_cast<script_map_t::size_type>(_numScripts)))
    {
#else
(_scripts.size() > (script_map_t::size_type)(_numScripts)))
            {
#endif
        return true;
    }

    // This cleans up when cache reached certain size
    if(_cacheSize >= 0)
    {
        if(_cacheSize == 0) return true; // Always cleanup
        int size = 0;
        for(script_map_t::const_iterator i = _scripts.begin();
                i != _scripts.end();
                i++) size += ((*i).second)->size();
        if(size > _cacheSize) return true; // cleanup if over set size
    }


    return false; // If we get here we must have enough space
}

/**
 * Mutator to perform cache cleanup,
 * If parameter is set to true do conditional cleanup, only cleaning up if
 * space is running short.
 **/
void ScriptCache::cleanup(bool lazy)
{
    while((TaskScript::someUnused()) && (!lazy || spaceNeeded()))
    {
        int chosenFact;
        script_map_t::iterator chosen = _scripts.end();
        for(script_map_t::iterator i = _scripts.begin(); i != _scripts.end(); i++)
        {
            if(!((*i).second->inUse()))
            {
                int fact = ((*i).second)->size() * _sizeWghtng;
                if((chosen == _scripts.end()) || (chosenFact < fact))
                {
                    chosen = i;
                    chosenFact = fact;
                }
            }
        }
        if(chosen != _scripts.end())
        {
            delete (*chosen).second;
            _scripts.erase(chosen);
        }
        else
        {   // Didn't find one (?), don't want to get into infinite loop
            break;
        }
    }
}

/**
 * Mutator to create a task loading a script or using already cached one.
 **/
ThreadTask* ScriptCache::makeTask(const string& scriptname)
{
    TaskScript* scriptP;
    script_map_t::const_iterator i = _scripts.find(scriptname);
    if(i == _scripts.end())
    {   // Script needs loading

        cleanup(); // Perform cache cleanup to save space before loading

        scriptP = new TaskScript(&_rule, scriptname); // Load

        _scripts[scriptname] = scriptP;

    }
    else
    {   // Script already loaded
        scriptP = ((*i).second);
        scriptP->addUse();
    }
    // Build thread specific task around this script
    ThreadTask* taskP = new ThreadTask(scriptP);
    // If fail to allocate task, release the script.
    if(taskP == NULL)
        scriptP->release();
    return taskP;
}
/*
 * Check to see if task already read
 */
bool ScriptCache::doesTaskExist(const string& scriptname) const
{
    script_map_t::const_iterator i = _scripts.find(scriptname);
    if(i == _scripts.end())
    {
        return false;
    }
    else
    {
        return true;
    }
}
/**
 * Mutator to set the parameters that control how the cache behaves
 **/
void ScriptCache::setParams(int numScripts, int cacheSize, int sizeWghtng)
{
    _numScripts = numScripts;
    _cacheSize = cacheSize;
    _sizeWghtng = sizeWghtng;
}

/**
 * Define operator << to allow ScriptCache to be output.
 **/
ostream& operator<<(ostream& os, const ScriptCache& cache)
{
    os << "Script Cache: " << cache._scripts.size() << " entries, ";
    if(cache._scripts.size() > 0 )
    {
        os << (TaskScript::someUnused() ? "some" : "none") <<
        " available for cleanup";
    }
    os << endl << " params: scripts=" << cache._numScripts <<
    " size=" << cache._cacheSize <<
    " size-weighting=" << cache._sizeWghtng;
    for(script_map_t::const_iterator i = cache._scripts.begin();
            i != cache._scripts.end();
            i++)
    {
        os << endl << " `" << (*i).first << "` " << *((*i).second);
    }

    return os;
}

/**
 * Define operator << to allow ScriptCache to be output to a ostrstream
 **/
ostrstream& operator<<(ostrstream& os, const ScriptCache& cache)
{
    os << "Script Cache: " << cache._scripts.size() << " entries, ";
    if(cache._scripts.size() > 0 )
    {
        os << (TaskScript::someUnused() ? "some" : "none") <<
        " available for cleanup";
    }
    os << "\r\n params: scripts=" << cache._numScripts <<
    " size=" << cache._cacheSize <<
    " size-weighting=" << cache._sizeWghtng;
    for(script_map_t::const_iterator i = cache._scripts.begin();
            i != cache._scripts.end();
            i++)
    {
        os << "\r\n `" << (*i).first << "` " << *((*i).second);
    }

    return os;
}

