/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TContext.h
#ifndef _CYR_TCONTEXT_H
#define _CYR_TCONTEXT_H
/**
 * Class Definition of an object that holds the context within which
 * a TScript executes.  Thus giving the ability to maintain some
 * state between the running of different TScripts.
 **/

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include <string>
#include <map>
#include <set>
#include <stack>
#include <list>

#include "odl.h"

#ifdef unix
#elif defined (__VMS)
#elif defined (WIN32)
#pragma warning(disable:4786)
using namespace std;
#else //unknown
#error Unknown Platform
#endif // unknown


// Local includes
#ifndef CYRANO_TP_STANDALONE
#include "CECMutexList.h"
#endif
#include "CECFileList.h"
#include "CECTofTxt.h"
#include "TConnect.h"
#include "TCommand.h"
#include "TScript.h"
#include "mdbstatus.h"
#include "varlist.h"
#include "cxxutils.h"
#include "mdbstl.h"
#include "mdbexcept.h"
#include "handler.h"
#include "CECTimer.h"
#include "tofprmdef.h"
// Forward declarations
class TEnvironment;

enum LogTypes
{/*These types match those found in the odl file for Logmsg*/
    AUDITMSG=0, HISTORYMSG=1, REPORTMSG=2, NOTEMSG=3, TRACEMSG=4,
    /*These next types are not for user use*/
    WRNMSG=5,ERRMSG=6};

// Logging levels
#define LOGDFLT AUDITMSG

// Numeric list of possible wait conditions
#define CW_NONE     0           // Not waiting for anything
#define CW_FILE     1           // Waiting for a file name
#define CW_VAR      2           // Waiting for a variable value
#define CW_ACK      3           // Waiting for acknowledge message

// Execution flags
#define CTXT_DEFAULT        0
#define CTXT_SKIP_ELSE      1
#define CTXT_OBJ_RESULTS    2


#define DEFAULT_THREAD_MAX 2000


typedef map<unsigned int, TConnect*, less<unsigned int> > cons_map_t;
typedef map<unsigned int, TCommand*, less<unsigned int> > cmds_map_t;
typedef map<int, Chandler*, less<int> >                   hnds_map_t;

//
//    Timer map. This retains knowledge of timers which have been activated
//
typedef map <string, CECTimer*, less<string> > TimerMap_t;
//
static const char * emptystr = "";

enum LOADTYPE   {LD_ACTIVETHRDS=0, LD_DATE, LD_DIRECTORY, LD_ERROR, LD_NODENAME,
                 LD_SCRIPT, LD_TEST, LD_THREAD, LD_TIME, LD_ACCOUNTING ,LD_TIMER};

// Exception class to use for Detatch thread
class DetachThreadException:public mdbRuntimeError
{
public:
DetachThreadException():mdbRuntimeError("Detaching Thread")
    {}
    ;
    virtual ~DetachThreadException()
    {}
    ;
};
class TContext
{
    //Keep these 3 accross all scripts within thread!
    cons_map_t _cons;
    cmds_map_t _cmds;
    hnds_map_t _hnds;
    TEnvironment* _environment;
#ifndef CYRANO_TP_STANDALONE
    CECMutexList m_MutexList;//list of mutexes currently used by this thread
    //(keep thread wide not script wide)
#endif//#ifndef CYRANO_TP_STANDALONE
public:
    // Constructors & Destructor
    TContext();
    virtual ~TContext();

    TScript*        Script() const
    {
        return m_ContextDataStack.top()->_script;
    }
    TEnvironment*   Environment() const
    {
        return _environment;
    }
    rcrd_stack_t&   OpenObjects()
    {
        return m_ContextDataStack.top()->_openObjects;
    }

    TConnect*               connection(unsigned int conid) const;
    TCommand*               command(unsigned int cmdid) const;

    TCommand*           getCommandAtInd(unsigned int ind) const;
    void                    log(const string& mesg, int level = LOGDFLT);
    void                    delay(unsigned int ms);
    bool            get_ObjResults () const
    {
        return objResults;
    }
    int                     get_status(void)
    {
        return status;
    }
    int                     get_state(void)
    {
        return state;
    }
    int                     get_phase(void)
    {
        return m_ContextDataStack.top()->phase;
    }
    const char *            get_phase_text(void)
    {
        if(m_ContextDataStack.top()->phasetxt != NULL)
            return m_ContextDataStack.top()->phasetxt;
        else
            return emptystr;
    }
    string                  get_file(void)
    {
        return m_ContextDataStack.top()->file;
    }

    // Override in specific Contexts (i.e. TestManager /TestExecuter) to retrieve name of test (not script) being run
    virtual string      get_test_name(void)
    {
        return "";
    }
    inline string ScriptName()
    {
        return m_ScriptNameStack.top();
    }
    Cvarlist&           get_varlist(void)
    {
        return m_ContextDataStack.top()->varlist;
    }
    int                     get_con_count(void)
    {
        return _cons.size();
    }
    int                     get_cmd_count(void)
    {
        return _cmds.size();
    }
    cons_map_t & get_connections(void)
    {
        return _cons;
    }

    int             GetLabelIndex(const string& label) const;
    int             GetNextEndTestCaseIndex() const;
    bool        MaskIsSet (const unsigned mask) const
    {
        return ((m_ContextDataStack.top()->flags & mask)?true:false);
    }
    const string&   ErrorMessage() const
    {
        return m_ContextDataStack.top()->_errorMessage;
    }
    string      GetCurrentReturnValue()
    {
        return CurrentReturnValue;
    }
    ;//return current return value of thread
    // Mutators
    void        SetScript(TScript* script)
    {
        m_ContextDataStack.top()->_script = script;
    }
    void        SetEnvironment(TEnvironment* envp)
    {
        _environment = envp;
    }
    void                cleanup(bool cflag = false);
    void                addCon(unsigned int conid, TConnect* conP);
    void                addCmd(unsigned int cmdid, TCommand* cmdP);
    void                remCon(unsigned int conid);
    void                remCmd(unsigned int cmdid);
    inline void         set_status(int stat)
    {
        status = stat;
    }
    inline void         set_state(int stat)
    {
        state = stat;
    }
    inline void         set_phase(int ph)
    {
        m_ContextDataStack.top()->phase = ph;
    }
    inline void         set_phase_text(const char * phtxt)
    {
        m_ContextDataStack.top()->phasetxt = phtxt;
    }
    inline void         set_file(string tfile)
    {
        m_ContextDataStack.top()->file = tfile;
    }
    inline void         set_oi(int index)
    {
        m_ContextDataStack.top()->oi = index;
    }
    void                set_verbose(bool tf)
    {
        verbose_mode = tf;
    }
    inline bool         is_aborting(void)
    {
        return abort_flag;
    }
    void        set_ObjResults ( bool flag )
    {
        objResults = flag;
    }
    void        SetMask(const unsigned mask)
    {
        m_ContextDataStack.top()->flags |= mask;
    }
    void        ClearMask(const unsigned mask)
    {
        m_ContextDataStack.top()->flags &= ~mask;
    }
    void        SetErrorMessage(const string& msgStr)
    {
        m_ContextDataStack.top()->_errorMessage = msgStr;
    }
    // Extract declaration (vardef for now) records
    virtual int                 declarations(trecset_t _decls);
    // Add/get event handlers
    void                add_handler(Chandler* hndlr)
    {
        hnds_map_t::iterator it;
        // If already an existing handler for key, remove it.
        it = _hnds.find(hndlr->get_key());
        if(it != _hnds.end())
        {
            delete (*it).second;
            _hnds.erase(it);
        }
        // Insert new handler
        _hnds[hndlr->get_key()] = hndlr;
    }

    const Chandler * get_handler_for(int key)
    {
        hnds_map_t::iterator it;
        // Find handler for key
        it = _hnds.find(key);
        if(it != _hnds.end())
            return (*it).second;
        // None exists
        return NULL;
    }

    hnds_map_t & get_handlers(void)
    {
        return _hnds;
    }

    virtual void        abort(void)
    {
        abort_flag = true;
    }
    // Set delay scaling percentage
    static void     delay_percent(unsigned d)
    {
        timefac = d;
    }
    // Get delay scaling percentage
    static unsigned get_delay_percent(void)
    {
        return timefac;
    }

    virtual void CleanUp()=0;//Remove Corba Object or other Thread type specific data at what is essentially the end of the thread
    virtual void SetNoteText(const string& mesg)=0;//Call thread type specific method for setting note Text

    // Start named timer
    int     StartTimer      (const string& name);
    // Stop named timer
    int     StopTimer       (const string& name);

    //Start/End a new Script method for any Thread type specific functionality
    //Called from TaskScript->run()
    virtual void    StartScript(const string& scriptname)
    {
        return;
    };
    virtual void    EndScript()
    {
        return;
    };

    virtual int ExitCommand(bool keepalive,const string& retvar)
    {
        CurrentReturnValue=retvar;
        return 0;
    };
    // @cmember Set value for a variable
    inline int      set_variable (const string& name,
                                  const string& value,
                                  string index = "")
    {
        // Set value locally
        int vstat = m_ContextDataStack.top()->varlist.set_variable_value(name, value, index);

        // Return set status
        return vstat;
    }

    // @cmember Get value for a variable
    //Chnaged from returning reference as more than 1 thread may have reference
    //at the same time
    inline string   get_variable(const string& name,
                                 bool getcur = false,
                                 string index = "")
    {
        // Need non-const version, calls below may result in
        // uppercasing of contents.
        string vname(name);

        // Get pointer to variable object for "name".
        CECVar* vp = m_ContextDataStack.top()->varlist.get_variable_ptr(vname);

        // If not defined throw execption to advise caller.
        if(vp == NULL)
        {
            string msg="Undefined variable: "+name;
            throw mdbParseException(msg.c_str());
        }
        // Get the variable value
        return vp->get_value(index);
    }
    void Load   (LOADTYPE Type, string ToVarName ,string FromItem);


    // VARIABLE MANIPULATION
    // =====================
    int Format(const string& targetVar, const string& targetIdx, const string& format,
               const TtValList* valListP);
    int FormatIntTarget(const string& targetVar, const string& targetIdx,
                        const string& format, const TtValList* valListP);
    int FormatCharTarget(const string& target, const string& targetIdx,
                         const string& format, const TtValList* valListP);

    // FILE MANIPULATION
    // =================
    int FileOpen(const string& filename,const string& fileid, int access, int options);
    int FileRead(const string& fileid, const TtVarRef* var);
    int FileWrite(const string& fileid,const string& data);
    int FileControl(int operation,const string& fileid);


    // FORMAL TEST CONTROL
    // ===================
    virtual int TCterminate(int result,string desc);//virtual so overriden in TM/TE threads
    //@cmember have we started a test case
    inline bool IsTestCaseStarted()const
    {
        return m_ContextDataStack.top()->m_have_valid_testcase;
    }
    //@cmember set Test Case current description
    inline void TestCaseDesc(string desc)
    {
        m_ContextDataStack.top()->m_cur_testcasedesc=desc;
        m_ContextDataStack.top()->m_have_valid_testcase=true;
    }
    //@cmember get Test Case current description
    inline string TestCaseDesc(void)const
    {
        return m_ContextDataStack.top()->m_cur_testcasedesc;
    }

    // Synchronisation
    // ===============
#ifndef CYRANO_TP_STANDALONE
    virtual bool LockMutex(string name,int scope);
    virtual bool UnLockMutex(string name,int scope);
#endif //#ifndef CYRANO_TP_STANDALONE
    //
    // Error Handling/Termination
    // ==========================
    void SetErrorLabel(int index)
    {
        m_ContextDataStack.top()->on_error_index=index;
    }
    void CancelOnError()
    {
        SetErrorLabel(0);
    }
    int GetErrorLabel()
    {
        return m_ContextDataStack.top()->on_error_index;
    }
    //
    // Subroutine Information
    // ==========================
public:
    class TContextData;
    class SubroutineData
    {
    public:
        SubroutineData(const string& name,int index)
    :SubroutineName(name),ReturnIndex(index)
        {}
        ;
        virtual ~SubroutineData()
        {}
        ;
        int ReturnIndex;
        string SubroutineName;
    };
public:
    inline  void    PushSubroutine(const string& name,int index)
    {
        SubroutineData sub(name,index);
        m_ContextDataStack.top()->m_SubroutineDataStack.push(sub);
    }
    inline  void    PopSubroutine()
    {
        m_ContextDataStack.top()->m_SubroutineDataStack.pop();
    }
    inline SubroutineData   GetCurrentSubroutineData()
    {
        return m_ContextDataStack.top()->m_SubroutineDataStack.top();
    }
    int SubParameters(TScript* script, string subNam, TtValList* prmListP);

    //not to be overridden by correct Thread class to return Identifier for class
    string GetThreadID()
    {
        return m_ThreadID;
    }

protected:

    // @cmember Help function to action the Load Timer command
    string LoadTimer (string FromTimer);

    //protected data

    //This class is Contextual Data that can be different for each script
    //called by a thread, a stack of these is held within the context
    //All access to these data members is done via the top member of the stack
public:

    class TContextData
    {
    public:
    TContextData():on_error_index(0)
        {
            m_SubroutineDataStack.push(TContext::SubroutineData("",0));//NULL subroutine on bottom of stack for start of script
        }
        ;
        virtual ~TContextData()
        {}
        ;
        // @mfunc return top of subroutine data stack
        virtual SubroutineData GetSubroutineData()
        {
            return m_SubroutineDataStack.top();
        }
    public:
        TScript*        _script;        // Ptr to Script being processed by this context.
        int             phase;          // Execution phase
        const char *    phasetxt;       // Execution phase text
        unsigned        flags;          // Execution flags
        string          file;       // Current transaction file
        int             oi;             // Current object (record) index within script
        Cvarlist        varlist;        // Variable list for this script
        CECFileList     filelist;       // File list for this script
        string          _errorMessage;  // Message reported by TRecord::execute methods

        // @cmember Map of active/complete timers - this is script specific and MUST be saved across
        //          script calls
        TimerMap_t      TimerMap;
        // @cmember has explicit EXIT been called on this script flag
        bool    exit_called_flag;
        rcrd_stack_t _openObjects;
        int     on_error_index;//index to go to in this script if error occurs
        stack<SubroutineData>  m_SubroutineDataStack;//stack of Subroutine information
        //@cmember current test case description
        string  m_cur_testcasedesc;
        //@cmember are we currently within a valid test case
        bool    m_have_valid_testcase;

    };
protected:
    stack<TContextData*>    m_ContextDataStack;
    stack<string>       m_ScriptNameStack;//stack of scripts called separate from m_ContextDataStack as needs to be pop()'ed at different time
    int             state;          // Execution state of thread
    int             status;         // Execution status of thread
    bool            abort_flag;     // Indicates abort asap.
    static unsigned timefac;        // Delay times %. 100 = 100% (as recorded),
    bool            verbose_mode;   // Flag set if running in verbose mode (execution tracing)
    bool            objResults;     // Flag set if transaction objects keep intermediate results
    int             intid;          // Internal id (usually thread id)
    string          strval;         // String arg for wake up

    string          CurrentReturnValue;//Current Return value from any script running
    string      m_ThreadID;//ThreadID for current thread


}

; // TContext

#endif // _CYR_TCONTEXT_H

