/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//////////////////////////////////////////////////////////////////////
// TExpression.cpp: implementation of the TExpression class.
//
// Class to evaluate Infix notation arithmetic expressions.
// Constructed with an Infix expression in string form, from which a
// Postfix (Polish) notation expression is built in the form of a token
// list, for easy evaluation.
// Parsed expression is evaluated by calling the ::Evaluate() method,
// which returns its result in string form, or the empty string on error.
// An error message can be obtained by calling the ::Reason() method.
//
//////////////////////////////////////////////////////////////////////

#include "TExpression.h"
#include "mdbparse.h"
#include "mdbexcept.h"
#include "TSegStr.h"
#include "TContext.h"
#include <cmath>


const string TExpression::stFALSE   = "0";
const string TExpression::stTRUE    = "1";


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TExpression::TExpression()
{
    ; // do nothing
}


TExpression::TExpression(const TExpression& trec)
{
    _reason = trec._reason;
    for (list<string>::const_iterator it = trec._postfix.begin(); it != trec._postfix.end(); it++)
        _postfix.push_back(*it);
    for (unsigned i=0; i < trec._varList.size(); i++)
        _varList.push_back(new TVarDetails(*(trec._varList[i])));
}


TExpression::~TExpression()
{
    for(vector<TVarDetails*>::iterator it = _varList.begin();
            it != _varList.end();
            it++) delete *it;

    _varList.erase(_varList.begin(), _varList.end());
}


bool TExpression::SetExpression(const string& inStr, int pos)
{
    static string opList("+-*/^~!()<>=&|@"); // List of operator symbols
    static char stEscape = '\\';

    // Discard previous expression (if any)
    while (!_opStack.empty()) _opStack.pop();
    _postfix.erase(_postfix.begin(), _postfix.end());

    for(vector<TVarDetails*>::iterator it = _varList.begin();
            it != _varList.end();
            it++) delete *it;
    _varList.erase(_varList.begin(), _varList.end());

    _operand = "";
    TSegStr infix(inStr.substr(pos)); // Build segmented string to seperate out variables

    // Build postfix (Polish) notation token list, for easy evaluation
    bool ok;
    TSegStr_iterator iter = infix.begin();
    while (iter != infix.end())
    {
        string rawSeg = *iter++;
        if (!rawSeg.empty())
        {
            // Do we have a variable ?
            if (rawSeg[0] == TSegStr::VAR_DELIM && rawSeg[rawSeg.length()-1] == TSegStr::VAR_DELIM)
            {
                // Append this token (variable) to the postfix string.
                _operand = rawSeg;
                // Save variable details to facilitate evaluation later on.
                TVarDetails* varPtr = new TVarDetails(rawSeg.substr(1,rawSeg.length()-2));
                _varList.push_back(varPtr);
            }
            else
            {
                // Process literal text (numbers and/or operators) ...

                // Build copy of rawSeg without any whitespace chars, as seg
                string seg;
                //Changed for slaris port from boolean to bool
                bool inQuotes = false;
                char quoteChar = '\0';
                pos = 0;
                while (pos < rawSeg.length())
                {
                    char c = rawSeg[pos];
                    if ((inQuotes && (c != stEscape || rawSeg[pos-1] == stEscape)) ||
                            (!inQuotes && Cmdbparse::wordBreak.find_first_of(c) == string::npos))
                        seg += c;

                    if ((c == '"' || c == '\'') &&
                            (!inQuotes || (c == quoteChar && rawSeg[pos-1] != stEscape)))
                    {
                        inQuotes = (inQuotes?false:true);
                        quoteChar = c;
                    }

                    ++pos;
                }

                if (inQuotes) //  Unterminated quoted string?
                {
                    _reason = "Quote mismatch";
                    return false;
                }

                // Build postfix (Polish) notation token list
                ok = true;
                string::iterator endSeg = seg.end();
                for (string::iterator it = seg.begin(); ok && it != endSeg; it++)
                    ok = _ProcessChar(it, endSeg);

                if (!ok) return false; // Error processing expression
            }

        }
    }

    string::iterator endPos = NULL;
    ok = _ProcessChar(endPos, endPos); // Flush stack.
    return ok;
}


// Evaluate the internal expression, postfix form built on construction.
string TExpression::Evaluate(TContext& ctxt) const
{
    static string opList = "+-/*^!<>=&|@"; // List of operator symbols (excluding brackets)
    static string unaryOps = "!";
    static string noResult = "";

    if (_postfix.empty())
    {
        _reason = "No expression to evaluate";
        return noResult;
    }

    // Process pre-compiled postfix (Polish) notation expression
    vector<TVarDetails*>::size_type varNo = 0; // Index into pre-compiled variables list
    stack<Operand*> operands;
    for (list<string>::const_iterator it = _postfix.begin(); it != _postfix.end(); it++)
    {
        string token = *it;
        // Is token an Operand or an Operator ?
        if (token.length() > 2 || opList.find_first_of(token[0]) == string::npos)
        {
            // If Operand is a variable, we need to get a value for it.
            if (token[0] == TSegStr::VAR_DELIM && token[token.length()-1] == TSegStr::VAR_DELIM )
            {
                // Select equivalent variable from pre-compiled list.
                if (varNo >= _varList.size())
                {
                    ctxt.SetErrorMessage("Illegal variable reference: " + token);
                    return noResult;
                }

                TVarDetails* var = _varList[varNo++];

                // Get value for variable
                try
                {
                    token = ctxt.get_variable(var->name(),
                                              var->current(),
                                              var->index());
                }
                catch(mdbParseException&)
                {
                    ctxt.SetErrorMessage("No value available for variable: " + token);
                    return noResult;
                }
            }

            // Strip off quotation marks
            if ((token[0] =='"' && token[token.length()-1] == '"') ||
                    (token[0] =='\'' && token[token.length()-1] == '\''))
                token = token.substr(1, token.length()-2);
            operands.push(new Operand(token));
        }
        else // Token is an Operator
        {
            // Get 1 or 2 Operands depending upon the operator type (unary, binary)
            if (operands.empty()) // Underflow
            {
                ctxt.SetErrorMessage("Stack underflow, insufficient operands for operator");
                return noResult;
            }
            Operand* rhs = operands.top();
            operands.pop();

            Operand* lhs =NULL;
            if (unaryOps.find_first_of(token[0]) == string::npos ||
                    token.length() > 1)
            {
                // Binary Operator, so pop 2nd operand
                if (operands.empty()) // Underflow
                {
                    ctxt.SetErrorMessage("Stack underflow, insufficient operands for binary operator");
                    return noResult;
                }
                lhs = operands.top();
                operands.pop();
            }

            // Operator defined for string operands ?
            bool stringsAllowed = _DefinedForStrings(token);

            // Check for string Operands in arithmetic only operation
            if (!stringsAllowed && (rhs->IsString() || (lhs && lhs->IsString())))
            {
                ctxt.SetErrorMessage("Operator '" + token + "' expects two numeric operands, found: ");
                _reason += ((lhs && lhs->IsString()) ? lhs->StrVal() : string(""));
                if (rhs->IsString()) _reason += " " + rhs->StrVal();
                return noResult;
            }

            // Check BOTH Operands are of same type
            if (stringsAllowed)
            {
                if ( ( lhs->IsString() && !rhs->IsString()) ||
                        (!lhs->IsString() &&  rhs->IsString()) )
                {
                    ctxt.SetErrorMessage("Mismatched Operands, need both Numeric, or both strings ");
                    return noResult;
                }
            }

            // Perform operation on operand(s), and push result back onto stack.
            Operand* temp = NULL;
            switch (token[0])
            {
            case '+':
                operands.push(new Operand(lhs->NumVal() + rhs->NumVal()));
                break;

            case '-':
                operands.push(new Operand(lhs->NumVal() - rhs->NumVal()));
                break;

            case '*':
                operands.push(new Operand(lhs->NumVal() * rhs->NumVal()));
                break;

            case '/':
                if (rhs->NumVal() != 0.0)
                    operands.push(new Operand(lhs->NumVal() / rhs->NumVal()));
                else
                {
                    ctxt.SetErrorMessage("Attempted divide by ZERO");
                    return noResult; // Divide by zero
                }

                break;

            case '@': // Exponentiation
                {
                    operands.push(new Operand(pow(lhs->NumVal(), rhs->NumVal())));
                    int numOpers = operands.size();
                    break;
                }


            case '!':
                if (token.length() > 1 && token[1] == '=') // Inequality operator
                {
                    if (lhs->IsString())
                        temp = new Operand(lhs->StrVal() != rhs->StrVal());
                    else
                        temp = new Operand(lhs->NumVal() != rhs->NumVal());
                }
                else // Logical NOT operator
                    temp = new Operand(!rhs->NumVal());

                operands.push(temp);
                break;

            case '<':
                if (token.length() > 1 && token[1] == '=') // Less than or equal
                {
                    if (lhs->IsString())
                        temp = new Operand(lhs->StrVal() <= rhs->StrVal());
                    else
                        temp = new Operand(lhs->NumVal() <= rhs->NumVal());
                }
                else // Less than
                {
                    if (lhs->IsString())
                        temp = new Operand(lhs->StrVal() < rhs->StrVal());
                    else
                        temp = new Operand(lhs->NumVal() < rhs->NumVal());
                }

                operands.push(temp);
                break;

            case '>':
                if (token.length() > 1 && token[1] == '=') // Greater than or equal
                {
                    if (lhs->IsString())
                        temp = new Operand(lhs->StrVal() >= rhs->StrVal());
                    else
                        temp = new Operand(lhs->NumVal() >= rhs->NumVal());
                }
                else // Greater than
                {
                    if (lhs->IsString())
                        temp = new Operand(lhs->StrVal() > rhs->StrVal());
                    else
                        temp = new Operand(lhs->NumVal() > rhs->NumVal());
                }

                operands.push(temp);
                break;

            case '=':
                if (lhs->IsString())
                    temp = new Operand(lhs->StrVal() == rhs->StrVal());
                else
                    temp = new Operand(lhs->NumVal() == rhs->NumVal());
                operands.push(temp);
                break;

            case '^': // Contains
                if (lhs->IsString())
                {
                    temp = new Operand(lhs->StrVal().find(rhs->StrVal()) != string::npos);
                    operands.push(temp);
                }
                //else // NOT defined for numeric operands, ignore

                break;


            case '&': // Logical AND
                operands.push(new Operand(lhs->NumVal() && rhs->NumVal()));
                break;

            case '|': // Logical OR
                operands.push(new Operand(lhs->NumVal() || rhs->NumVal()));
                break;

            default:
                ctxt.SetErrorMessage("Unknown arithmetic/logical operator '" + token + "' encountered");
                return noResult;
            }

            delete lhs;
            delete rhs;
        }
    }

    if (operands.empty())
    {
        ctxt.SetErrorMessage("Illformed expression, no result left on stack at end");
        return noResult; // Underflow
    }

    else if (operands.size() > 1)
    {
        ctxt.SetErrorMessage("Illformed expression, too many results left on stack at end");
        return noResult; // Underflow
    }

    // Get the result, and convert to a string representation (no trailing zero's)

    string resultStr;
    Operand* result = operands.top();
    if (result->IsString())
        resultStr = result->StrVal();
    else
    {
        char resStr[32];
        sprintf(resStr, "%g", result->NumVal());
        resultStr = string(resStr);
    }

    delete result;
    return resultStr;
}


// Process next character in infix stream to postfix stream conversion
// In the case of a two character operator (!=, <= et al), the iterator 'it'
// will be incremented and two characters processed.
bool TExpression::_ProcessChar(string::iterator& it, string::iterator endPos)
{
    static string twoCharOperators = "!<>"; // Initial char of a two char operator (maybe)

    char c = ((it != endPos) ? *it : '\0'); // NULL iterator used to flush stack.
    string oper (1, c);
    string::iterator tmp_it=it;//tmp pointer to next char in string
    tmp_it++;
    if (twoCharOperators.find_first_of(c) != string::npos &&
            ((tmp_it++) != endPos && *(tmp_it++) == '='))
    {
        oper += *++it; // handle two character operators: <=, >= or !=
    }


    int priority; // Operator priority
    if ((priority = _Priority(oper)) == 0)
    {
        // Operand, or part there of, append to current Operand
        _operand += c;
        return true;
    }

    // Operator encountered, flush current operand (if any)
    if (_operand.length())
    {
        _postfix.push_back(_operand);
        _operand = "";
    }

    if (_opStack.empty() || c == '(')
    {
        // Empty stack, or start of sub-expression
        _opStack.push(oper);
        return true;
    }

    if (c == ')')
    {
        // End of sub-expression,
        // Pop & copy to output until '(', then discard parentheses
        while ( !_opStack.empty() && _opStack.top() != "(" )
        {
            _postfix.push_back(_opStack.top());
            _opStack.pop();
        }

        if (_opStack.empty())
        {
            _reason = "Bracket mismatch, too many ')'s";
            return false;
        }
        else
            _opStack.pop();
        return true;
    }

    if (priority > _Priority(_opStack.top()))
    {
        // New Operator has higher priority than last, so add to _opStack
        _opStack.push(oper);
        return true;
    }

    while ( (!_opStack.empty()) && (priority <= _Priority(_opStack.top())) )
    {
        // New Operator has lower priority than last, so
        // Pop & copy to output until lower priority operator is found
        _postfix.push_back(_opStack.top());
        _opStack.pop();
    }

    // Add new operator to _opStack, unless flushing stack.
    if (oper[0] != '\0')
        _opStack.push(oper);
    else
    {
        if ( !_opStack.empty() )
        {
            if ((_opStack.top())[0] == '(')
                _reason = "Bracket mismatch, too many '('s";
            else
                _reason = "Illformed expression, unable to complete parse";
            return false;
        }
    }

    return true;
}


// Determine operator precedence.
int TExpression::_Priority(const string& oper) const
{
    switch (oper[0])
    {
        // Arithmetic Operators ...
    case '!':
        if (oper.length() > 1 && oper[1] == '=')
            return 5;   // Inequality '!='

    case '/':
    case '*':
        return 8;

    case '+':
    case '-':
        return 7;

        // Relational Operators ...

    case '<':
    case '>':
        return 6;

    case '^':   // Contains
    case '=':
        return 5;

        // Logical Operators ...

    case '&':
        return 4;

    case '@':
        return 3;

    case '|':
        return 2;

        // Grouping Operators ...

    case '\0': // End of expression marker, used to flush stack at end.
        return 1;

    case '(':
    case ')':
        return -1;

        // Anything else, ie Operands ...
    default:
        return 0;
    }

    return 0; // should never get here!
}


bool TExpression::_DefinedForStrings(const string& oper) const
{
    switch (oper[0])
    {

    case '<':
    case '>':
    case '=':
    case '^': // Contains
        return true;
    case '!':
        return (oper.length() > 1 && oper[1] == '=');

    default:
        return false;
    }

    return false;
}
