/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// TEHttpRecords.h
// ---------------
//=========================================================================

#ifndef CYR_TEHTTPRECORDS_H
#define CYR_TEHTTPRECORDS_H

#include "HTTPconn.h"
#include <string>

#ifdef WIN32
#pragma warning(disable:4786)
using namespace std;
#endif

#include <TRecord.h>


///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdConnect | This class is used to Connect in an HTTP environment
//
// @base    public | TRecord.
//
class TRcrdConnect : public TRecord
{

    // @access Public Members.
public:
    // @cmember Set variable record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
TRcrdConnect (const TParamRule* headRuleP) : TRecord(headRuleP)
    , _exprP(NULL)
    {}
    // @cmember Destructor.

    ~TRcrdConnect ()
    {
        delete _exprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute (TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdConnect (*this);
    };

    // @access Private Members
private:
    // @cmember
    static const bool _registeredCreator;
    // @cmember
    static const bool _RegisterCreator();

    TtChrExpr* _exprP;
    static CTMutex m_mutex;
}
; // TRcrdConnect


/**
 * Definition of Request Record class
 **/
class TRcrdRequest : public TRecord
{

public:
    static const AtvfKey KEY;
    // Constructors & Destructor
TRcrdRequest(const TParamRule* headRuleP) : TRecord(headRuleP),
    _uriP(NULL), _headerP(NULL), _withP(NULL),
    _woutP(NULL), _bodyP(NULL)
    {}

    // @cmember Destructor.

    ~TRcrdRequest ()
    {
        if (_uriP) delete _uriP;
        if (_headerP) delete _headerP;
        if (_withP) delete _withP;
        if (_woutP) delete _woutP;
        if (_bodyP) delete _bodyP;
    }

    // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdRequest(*this);
    };
    // Mutators

    string getRequestString(int eType);

private:
    static const bool _registeredCreator;
    static const bool _RegisterCreator();
    TtChrExpr* _uriP;
    TtValList* _headerP;
    TtValList* _withP;
    TtValList* _woutP;
    TtChrExpr* _bodyP;
    static CTMutex evaluateMtx;

}
; // TRcrdRequest

/**
 * Definition of Disconnect Record class
 **/
class TRcrdDisconnect : public TRecord
{

public:
    static const AtvfKey KEY;
    // Constructors & Destructor
TRcrdDisconnect(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
    // Accessors

    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdDisconnect(*this);
    };
    // Mutators

private:
    static const bool _registeredCreator;
    static const bool _RegisterCreator();
}
; // TRcrdDisc


/**
 * Definition of HttpResponse Record class
 **/
class TRcrdHttpResponse : public TRecord
{

public:
    static const AtvfKey KEY;
    // Constructors & Destructor
TRcrdHttpResponse(const TParamRule* headRuleP) : TRecord(headRuleP)
    , _varP(NULL), _exprP(NULL), _identP(NULL), _varSize(0)
    {}
    // @cmember Destructor.

    ~TRcrdHttpResponse()
    {
        if (_varP) delete _varP;
        if (_exprP) delete _exprP;
        if (_identP) delete _identP;
    }
    // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdHttpResponse(*this);
    };
    // Mutators

private:
    int _varSize; // variable size

    // @cmember Destination label index in script
    int _labelIndex;
    // @cmember used to allow safe writes during run-phase
    static CTMutex m_mutex;
    // @cmember Pointer to set variable 'name' object.
    TtVarRef* _varP;
    // @cmember Pointer to set variable 'expression' object.
    CECTofTxt* _exprP;
    // @cmember Pointer to set variable 'identity' object.
    TtChrExpr* _identP;
    // @cmember
    static const bool _registeredCreator;
    // @cmember
    static const bool _RegisterCreator();
}
; // TRcrdHttpResponse


/**
 * Definition of Synchronize Record class
 **/
class TRcrdSynchronize : public TRecord
{

public:
    static const AtvfKey KEY;
    // Constructors & Destructor
TRcrdSynchronize(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
    // Accessors

    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdSynchronize(*this);
    };
    // Mutators

protected:
    int _labelIndex;            // Destination label index in script

private:
    static CTMutex m_mutex;

    static const bool _registeredCreator;
    static const bool _RegisterCreator();
}
; // TRcrdDisc


int server_specific_init(void);
int server_specific_cleanup(void);
#endif
