/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


///////////////////////////////////////////////////////////////////////////////
// @doc     TestLevel_i
// @module  TestLevel_i | This is the implementation of the interface to
//  the status of a single thread running a Test within a test manager
//
///////////////////////////////////////////////////////////////////////////////
#include <TPLog.h>
#include <TestLevel_i.h>
#include <CORBA/CORBAGeneral.h>
#include <TMCORBARepository.h>
#include <CECSpawnObject.h>
#include <ctime>
#include <iostream>
#include <strstream>
#include "mdbexcept.h"
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include "cxxutils.h"
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor
// @parm    TestName to be run
// @parm    Node on which test is being executed
// @parm    ThreadID for this process
TestLevelImpl::TestLevelImpl(string TestName,string Node,TestLevel_ptr parent,string ThreadID):m_no_executers_to_finish(0),m_closing_down(false)
{
    CTTime tm;
    tm.now();
    m_StartTime=tm.getSeconds();
    m_TestName=TestName;
    m_ScriptName=TestName;
    m_NodeName=Node;
    m_ident=ThreadID;
    m_ParentTestLevel=TestLevel::_duplicate(parent);
    m_TestCasePass=0;
    m_TestCaseFail=0;

}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor
TestLevelImpl::~TestLevelImpl()
{
    NodeExecuterMap_t::iterator nemi;
    TargetExecuterMap_t::iterator temi;
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli=(*temi).second.begin();
            while(teli!=(*temi).second.end())
            {
                try
                {
                    logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Test for TestExecuter existence");
                    //No simple 'ping' able type operation available on TestExecuters
                    //so use _non_existent (for omniORB this appears to suffice)
                    if((*teli)->_non_existent())
                    {
                        teli++;//assume gone, try next
                    }

                    else
                    {
                        CyrSleep(5000);//wait for 5 seconds
                    }

                }
                catch(...)//catch everything and treat as invalid TestExecuter
                {
                    teli++;//assume gone, try next
                }

            }
        }
    }
    try
    {
        CORBA::release(m_ParentTestLevel);
    }
    catch(...)
    {
        //Too late to worry
    }

    logr->AddTrace(DBGSTD,DEBUGLEVEL,"TestLevel Destructor");
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Call Stop on child TestLevels and Executers for close down
void TestLevelImpl::Stop()
{
    NodeExecuterMap_t::iterator nemi;
    TargetExecuterMap_t::iterator temi;
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli;
            for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
            {
                try
                {
                    logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Call Stop on TestExecuter");
                    (*teli)->Stop();
                    m_no_executers_to_finish++;//called Stop successfully so must wait for reply
                }

                catch(...)//catch everything and treat as invalid TestExecuter
                {}


            }
        }
    }
    //now wait for them all to reply that they are finished
    CTTime tm;
    tm.now();
    unsigned long StartTime=tm.getSeconds();
    while(m_no_executers_to_finish>0)
    {
        CTTime now;
        now.now();
        long t=now.getSeconds()-StartTime;
        if(t<0)t*=-1;
        if(t>300)break;//If waited longer than 5 mins then shutdown assume something gone wrong
        CyrSleep(5000);//Threads still running wait for 5 seconds
    }

}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return TestName
// @rdesc   TestName
char*  TestLevelImpl::TestName()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_TestName.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return Current ScriptName running in Thread
// @rdesc   ScriptName
char*  TestLevelImpl::ScriptName()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_ScriptName.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   set Current ScriptName running in Thread (not in CORBA interface)
// @parm    ScriptName
void TestLevelImpl::ScriptName(string scriptname)
{
    m_ScriptName=scriptname;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return Time Started as a long
// @rdesc   Start Time in number of seconds since 00:00:00 January 1, 1901 GMT.
CORBA::ULong  TestLevelImpl::StartTime()
{
    if(m_closing_down)throw CyrClosingDownException();
    return m_StartTime;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return Node on which Test is being executed
// @rdesc   Node Name
char*  TestLevelImpl::GetNode()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_NodeName.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc retrieves the note text
// @rdesc The last note message
char*  TestLevelImpl::GetNoteText()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_NoteText.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc Sets the note text,NOT a CORBA method internal use only
// @parm text for note message
void  TestLevelImpl::SetNoteText(string ntext)
{
    m_NoteText=ntext;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return Number of Child TestLevels being executed
// @rdesc   Number of Tests Being Executed by this 1 test level
CORBA::Long  TestLevelImpl::NoChildTestLevels(CORBA::Boolean  recursive)
{
    if(m_closing_down)throw CyrClosingDownException();
    CORBA::Long no_levels=0;
    if(recursive)//iterate through test levels and ask for same thing
    {
        TestLevelMap::iterator i;
        for(i=m_TestLevelMap.begin();i!=m_TestLevelMap.end();i++)
        {
            try
            {
                no_levels+=(*i).second->NoChildTestLevels(recursive);
            }
            catch(const CyrClosingDownException&)
            {
                //if closing down add 0
            }

            catch(...)
            {
                logr->AddTrace(DBGSTD,DEBUGLEVEL,"NoChildTestLevels caught an Exception, probably an abnormal termination of a TestManager/Executer, continuing");
            }
        }
    }
    no_levels+=m_TestLevelMap.size();
    return no_levels;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return Number of Threads that are being executed
// @rdesc   Number of Test Threads Being Executed by this 1 test level
CORBA::Long  TestLevelImpl::NoChildThreads(CORBA::Boolean  recursive)
{
    if(m_closing_down)throw CyrClosingDownException();
    logr->AddTrace(DBGSTD,DETAILLEVEL,"Called NoChildThreads");
    CORBA::Long no_levels=0;
    if(recursive)//iterate through test levels and ask for same thing
    {
        TestLevelMap::iterator i;
        for(i=m_TestLevelMap.begin();i!=m_TestLevelMap.end();i++)
        {
            try
            {
                no_levels+=(*i).second->NoChildThreads(recursive);
            }
            catch(const CyrClosingDownException&)
            {
                //do nothing
            }

            catch(...)
            {
                logr->AddTrace(DBGSTD,DEBUGLEVEL,"NoChildTestThreads caught an Exception, probably an abnormal termination of a TestManager/Executer, continuing");
            }
        }
    }
    NodeExecuterMap_t::iterator nemi;
    TargetExecuterMap_t::iterator temi;
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli;
            for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
            {
                try
                {
                    logr->AddTrace(DBGSTD,DETAILLEVEL,"Call GetNoThreads on Executer");
                    if(!((*teli)->_non_existent()))
                    {
                        if(!CORBA::is_nil((*teli)))
                        {
                            TestExecuter_var TEvar=TestExecuter::_duplicate((*teli));
                            no_levels+=TEvar->GetNoThreads(CORBA::String_var((const char*)"").inout());
                        }
                    }
                }
                catch(const CyrClosingDownException&)
                {
                    //do nothing
                }

                catch(...)
                {
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,"NoChildTestExecuters caught an Exception, probably an abnormal termination of a TestManager/Executer, continuing");
                }

            }
        }
    }
    logr->AddTrace(DBGSTD,DETAILLEVEL,"Leaving NoChildThreads");
    return no_levels;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return \'|\' separated list of unique identifiers for all Child tests
// @rdesc   list of identifiers
char*  TestLevelImpl::GetChildTestLevelIdents()
{
    TestLevelMap::iterator i;
    string strlist="";
    for(i=m_TestLevelMap.begin();i!=m_TestLevelMap.end();i++)
    {
        strlist+=(*i).second->ident();
        strlist+='|';
    }
    strlist+='|';
    return CORBA::string_dup(strlist.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Return reference to Child Test Level
// @parm    unique identifier for this Child test
// @rdesc   Child Test Level for the given identifier or TestLevel::_nil() if
//      invalid
TestLevel_ptr  TestLevelImpl::GetChildTestLevel(const char*  ident)
{
    if(m_TestLevelMap.find(ident)!=m_TestLevelMap.end())
        return  TestLevel::_duplicate(m_TestLevelMap[ident]);
    else
        return TestLevel::_nil();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   return details on list of ChildTestLevels for this Test
// @parm    TestLEvelList to be filled with information
// @rdesc   number of entries in list
CORBA::ULong  TestLevelImpl::GetChildTestLevelList(TestLevelList & TLList)
{
    if(m_closing_down)throw CyrClosingDownException();
    TestLevelMap::iterator i;
    int j=0;//counter within TLList
    for(i=m_TestLevelMap.begin();i!=m_TestLevelMap.end();i++)
    {
        TLList.length(TLList.length()+1);//increase size of list (probably not required)
        TLList[j]=(*i).second;
        j++;//increment TLList
    }

    return TLList.length();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Add a child test level to this test level
//      This checks if tests of this name have already been started
//      and assigns an identifier accordingly, checks for if a test has finished
//      can be done by seeing if Any tests of a name have ever been run
//      and if so if any are currently running
// @parm    Child Test Level to add, this will receive an identifier
void TestLevelImpl::AddChildTestLevel(TestLevel_ptr &  newTL_ptr)
{
    TestLevel_var newTL=TestLevel::_duplicate(newTL_ptr);
    int i;
    string name=newTL->TestName();
    //Record the starting of another Test
    if(m_TestsStartedMap.find(name)!=m_TestsStartedMap.end())
    {
        i=m_TestsStartedMap[name];
    }
    else
    {
        i=0;
    }
    m_TestsStartedMap[name]=++i;
    //record this test level against ident
    string s_ident=newTL->ident();
    m_TestLevelMap.insert(TestLevelMap::value_type(s_ident,newTL));
    //Record that this test is running
    if(m_TestsRunningMap.find(name)!=m_TestsRunningMap.end())
    {
        i=m_TestsRunningMap[name];
    }
    else
    {
        i=0;
    }
    m_TestsRunningMap[name]=++i;

}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Remove child test level because of completion or failure
// @parm    TestLevel to Remove (notice this is just removed as an active sub test
//      not necessarily a test level that no longer exists)
void TestLevelImpl::RemoveChildTestLevel(TestLevel_ptr & TL)
{
    string name=TL->TestName();
    m_TestsRunningMap[name]--;//decrement the number of tests running do not decrement number of tests started
    int i;
    if(m_TestsFinishedMap.find(name)!=m_TestsFinishedMap.end())
    {
        i=m_TestsFinishedMap[name];
    }
    else
    {
        i=0;
    }
    m_TestsFinishedMap[name]=++i;//increment number of test of this name that have finished
    m_TestLevelMap.erase(string(TL->ident()));//remove pointer to this testlevel
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Collect A List of TestExecuter Managed by this TestLevel
// @parm    TestExecuterList to fill
// @rdesc   number of executers in list

CORBA::ULong    TestLevelImpl::GetTestExecuterList(TestExecuterList & TElist)
{
    if(m_closing_down)throw CyrClosingDownException();
    NodeExecuterMap_t::iterator nemi;
    TargetExecuterMap_t::iterator temi;
    int j=0;
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli;
            for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
            {
                try
                {
                    TElist.length(TElist.length()+1);
                    TElist[j]=TestExecuter::_duplicate((*teli));
                    j++;//increment TEList
                }

                catch(const CyrClosingDownException&)
                {
                    TElist.length(TElist.length()-1);
                }
            }
        }
    }
    return TElist.length();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   return identifier
char* TestLevelImpl::ident()
{
    if(m_closing_down)throw CyrClosingDownException();
    return CORBA::string_dup(m_ident.c_str());
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Loop until no sub tests/threads are running, this goes through each
//      sub test/thread held in internal containers and checks each one to see if
//      it is still there. It does this by calling a method on the sub object.
//      If the object responds (i.e. no exceptions thrown) then the loop continues
//      if an exception is trown then the object is assumed to have stopped abnormaly
//      and it's entry is removed from the container. When the container is empty
//      no more obejcts are running. This method is for use when a TestLevel
//      wants to close down, before removing itself from it's parent it will
//      check it has nothing running.
void TestLevelImpl::WaitForEmpty()
{
    WaitForNoSubTests();
    WaitForNoSubExecuters();
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc loop until no valid SubTests are running, called from waitforempty
void TestLevelImpl::WaitForNoSubTests()
{
    //if a test has been started
    if(!m_TestsStartedMap.empty())
    {
        //wait until at least one test has registered itself as running or finished
        while(m_TestsRunningMap.empty() && m_TestsFinishedMap.empty() );
    }
    //then wait for all to finish
    while(!m_TestLevelMap.empty())
    {
        //get identifier of beginning of map
        string l_ident=(*(m_TestLevelMap.begin())).first;
        //try to call method on object in beginning of map
        string obj_ident;
        try
        {
            obj_ident=(*(m_TestLevelMap.begin())).second->ident();
            CyrSleep(5000);//Tests still running wait for 5 seconds
        }

        catch(CORBA::Exception)//another communication problem. do the same for now
        {
            //erase by key
            try
            {
                m_TestLevelMap.erase(l_ident);
            }
            catch(...)
            {
                //just do it, clearing down of invalid _var type might
                //cause all sorts of problems
                logr->AddTrace(DBGSTD,DEBUGLEVEL,"WaitForEmpty:caught Exception, TestManager May require Manual intervaention to stop");
            }
        }
        //any other exceptions to be treated as normal and handled as such
    }

}
///////////////////////////////////////////////////////////////////////////////
// @mfunc loop until no valid SubExecuters are running, called from waitforempty
void TestLevelImpl::WaitForNoSubExecuters()
{
    int no_threads=0;
    string obj_ident;

    NodeExecuterMap_t::iterator nemi;
    TargetExecuterMap_t::iterator temi;
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli;
            for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
            {
                logr->AddTrace(DBGSTD,DEBUGLEVEL,"WaitForNoSubExecuters Threads");
                //try to call method on object in beginning of map
                try
                {
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,"WaitForNoSubExecuters Thread:Get Name");
                    //Check that TestExecuter is still valid
                    TestExecuter_var TEx= TestExecuter::_duplicate(*teli);
                    obj_ident=TEx->Name();
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,obj_ident.c_str());
                    //Check Number of Threads running within this
                    //TestExecuter and loop while greater than 0
                    CORBA::ULong no_threads=0;
                    while((no_threads=TEx->GetNoThreads(CORBA::String_var((const char*)"").inout()))>0)
                    {
                        CyrSleep(10000);//Thread still running wait for 10 seconds (Too long??)
                    }

                }
                catch(CORBA::Exception)//communication problem. treat as executer has died abnormally, move on to next one
                {
                    //log that there was a problem
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,"WaitForNoSubExecuters:caught Exception, TestManager May require Manual intervention to stop(Getting number of threads)");
                }
                //any other exceptions to be treated as normal and handled as such
            }

        }
    }
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli;
            for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
            {
                try
                {
                    TestExecuter_var TEx= TestExecuter::_duplicate(*teli);
                    obj_ident=TEx->Name();
                    string msg="Call Stop on "+obj_ident;
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,msg.c_str());
                    TEx->Stop();//now finished tell it to shut down
                    m_no_executers_to_finish++;//called Stop successfully so must wait for reply
                }

                catch(CORBA::Exception)//communication problem. treat as executer has died abnormally, move on to next one
                {
                    //log that there was a problem
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,"WaitForNoSubExecuters:caught Exception, TestManager May require Manual intervention to stop(Calling Stop)");
                }
                //any other exceptions to be treated as normal and handled as such
            }

        }
    }

    //now wait for them all to reply that they are finished
    CTTime tm;
    tm.now();
    unsigned long StartTime=tm.getSeconds();
    while(m_no_executers_to_finish>0)
    {
        CTTime now;
        now.now();
        long t=now.getSeconds()-StartTime;
        if(t<0)t*=-1;
        if(t>60)
        {
            logr->AddTrace(DBGSTD,DEBUGLEVEL,"Timed out waiting for TestExecuter");
            break;//If waited longer than 1 min then shutdown assume something gone wrong
        }

        CyrSleep(5000);//Threads still running wait for 5 seconds
    }

}

TestLevel_ptr  TestLevelImpl::GetParentTestLevel()
{
    if(m_closing_down)throw CyrClosingDownException();
    return TestLevel::_duplicate(m_ParentTestLevel);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Is named test running
// @rdesc   true if running
// @parm    name of test
bool TestLevelImpl::isTestRunning(const string& testname)
{
    std::map<string,int>::iterator i=m_TestsStartedMap.find(testname);
    if(i==m_TestsStartedMap.end())return false;//never started so isn't running
    int no_started=(*i).second;
    std::map<string,int>::iterator j=m_TestsFinishedMap.find(testname);
    if(j==m_TestsFinishedMap.end())return true;//has started but hasn't finished
    int no_finished=(*j).second;
    if(no_finished<no_started)return true;//some left
    return false;//all must be finished

}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Is named Thread running
// @rdesc   true if running
// @parm    name of Thread

bool TestLevelImpl::isThreadRunning(const string& scriptname)
{
    CORBA::ULong no_threads=0;
    CORBA::String_var SName=scriptname.c_str();

    NodeExecuterMap_t::iterator nemi;
    TargetExecuterMap_t::iterator temi;
    for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
    {
        for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
        {
            // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
            // iterate throught each TestExecuterList_t.
            TestExecuterList_t::iterator teli;
            for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
            {
                try
                {
                    TestExecuter_var TEx=(*teli);
                    no_threads+=TEx->GetNoThreads(SName.inout());
                }
                catch(...)//catch everything and treat as invalid TestExecuter
                {}


            }
        }
    }
    if(no_threads>0)
        return true;
    else
        return false;
}
void  TestLevelImpl::AddTestCase(TestCaseStatus status)
{
    switch(status)
    {
    case PASSED:
        m_TestCasePass++;
        break;
    case FAILED:
        m_TestCaseFail++;
        break;
    default:
        break;
    }
}
CORBA::Long  TestLevelImpl::GetTestCaseResults(TestCaseStatus  Status, CORBA::Boolean  recursive)
{
    if(m_closing_down)throw CyrClosingDownException();
    CORBA::Long no_testcases=0;
    if(recursive)//iterate through test levels and ask for same thing
    {
        TestLevelMap::iterator i;
        for(i=m_TestLevelMap.begin();i!=m_TestLevelMap.end();i++)
        {
            try
            {
                no_testcases+=(*i).second->GetTestCaseResults(Status,recursive);
            }
            catch(const CyrClosingDownException&)
            {
                //if TestLevel closing down then add 0 (do nothing)
            }

            catch(...)
            {
                logr->AddTrace(DBGSTD,DEBUGLEVEL,"GetTestCaseResults caught an Exception, probably an abnormal termination of a TestManager/Executer, continuing");
            }
        }
        NodeExecuterMap_t::iterator nemi;
        TargetExecuterMap_t::iterator temi;
        for(nemi=m_NodeExecuterMap.begin();nemi!=m_NodeExecuterMap.end();nemi++)
        {
            for(temi=(*nemi).second.begin();temi!=(*nemi).second.end();temi++)
            {
                // TestExecuterMap iterators point to a pair<TARGETID,TestExecuterList_t>
                // iterate throught each TestExecuterList_t.
                TestExecuterList_t::iterator teli;
                for(teli=(*temi).second.begin();teli!=(*temi).second.end();teli++)
                {
                    try
                    {
                        no_testcases+=(*teli)->GetTestCaseResults(Status);
                    }
                    catch(const CyrClosingDownException&)
                    {
                        //if TestLevel closing down then add 0 (do nothing)
                    }

                    catch(...)
                    {
                        logr->AddTrace(DBGSTD,DEBUGLEVEL,"GetTestCaseResults caught an Exception, probably an abnormal termination of a TestManager/Executer, continuing");
                    }
                }
            }
        }
    }
    switch (Status)
    {
    case PASSED:
        no_testcases+=m_TestCasePass;
        break;
    case FAILED:
        no_testcases+=m_TestCaseFail;
        break;
    default:
        break;
    }
    return no_testcases;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Execute a thread using a script
// @parm    Script To execute
// @parm    Thread definition to use in execution
// @parm    Parent TestLevel to which thread will belong
// @rdesc   0 for success
CORBA::Long  TestLevelImpl::ExecuteThreadScript(const char*  scriptname, const char* nodename,const ThreadDef & thrddef, TestLevel_ptr & TL)
{
    string Node=nodename;//Save nodename for later use in Creating Executer
    TestLevel_var TL_var=TestLevel::_duplicate(TL);
    // Get A TestExecuter for the correct Node
    TargetExecuterMap_t::iterator temi;
    NodeExecuterMap_t::iterator nemi=m_NodeExecuterMap.find(Node);
    //if there wasn't one
    if(nemi==m_NodeExecuterMap.end())
    {
        logr->AddTrace(DBGSTD,DEBUGLEVEL,"Create new Node list of executers");
        //Create a brand new node list of executers
        TargetExecuterMap_t newmap;
        //insert returns a pair<iterator,bool> the iterator should point to the new list we have created
        nemi=m_NodeExecuterMap.insert(NodeExecuterMap_t::value_type(Node,newmap)).first;
        logr->AddTrace(DBGSTD,DEBUGLEVEL,"Create new list of executers");
        //Create a brand new list of executers
        TestExecuterList_t newlist;
        //insert returns a pair<iterator,bool> the iterator should point to the new list we have created
        temi=(*nemi).second.insert(TargetExecuterMap_t::value_type((TARGETID)thrddef.Target,newlist)).first;
    }
    else
    {
        // Get A TestExecuter for correct Target
        temi=(*nemi).second.find((TARGETID)thrddef.Target);
        // if there wasn't one
        if(temi==(*nemi).second.end())
        {
            logr->AddTrace(DBGSTD,DEBUGLEVEL,"Create new list of executers");
            //Create a brand new list of executers
            TestExecuterList_t newlist;
            //insert returns a pair<iterator,bool> the iterator should point to the new list we have created
            temi=(*nemi).second.insert(TargetExecuterMap_t::value_type((TARGETID)thrddef.Target,newlist)).first;
        }
        else
        {
            // reverse loop through all executers of this target,
            // should give us the last one we used (i.e. most likely to have a free thread)
            TestExecuterList_t::reverse_iterator teli;
            for(teli=(*temi).second.rbegin();teli!=(*temi).second.rend();teli++)
            {
                // Execute a thread on this executer
                try
                {
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,"Execute Thread on TestExecuter2");
                    CORBA::String_var cb_projectname=CORBA::string_dup(Repository->GetProjectName().c_str());
                    CORBA::String_var cb_scriptname=CORBA::string_dup(scriptname);
                    CORBA::Long  status=(*teli)->ExecuteThread(cb_projectname.in(),cb_scriptname.in(),  TL_var.inout(),thrddef);
                    if(status==0)return status;// if all O.K break out of here now
                    else
                    {
                        logr->AddTrace(DBGSTD,DEBUGLEVEL,"TestExecuter Failed to Execute Thread");
                        throw mdbParseException("TestExecuter Failed to Execute Thread");
                    }
                }
                catch( const CyrNoFreeThreadException& )
                {
                    logr->AddTrace(DBGSTD,DEBUGLEVEL,"TestExecuter Full !!");
                    // if not successful due to lack of threads try another one!
                }

            }
        }
    }
    // if no room on any executers create new one and execute thread
    TestExecuter_var TE=CreateNewTestExecuter((TARGETID)thrddef.Target,Node);
    if(CORBA::is_nil(TE))return -1;
    (*temi).second.push_back(TE);
    CORBA::Long  status=TE->Initialise(Repository->GetDataServerNodeName().c_str()
                                       ,Repository->GetUserName().c_str()
                                       ,Repository->GetPassword().c_str()
                                       ,Repository->GetTestRunString().c_str());
    logr->AddTrace(DBGSTD,DEBUGLEVEL,"Execute Thread on TestExecuter1");
    CORBA::String_var cb_projectname=CORBA::string_dup(Repository->GetProjectName().c_str());
    CORBA::String_var cb_scriptname=CORBA::string_dup(scriptname);
    status=TE->ExecuteThread(cb_projectname.in(),cb_scriptname.in(), TL_var.inout(),thrddef);
    if(status!=0)
    {
        logr->AddTrace(DBGSTD,DEBUGLEVEL,"TestExecuter Failed to Execute Thread");
        throw mdbParseException("TestExecuter Failed to Execute Thread");
    }
    // if just not successful return error
    return status;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Spawn a new test executer for appropriate target
// @parm    Target
// @rdesc   New TestExecuter
TestExecuter_ptr TestLevelImpl::CreateNewTestExecuter(TARGETID tgt,const string& nodename)
{
    char* c_tlevel=getenv(TP_TRACE_LEVEL);
    string tlevel="";
    if(c_tlevel!=NULL)
    {
        tlevel="-v "+string(c_tlevel);
    }
    else
    {
        tlevel=" -v 20";
    }
    char* image_ptr=new char[512];
    sprintf(image_ptr,"Start Target Executer %d",(long)tgt);
    logr->AddTrace(DBGSTD,DEBUGLEVEL,image_ptr);
    delete image_ptr;
    CECNodeDefinition nodedef=Repository->GetNodeDefinition(nodename);
    string msgstring ="Run on Node " + nodedef.GetNodeIPAddrAsString();
    logr->AddTrace(DBGSTD,DEBUGLEVEL,msgstring.c_str());
    CECSpawnObject* _SpawnObject=new CECSpawnObject(nodedef,"Engines",TEXECUTER,tgt,tlevel.c_str(),TestExecuter::CLASS_NAME,nodedef.GetNodeIPAddrAsString(),pCServer);
    logr->AddTrace(DBGSTD,DEBUGLEVEL,"Created Spawn Object");
    CORBA::Object_ptr obj=_SpawnObject->RetrieveObject();
    logr->AddTrace(DBGSTD,DEBUGLEVEL,"Retrieved Object");
    TestExecuter_var _TestExecuter = narrow_object((TestExecuter*)NULL,obj);
    logr->AddTrace(DBGSTD,DEBUGLEVEL,"Converted Object to TestExecuter");
    delete _SpawnObject;
    return TestExecuter::_duplicate(_TestExecuter);
}
void TestLevelImpl::TestExecuterFinished()
{
    m_no_executers_to_finish--;
}
